#include "LDPC.h"

/************************************************************************************************************************
 * NAME : void init_H_para(short *mat_para, FILE *f_H_data)
 * ABSRTACT : initilize the matrix parameter by reading the data from the file 
 * ENTRY : input : FILE *f_H_data : the exited file download from the Mackay web 
 	   output : short *mat_para : contain these parameters,see detail in program 
 * return: void
 *************************************************************************************************************************/
void init_mat_para(short *mat_para, FILE *f_H_data)
{
	short COLUMN,H_ROW;
	short Max_row_weight, Max_col_weight;//the max weight of the row and column,also suitable for irregular GF(2) LDPC
	
        fscanf(f_H_data,"%hd",&COLUMN);//read from the file
	fscanf(f_H_data,"%hd",&H_ROW);
	fscanf(f_H_data,"%hd",&Max_col_weight);
	fscanf(f_H_data,"%hd",&Max_row_weight);	
	
	mat_para[0] = COLUMN;//matrix column for both H and G
	mat_para[1] = H_ROW;//matrix row for H
	mat_para[2] = mat_para[0] - mat_para[1];//G_ROW : row of Generator matrix
	mat_para[3] = Max_col_weight;
	mat_para[4] = Max_row_weight;
}

/************************************************************************************************************************
 * NAME : void array_sort (short *in , short length , short init_min)
 * ABSTRACT : the total function aims to sort the series in decensend 
 * ENTRY : input : short *in;
                   short length;
                   short init_min;
           output :short *in;
 * RETURN : void
 ***********************************************************************************************************************/
void array_sort (short *in , short length , short init_min)
{
	short i;
	short j;
	short temp;
	short min;
	short min_index;
	
	for(i=0;i<length;i++)
	{
		min = init_min;
		
		//sort the min among the rest datas,record the index
		for(j=i;j<length;j++)
		{
			if(in[j]<min)
			{
				min = in[j];
				min_index  = j;
			}
		}
		
		//change the datas
		temp = in[i];
		in[i] = min;
		in[min_index] = temp;
	}
}

/*************************************************************************************************************************
 * NAME : void Construct_H_Matrix(short *mat_para,short *mat_col,short *mat_row,FILE *f_H_data)
 * ABSTRACT : this function is to construct the H matrix by tracking its 1 postion,not the whole matrix 
 * ENTRY : input : short *mat_para : the same as function init_mat_para()
                   FILE *f_H_data  : the same as function init_mat_para()
           output : short *mat_col : tracking the weigh of '1' in H matrix's column and its postion
                    short *mat_row : tracking the weigh of '1' in H matrix's row and its postion
 * RETURN : void
 ************************************************************************************************************************/
void Construct_H_Matrix(short *mat_para,short *mat_col,short *mat_row,FILE *f_H_data)
{
	short i,j;
	short temp;
	short pos;
	short COLUMN,G_ROW,H_ROW;
	short Max_row_weight , Max_col_weight;
	
	COLUMN = mat_para[0];
	H_ROW  = mat_para[1];
	G_ROW  = mat_para[2];
	Max_col_weight = mat_para[3];
	Max_row_weight = mat_para[4];
	
	//store each column's weight to mat_col
	for(i=0;i<COLUMN;i++)
	{
		fscanf(f_H_data,"%hd",&temp);
		mat_col[i*(Max_col_weight+1)] = temp;		
	}
        
        //store each row's weight to mat_row
	for(i=0;i<H_ROW;i++)
	{
		fscanf(f_H_data,"%hd",&temp);
		mat_row[i*(Max_row_weight+1)] = temp;
	}
	
	//store each column's 1 postion to mat_col in descendly order 
	for(i=0;i<COLUMN;i++)
	{
		temp = mat_col[i*(Max_col_weight+1)];
		
		for(j=1;j<=temp;j++)
		{
			fscanf(f_H_data,"%hd",&pos);
			if(pos==0) j--;
			else mat_col[i*(Max_col_weight+1)+j] = pos-1;
		}
		
		array_sort(&mat_col[i*(Max_col_weight+1)+1],temp,COLUMN);//descendly order
	}
	
	//store each row's 1 postion to mat_row in descendly order 
	for(i=0;i<H_ROW;i++)
	{
		temp = mat_row[i*(Max_row_weight+1)];
		
		for(j=1;j<=temp;j++)
		{
			fscanf(f_H_data,"%hd",&pos);
			if(pos==0) j--;
			else mat_row[i*(Max_row_weight+1)+j] = pos-1;
		}
		
		array_sort(&mat_row[i*(Max_row_weight+1)+1],temp,COLUMN);//descendly order
	}
}

/************************************************************************************************************************
 * NAME : void gauss_elimate_to_P_I (short *mat_col , short *mat_row , short **t , short *rank , short *mat_para)
 * ABSTRACT : this function is aimed to complete gauss elimate
 * ENTRY : input  : short *mat_para : contain these parameters
                    short *mat_col  : tracking the weigh of '1' in H matrix's column and its row postion for each column
                    short *mat_row  : tracking the weigh of '1' in H matrix's row and its column postion for each row
           output : 
           	    short *mat_col  : tracking the weigh of '1' in H matrix's column and its row postion for each column after ordering
                    short *mat_row  : tracking the weigh of '1' in H matrix's column and its column postion for each row after ordering 	
           	    short **t : H matrix after reorder
                    short *rank : 
 * RETURN : void
 ************************************************************************************************************************/		
void gauss_elimate_to_P_I (short *mat_col , short *mat_row , short **t , short *rank , short *mat_para)
{
	int	i;
	int j;
	int k;
	short  reorder_flag=0;
	short  temp;
	short  delta;
	short  COLUMN;
	short  G_ROW;
	short  H_ROW;
	short  Max_row_weight;
	short  Max_col_weight;
	short  *temp_col;
	
	COLUMN = mat_para[0];
	H_ROW  = mat_para[1];
	G_ROW  = mat_para[2];
	Max_col_weight = mat_para[3];
	Max_row_weight = mat_para[4];

	if ((temp_col = (short *) malloc(sizeof(short)*(Max_col_weight+1))) == NULL)
	{
		printf("can not malloc \n");
	}
	
	while(reorder_flag>=0)
	{
		reorder_flag = -1;
		
		memset((*t),0,sizeof(short)*H_ROW*COLUMN);//(* t)[] is a pointer varibles pointing to a array contain H_ROW*COLUMN element
		
		//build the H matrix
		for(i=0;i<COLUMN;i++)
		{
			temp = mat_col[i*(Max_col_weight+1)];//Max_col_weight of this column
			for(j=1;j<=temp;j++)
			{
				(*t)[ mat_col[i*(Max_col_weight+1)+j] * COLUMN + i ] = 1;//mat_col[i*(Max_col_weight+1)+j] is the 1's row position
			}
		}//so we have the H matrix here
		
		//try to convert the matrix to [P|I]
		for(i=1;i<=H_ROW;i++)
		{
			//detect if the t[H_ROW-i][COLUMN-i]==1,and try to set it
			if((*t)[(H_ROW-i+1)*COLUMN-i] == 0)
			{
				j = i+1;
				while(j<=H_ROW)
				{
					if((*t)[(H_ROW-j+1)*COLUMN-i] == 1)
					{
						for(k=0;k<COLUMN;k++)
						{
							(*t)[(H_ROW-i)*COLUMN + k] 
							= (*t)[(H_ROW-i)*COLUMN + k] ^ (*t)[(H_ROW-j)*COLUMN + k];
						}	
						j = H_ROW;//if there is 1 in the upper tringle matrix,end up loop;
					}
					j++;
				}
			}
			//detect if the t[H_ROW-i][COLUMN-i]==1,and try to set it
			
			//after setting through above operations
			//detect whether t[H_ROW-i][COLUMN-i] set successfully
			//if not succeed, set the reorder_flag = index of that column
			if(((*t)[(H_ROW-i+1)*COLUMN-i]==0)&&(reorder_flag<0))
			{
				reorder_flag = COLUMN-i;//can know the index of that column
			}
			
			//cancel the other 1 position in (COLUMN-i) column of the upper tringle matrix
			for(j=i+1;j<=H_ROW;j++)
			{
				if((*t)[(H_ROW-j+1)*COLUMN-i] == 1)
				{
					for(k=0;k<COLUMN;k++)
					{
						(*t)[(H_ROW-j)*COLUMN + k]
					        = (*t)[(H_ROW-j)*COLUMN + k] ^ (*t)[(H_ROW-i)*COLUMN + k];
					}
					        
				}
			}
			//cancel the other 1 position in (COLUMN-i) column of the upper tringle matrix
		}
		//try to convert the matrix to [P|I]
		//so there are all 0 in (COLUMN-i) column of the upper tringle matrix
		//H_matrix becomes a nether triangle matrix	
			
		//construct [P|I]
		for(j=G_ROW;j<COLUMN;j++)
		{
			for(i=(j-G_ROW+1);i<H_ROW;i++)
			{
				if((*t)[i*COLUMN + j] == 1)
				{
					for(k=0;k<COLUMN;k++)
					{
						(*t)[i*COLUMN + k]
						= (*t)[i*COLUMN + k] ^ (*t)[(j-G_ROW)*COLUMN + k];
					}
				}
			}
		}
		//construct [P|I]
		
		//detect all zero row 
		i=0;
		j=0;
		temp=0;
		delta=0;
		
		while(i<H_ROW)
		{
			j=0;
			while(j<COLUMN)
			{
				if((*t)[i*COLUMN + j] !=0)
				{
					j = COLUMN;
					delta ++ ;//if the row number which has 1, delta ++,then escape this row
				}
				
				j++;
			}
			
			if((delta-temp)==0)
			{
				delta++;
				temp++;//the zero row number
			}
			
			i++;
		}
		
		(*rank) = temp;//the number of zero row
		
		//right to be the first rows,so no need to reord culumn
		if((COLUMN-reorder_flag+temp-1)==H_ROW)		reorder_flag=-1;
	        //detect all zero row
	        
	        //for test
		//printf("%d\n",reorder_flag);
		//for test
		
		//if reorder_flag ,reorder column
		if(reorder_flag>=0)		
		{
			//store the all zeros row's postion temparoly
			for(j=0;j<=Max_col_weight;j++)
				temp_col[j] = mat_col[reorder_flag*(Max_col_weight+1)+j];
                        
                        //column shift
			for(k=(reorder_flag-1);k>=0;k--)
				for(j=0;j<=Max_col_weight;j++)
					mat_col[(k+1)*(Max_col_weight+1)+j] = mat_col[k*(Max_col_weight+1)+j];

			for(j=0;j<=Max_col_weight;j++)	mat_col[j] = temp_col[j];
		}
		//if reorder_flag ,reorder column

	} //end of while 

	free(temp_col);
}

/****************************************************************************************************************
 * NAME : void Construct_G_Matrix (short *mat_col , short *mat_row , short **p_gen, short *mat_para)
 * ABSTRACT : constuct G matrix
 * ENTRY : input : short *mat_col :
                   short *mat_row :
                   short *mat_para:
           output : short **p_gen : G matrix
 * RETURN : void
 ****************************************************************************************************************/
void Construct_G_Matrix (short *mat_col , short *mat_row , short **p_gen, short *mat_para)
{
        int	i,j,k;
	short   *t;
	short	*gen;
	short   reorder_flag=0,temp;
	short   COLUMN , G_ROW , H_ROW;
	short   Max_row_weight , Max_col_weight;

	COLUMN = mat_para[0];
	H_ROW  = mat_para[1];
	G_ROW  = mat_para[2];
	Max_col_weight = mat_para[3];
	Max_row_weight = mat_para[4];
	gen = *p_gen;

	if ((t = (short *) malloc(sizeof(short)*(H_ROW*COLUMN))) == NULL)
	{
		printf("can not malloc \n");
	}

	gauss_elimate_to_P_I(mat_col,mat_row,&t,&temp,mat_para);

	printf("the system check_sum matrix\n");

	/*construct generate matrix*/
	G_ROW = G_ROW + temp;
	mat_para[2] = G_ROW;

	if(temp!=0)
	{
		free(gen);
		if ((gen = (short *) malloc(sizeof(short)*(G_ROW*COLUMN))) == NULL)
			printf("can not malloc \n");
	}

	for(i=0;i<G_ROW;i++)
		for(j=0;j<G_ROW;j++)
			if(i==j)	gen[j*COLUMN+i]=1;
			else		gen[j*COLUMN+i]=0;	
	
	for(i=G_ROW;i<COLUMN;i++)
	{
		for(j=0;j<G_ROW;j++)
		{
			gen[j*COLUMN+i] = t[(i-G_ROW+temp)*COLUMN+j];			
		}				
	}	
	printf("the generate matrix\n");

	/*reconstruct the coordinate of checksum matrix row*/		
	//for(i=0;(i<H_ROW*COLUMN);i++)	t[i] = 0;

	memset(t,0,sizeof(short)*H_ROW*COLUMN);		

	for(i=0;i<COLUMN;i++)
	{
		temp = mat_col[i*(Max_col_weight+1)];
		for(j=1;j<=temp;j++)
			t[mat_col[i*(Max_col_weight+1)+j]*COLUMN + i] = 1;
	}

	for(i=0;i<H_ROW;i++)
	{
		k = 1;
		for(j=0;j<COLUMN;j++)
		{
			if(t[i*COLUMN + j] == 1)
			{
				mat_row[i*(Max_row_weight+1)+k] = j;
				k++;
			}				
		}
	}
	*p_gen = gen;
	printf("The G_matrix has been created\n");
	
	/*reconstruct the coordinate of checksum matrix row*/
	free(t);
	
}

/**************************************************************************************************************************
 * NAME     :  void LDPC_init(struct LDPCParam *p_ldpc_param, FILE *fpLDPC)
 * ABSTRACT :  initialize LDPC
 * ENTRY    :  input  : FILE *fpLDPC
 	       output : struct LDPCParam *p_ldpc_param 
 * RETURN   :  void
 *************************************************************************************************************************/
void LDPC_init(struct LDPCParam *p_ldpc_param, FILE *fpLDPC)
{
	short	mat_param[5];

	init_mat_para(mat_param, fpLDPC);
	
	p_ldpc_param->mat_param[0]	=	mat_param[0];//COLUMN
	p_ldpc_param->mat_param[1]	=	mat_param[1];//H_ROW
	p_ldpc_param->mat_param[2]	=	mat_param[2];//G_row
	p_ldpc_param->mat_param[3]	=	mat_param[3];//MAX_COLUMN_WEIGHT
	p_ldpc_param->mat_param[4]	=	mat_param[4];//MAX_ROW_WEIGHT
	p_ldpc_param->MatCol		=	(short*) malloc(mat_param[0]*(mat_param[3]+1)*sizeof(short));
	p_ldpc_param->MatRow		=	(short*) malloc(mat_param[1]*(mat_param[4]+1)*sizeof(short));
	p_ldpc_param->GMatrix		=	(short*) malloc(mat_param[0]*mat_param[2]*sizeof(short));
	p_ldpc_param->iterC		=	100;

	Construct_H_Matrix(p_ldpc_param->mat_param, p_ldpc_param->MatCol, p_ldpc_param->MatRow, fpLDPC);
	Construct_G_Matrix(p_ldpc_param->MatCol, p_ldpc_param->MatRow, &(p_ldpc_param->GMatrix), p_ldpc_param->mat_param);
	
}

void LDPC_encode(short *gen, short *input, short *output, short *mat_para)
{
	//printf("LDPC_encode started\n"); fflush(stdout);
	short  i , j;
	short  COLUMN , G_ROW , H_ROW;
	short  Max_row_weight , Max_col_weight;
	
	COLUMN = mat_para[0];
	H_ROW  = mat_para[1];
	G_ROW  = mat_para[2];
	Max_col_weight = mat_para[3];
	Max_row_weight = mat_para[4];

	//printf("COLUMN: %d, H_ROW: %d, G_ROW: %d\n", COLUMN, H_ROW, G_ROW); fflush(stdout);
	
	memset(output,0,sizeof(short)*COLUMN);

	for(i=0;i<COLUMN;i++)
		for(j=0;j<G_ROW;j++)
			output[i] = output[i]^(input[j]&gen[j*COLUMN+i]);	
 
	//printf("LDPC_encode ended\n"); fflush(stdout);
}

double saturate(double x, double Inf)
{
	double	tmp;
	
	tmp = x;
	
	if (x > Inf)
	{
		tmp = Inf;
	}
	if (x < -Inf)
	{
		tmp = -Inf;
	}
	
	return(tmp);
}

/**************************************************************************************************************************
 * NAME : short check_sum(short *mat_row, short *cc, short length, short row_weight)
 * ABSTRACT : check whether the parity equations satisfied  based on the encoded bits
 * ENTRY : input : short *mat_row :
                    short *cc : the encoded bits
                    short length : the check equations number
                    short row_weight : the max_row _weight
 * RETURN : flag whether the parity equations satisfied : flag = 0 ,satisfied 
 *************************************************************************************************************************/
short check_sum(short *mat_row, short *cc, short length, short row_weight)
{
	int    i,j;
    	short  temp , flag=0;
	short  weight , col;

	for(i=0;i<length;i++)
	{
		temp = 0;
		weight = mat_row[i*(row_weight+1)];
		for(j=1;j<=weight;j++)
		{
			col = mat_row[i*(row_weight+1)+j];
			temp = temp ^ cc[col];   
		}

		if (temp==1)	flag ++;
	}

	return(flag);
}
/**************************************************************************************************************************
 * NAME : double Horiz_compute_llr(double *vcQ, int row_weight, int ex_vode)
 * ABSTRACT : compute the extrinsic information passing from the check node j to the variable node i ; 
              extrinsic information is the LLR that the check node j is satified based on messages
              from connected variable nodes except i.
 * ENTRY : input : double *vcQ : array contain the extrinsic informations
                               passing from all variable nodes connected the check node to the check node.
                   int row_weight : the  Max_row_weight for current row.
                   int ex_vode: the variable node excluded.
 * RETURN :the extrinsic information passing from the check node j to the variable node i. 
 *************************************************************************************************************************/
/*double Horiz_compute_llr(double *vcQ, int row_weight, int ex_vnode)
{
	double	T;
	int	k;
	double	tmp;

	T = 1;

	for(k=0; k<row_weight; k++)
	{
		if (k != ex_vnode)
		{
			T = T*(1-exp(vcQ[k]))/(1+exp(vcQ[k]));
		}
	}
	
	tmp = log((1-T)/(1+T));
	
	tmp = saturate(tmp, 100);

	return(tmp);

}*/

/*
//The SPA decoder
double Horiz_compute_llr(double *vcQ, int row_weight, int ex_vnode)
{
	double	T,S;
	int	k;
	double	tmp,a,b;

	T = 1;
    S = 0;

	for(k=0; k<row_weight; k++)
	{
		if (k != ex_vnode)
		{
		    if(vcQ[k]>0)
			{
			    a = 1;
				b = vcQ[k];
		    }
			else
			{
				a = -1;
				b = -vcQ[k];
			}

			T = T*a; 
			 
			S = S + log((exp(b)+1)/(exp(b)-1));          
		}
	}
	
	tmp = T * log((exp(S)+1)/(exp(S)-1));
	
	tmp = saturate(tmp, 100);

	return(tmp);

}*/

/*
// the Min-Sum Decoder
double Horiz_compute_llr(double *vcQ, int row_weight, int ex_vnode)
{
	double	T;
	int	k;
	double	tmp,a,b,min;

	T = 1;
	min = 1000000;

	for(k=0; k<row_weight; k++)
	{
		if (k != ex_vnode)
		{
		    if(vcQ[k]>0)
			{
			    a = 1;
				b = vcQ[k];
		    }
			else
			{
				a = -1;
				b = -vcQ[k];
			}

			T = T*a;
			
			if(b<min)
			{
			    min = b;
			}
        
		}
	}
	
	tmp = T * min;
	
	tmp = saturate(tmp, 100);

	return(tmp);

}*/	

// the Min-Sum-Plus-Correction-Factor Decoder
double correct_factor(double x,double y)
{
     if((fabs(x+y)<2)&&(fabs(x-y)>(2*fabs(x+y))))
	        return 0.5;
	 else if( (fabs(x-y)<2) && ( (fabs(x+y)) > (2*fabs(x-y)) ) )
	        return -0.5;
	 else return 0.0;
}

double Horiz_compute_llr(double *vcQ, int row_weight, int ex_vnode)
{
	double	T;
	int	k;
	double	tmp,a,b,min,factor,tmpa,tmpb;

	T = 1;
	min = 1000000;
	tmp = T * min;

	for(k=0; k<row_weight; k++)
	{
		if (k != ex_vnode)
		{
		    factor = correct_factor(tmp,vcQ[k]);
		    
		    if(vcQ[k]>0)
			{
			    a = 1;
				b = vcQ[k];
		    }
			else
			{
				a = -1;
				b = -vcQ[k];
			}

		    if(tmp>0)
			{
			    tmpa = 1;
				tmpb = tmp;
		    }
			else
			{
				tmpa = -1;
				tmpb = -tmp;
			}

            if(b<tmpb)
			{
			    tmpb = b;
			}
			
             tmp = a * tmpa * tmpb + factor;      
		}
	}
	
	
	tmp = saturate(tmp, 100);

	return(tmp);

}  

 											
/**************************************************************************************************************************
 * NAME : double Verti_compute_llr(double *vcL, int col_weight, double f, int ex_cnode)
 * ABSTRACT : compute the extrinsic information passing from the variable node i to the check node j ; 
              extrinsic information is the LLR of the varible node i based on message from  all check node connected i except j 
 * ENTRY : input : double * vcL : array contain the extrinsic information passing from all check node to the varible node i
                   int col_weight : the  Max_clo_weight for current column.
                   double f : the current LLR of all varible nodes 
                   int ex_cnode : the check node excluded
 * RETURN : the extrinsic information passing from the variable node i to the check node j . 
 *************************************************************************************************************************/					
double Verti_compute_llr(double *cvL, int col_weight, double f, int ex_cnode)
{
	int	k;
	double	tmp;

	tmp = f;

	for(k=0; k<col_weight; k++)
	{
		if (k != ex_cnode)
		{
			tmp = tmp + cvL[k];
		}
	}

	return(tmp);
}

/**************************************************************************************************************************
 * NAME : 
 * ABSTRACT :
 * ENTRY : 
 * RETURN : 
 *************************************************************************************************************************/
void BP_decoding(double		*llr_in,
		 double		*llr_out,
		 short		*dec_out,
		 short		*mat_col,
		 short		*mat_row,
		 short		*mat_param,
		 short		iterN)
{
	double		*table_row;
	double		*table_col;

	short		*row_indx;
	short		*col_indx;

	short		Column, H_row, G_row, Max_col_weight, Max_row_weight;

	short		flag = 1;


	int		i,j,k,iter = 0;
	short		tmp_weight, tmp_col, tmp_row;

	double		tmp;

	//parameter initialize
	Column = mat_param[0];
	H_row = mat_param[1];
	G_row = mat_param[2];
	Max_col_weight = mat_param[3];
	Max_row_weight = mat_param[4];

	//printf("Column: %d, H_row: %d, G_row: %d, Max_col_weight: %d, Max_row_weight: %d\n", Column, H_row, G_row, Max_col_weight, Max_row_weight);
	fflush(stdout);

	
	table_col = (double*) malloc(Column*Max_col_weight*sizeof(double));
	table_row = (double*) malloc(H_row*Max_row_weight*sizeof(double));
	row_indx = (short*) malloc(H_row*sizeof(short));
	col_indx = (short*) malloc(Column*sizeof(short));


	//initialize the first llr
	for(i=0; i<H_row; i++)//for all check node 
	{
		//firstly,set the number of varible nodes connected to the check node
		tmp_weight = mat_row[i*(Max_row_weight+1)];
		
		for(j=1; j<=tmp_weight; j++)//for all the varible nodes connected to the check node
		{
			tmp_col = mat_row[i*(Max_row_weight+1)+j];// 1's postion in column for specified row
			table_row[i*Max_row_weight+j-1] = llr_in[tmp_col];//table_row contain the llr of variable node in each row
		}
	}

	printf("iterN: %d\n", iterN); fflush(stdout);
	while((flag>0)&&(iter<iterN))
	{
		memset(col_indx, 0, Column*sizeof(short));
		memset(row_indx, 0, H_row*sizeof(short));
		
		//Horizontal step
		for(i=0; i<H_row; i++)
		{
			tmp_weight = mat_row[i*(Max_row_weight+1)];//set the number of variable nodes connected to the check node
	                
	        for(j=1;j<=tmp_weight;j++)
	        {
	           //the extrinsic information passing from the check node i to the jth varible node
	           tmp = Horiz_compute_llr(&table_row[i*Max_row_weight], tmp_weight, j-1);
	           tmp_col = mat_row[i*(Max_row_weight+1) + j];//variable node is at tmp_col'th  column
	           table_col[tmp_col*Max_col_weight+col_indx[tmp_col]] = tmp;
	           col_indx[tmp_col]++;
	        }
	    }
	        
	    //Vertical step
		for(i=0; i<Column; i++)
		{
			tmp_weight = mat_col[i*(Max_col_weight+1)];//set the number of check nodes connected to the variable nodes
			
			for(j=1; j<=tmp_weight; j++)
			{       
				//the extrinsic information passing from the variable node i to the jth check node
				tmp = Verti_compute_llr(&table_col[i*Max_col_weight], tmp_weight, llr_in[i], j-1);
				tmp_row = mat_col[i*(Max_col_weight+1)+j];//check node is at tmp_col'th  row
				table_row[tmp_row*Max_row_weight+row_indx[tmp_row]] = tmp;
				row_indx[tmp_row]++;
			}
		}
		
		//compute pseudo llr
		for(i=0; i<Column; i++)
		{
			tmp = llr_in[i];
			tmp_weight = mat_col[i*(Max_col_weight+1)];
			for(k=0; k<tmp_weight; k++)
			{
				tmp = tmp + table_col[i*Max_col_weight+k];
			}
			llr_out[i] = tmp;
			if (tmp >0)
			{
				dec_out[i] = 1;
			}
			else 
			{
				dec_out[i] = 0;
			}
		}

		struct timeval tv1, tv2;
		time_t time1, time2;

		flag = check_sum(mat_row, dec_out, H_row, Max_row_weight);

		iter++;
		
	}
		                	
	if (flag == 0)
	{
		for(i=0; i<Column; i++)
		{
			llr_out[i] = (llr_out[i] > 0)? 1000:-1000;
		}
	}

	free(table_col);
	free(table_row);
	free(row_indx);
	free(col_indx);	                	
        //printf("done completed!\n"); fflush(stdout);
}				
