/* -*- c++ -*- */
/*
 * Copyright 2005 Free Software Foundation, Inc.
 * 
 * This file is part of GNU Radio
 * 
 * GNU Radio is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 * 
 * GNU Radio is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with GNU Radio; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street,
 * Boston, MA 02110-1301, USA.
 */

/*
 * See also ISO 3309 [ISO-3309] or ITU-T V.42 [ITU-V42] for a formal specification.
 */

//#define DEBUG_FEC 0

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <gr_crc32.h>

// apurv++ starts //
#include <stdlib.h>     // char.h combines with gr_crc32.h
#include <string.h>
#include <bitset>
#include <strings.h>
#include <stdio.h>      //for debugging
#include <vector>
#include <map>

#define NULL ((void *)0)
//#define DEBUG 0 
//#define DEBUG_LDPC 0

#define min(a,b)        ((a) < (b) ? (a) : (b)) //for decoding
// apurv++ ends //


// for file logging //
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>

#ifdef HAVE_IO_H
#include <io.h>
#endif
#ifdef O_BINARY
#define OUR_O_BINARY O_BINARY
#else
#define OUR_O_BINARY 0
#endif

#ifdef O_LARGEFILE
#define OUR_O_LARGEFILE O_LARGEFILE
#else
#define OUR_O_LARGEFILE 0
#endif
// end for file logging //

   
// Automatically generated CRC function
// polynomial: 0x104C11DB7
unsigned int
gr_update_crc32(unsigned int crc, const unsigned char *data, size_t len)
{
    static const unsigned int table[256] = {
    0x00000000U,0x04C11DB7U,0x09823B6EU,0x0D4326D9U,
    0x130476DCU,0x17C56B6BU,0x1A864DB2U,0x1E475005U,
    0x2608EDB8U,0x22C9F00FU,0x2F8AD6D6U,0x2B4BCB61U,
    0x350C9B64U,0x31CD86D3U,0x3C8EA00AU,0x384FBDBDU,
    0x4C11DB70U,0x48D0C6C7U,0x4593E01EU,0x4152FDA9U,
    0x5F15ADACU,0x5BD4B01BU,0x569796C2U,0x52568B75U,
    0x6A1936C8U,0x6ED82B7FU,0x639B0DA6U,0x675A1011U,
    0x791D4014U,0x7DDC5DA3U,0x709F7B7AU,0x745E66CDU,
    0x9823B6E0U,0x9CE2AB57U,0x91A18D8EU,0x95609039U,
    0x8B27C03CU,0x8FE6DD8BU,0x82A5FB52U,0x8664E6E5U,
    0xBE2B5B58U,0xBAEA46EFU,0xB7A96036U,0xB3687D81U,
    0xAD2F2D84U,0xA9EE3033U,0xA4AD16EAU,0xA06C0B5DU,
    0xD4326D90U,0xD0F37027U,0xDDB056FEU,0xD9714B49U,
    0xC7361B4CU,0xC3F706FBU,0xCEB42022U,0xCA753D95U,
    0xF23A8028U,0xF6FB9D9FU,0xFBB8BB46U,0xFF79A6F1U,
    0xE13EF6F4U,0xE5FFEB43U,0xE8BCCD9AU,0xEC7DD02DU,
    0x34867077U,0x30476DC0U,0x3D044B19U,0x39C556AEU,
    0x278206ABU,0x23431B1CU,0x2E003DC5U,0x2AC12072U,
    0x128E9DCFU,0x164F8078U,0x1B0CA6A1U,0x1FCDBB16U,
    0x018AEB13U,0x054BF6A4U,0x0808D07DU,0x0CC9CDCAU,
    0x7897AB07U,0x7C56B6B0U,0x71159069U,0x75D48DDEU,
    0x6B93DDDBU,0x6F52C06CU,0x6211E6B5U,0x66D0FB02U,
    0x5E9F46BFU,0x5A5E5B08U,0x571D7DD1U,0x53DC6066U,
    0x4D9B3063U,0x495A2DD4U,0x44190B0DU,0x40D816BAU,
    0xACA5C697U,0xA864DB20U,0xA527FDF9U,0xA1E6E04EU,
    0xBFA1B04BU,0xBB60ADFCU,0xB6238B25U,0xB2E29692U,
    0x8AAD2B2FU,0x8E6C3698U,0x832F1041U,0x87EE0DF6U,
    0x99A95DF3U,0x9D684044U,0x902B669DU,0x94EA7B2AU,
    0xE0B41DE7U,0xE4750050U,0xE9362689U,0xEDF73B3EU,
    0xF3B06B3BU,0xF771768CU,0xFA325055U,0xFEF34DE2U,
    0xC6BCF05FU,0xC27DEDE8U,0xCF3ECB31U,0xCBFFD686U,
    0xD5B88683U,0xD1799B34U,0xDC3ABDEDU,0xD8FBA05AU,
    0x690CE0EEU,0x6DCDFD59U,0x608EDB80U,0x644FC637U,
    0x7A089632U,0x7EC98B85U,0x738AAD5CU,0x774BB0EBU,
    0x4F040D56U,0x4BC510E1U,0x46863638U,0x42472B8FU,
    0x5C007B8AU,0x58C1663DU,0x558240E4U,0x51435D53U,
    0x251D3B9EU,0x21DC2629U,0x2C9F00F0U,0x285E1D47U,
    0x36194D42U,0x32D850F5U,0x3F9B762CU,0x3B5A6B9BU,
    0x0315D626U,0x07D4CB91U,0x0A97ED48U,0x0E56F0FFU,
    0x1011A0FAU,0x14D0BD4DU,0x19939B94U,0x1D528623U,
    0xF12F560EU,0xF5EE4BB9U,0xF8AD6D60U,0xFC6C70D7U,
    0xE22B20D2U,0xE6EA3D65U,0xEBA91BBCU,0xEF68060BU,
    0xD727BBB6U,0xD3E6A601U,0xDEA580D8U,0xDA649D6FU,
    0xC423CD6AU,0xC0E2D0DDU,0xCDA1F604U,0xC960EBB3U,
    0xBD3E8D7EU,0xB9FF90C9U,0xB4BCB610U,0xB07DABA7U,
    0xAE3AFBA2U,0xAAFBE615U,0xA7B8C0CCU,0xA379DD7BU,
    0x9B3660C6U,0x9FF77D71U,0x92B45BA8U,0x9675461FU,
    0x8832161AU,0x8CF30BADU,0x81B02D74U,0x857130C3U,
    0x5D8A9099U,0x594B8D2EU,0x5408ABF7U,0x50C9B640U,
    0x4E8EE645U,0x4A4FFBF2U,0x470CDD2BU,0x43CDC09CU,
    0x7B827D21U,0x7F436096U,0x7200464FU,0x76C15BF8U,
    0x68860BFDU,0x6C47164AU,0x61043093U,0x65C52D24U,
    0x119B4BE9U,0x155A565EU,0x18197087U,0x1CD86D30U,
    0x029F3D35U,0x065E2082U,0x0B1D065BU,0x0FDC1BECU,
    0x3793A651U,0x3352BBE6U,0x3E119D3FU,0x3AD08088U,
    0x2497D08DU,0x2056CD3AU,0x2D15EBE3U,0x29D4F654U,
    0xC5A92679U,0xC1683BCEU,0xCC2B1D17U,0xC8EA00A0U,
    0xD6AD50A5U,0xD26C4D12U,0xDF2F6BCBU,0xDBEE767CU,
    0xE3A1CBC1U,0xE760D676U,0xEA23F0AFU,0xEEE2ED18U,
    0xF0A5BD1DU,0xF464A0AAU,0xF9278673U,0xFDE69BC4U,
    0x89B8FD09U,0x8D79E0BEU,0x803AC667U,0x84FBDBD0U,
    0x9ABC8BD5U,0x9E7D9662U,0x933EB0BBU,0x97FFAD0CU,
    0xAFB010B1U,0xAB710D06U,0xA6322BDFU,0xA2F33668U,
    0xBCB4666DU,0xB8757BDAU,0xB5365D03U,0xB1F740B4U,
    };
  
    while (len > 0)
    {
      crc = table[*data ^ ((crc >> 24) & 0xff)] ^ (crc << 8);
      data++;
      len--;
    }
    return crc;
}

unsigned int
gr_update_crc32(unsigned int crc, const std::string s)
{
  return gr_update_crc32(crc, (const unsigned char *) s.data(), s.size());
}
    
unsigned int
gr_crc32(const unsigned char *buf, size_t len)
{
  return gr_update_crc32(0xffffffff, buf, len) ^ 0xffffffff;
}

unsigned int
gr_crc32(const std::string s)
{
  return gr_crc32((const unsigned char *) s.data(), s.size());
}

// convert the entire data into bits //
std::string
getBitString(const unsigned char *data, int len)
{
  std::string orig_s;
  for(int i = 0; i < len; i++)
  {
      std::bitset<sizeof(unsigned char) * 8> s_bits(data[i]);
      std::string tmp_string = s_bits.to_string<char,std::char_traits<char>,std::allocator<char> >();
      orig_s.append(tmp_string);
  }
  return orig_s;
}

// converts a bitstring back into a proper msg //
std::string
getMsg(std::string str, unsigned char *data)
{
  if(str.size() % 8 != 0)
	printf("%d ", str.size()); fflush(stdout);
  assert(str.size() % 8 == 0);

  int s_len = str.length();
  int k = 0;
#ifdef DEBUG
  printf("getMsg, s_len: %d\n", s_len); fflush(stdout);
#endif
  for(int i = 0; i < s_len; i=i+8)
  {
      std::bitset<sizeof(unsigned char)*8> s_bits(str.substr(i, 8));
      data[k] = (unsigned char) s_bits.to_ulong();
      k++;
  }

#ifdef DEBUG
  printf("Final Msg: \n");
  for(int i = 0; i < k; i++)
  {
     printf("%d ", data[i]);
  }
  printf("\n");
  printf("k: %d\n", k); fflush(stdout);
#endif

  std::string ret_s((const char*) data, k);
  return ret_s;
}

// -- alternate interleaving - bestN method //

// only used for the bestN strategy //
static void initializeSubcarriers(int *best_sc_arr, int *rem_sc_arr, int total_subcarriers, int n)
{
   int rem_subcarriers = total_subcarriers - n;
   for(int i = 0, j = 0; i < n * 2; i+=2, j++)
   {
       best_sc_arr[j] = i;
   }   
   
   int k = 0;
   for(int i = n; i < total_subcarriers; i++)
   {
       //printf("inserting sub%d at pos%d\n", k, i);
       rem_sc_arr[k++] = i;
   } 
}

static void initializeSubcarriersFromFile(int *best_sc_arr, int *rem_sc_arr, const int total_subcarriers, const int n)
{
   const char *filename1 = "bestsubcarriers.dat";
   int fd;
   FILE *d_fp;
   unsigned char sc[total_subcarriers];
   if ((fd = open (filename1,
                   O_RDONLY|OUR_O_LARGEFILE,
                   0664)) < 0){
       perror(filename1);
   }else {
       if((d_fp = fdopen (fd, true ? "r" : "rb")) == NULL) {
           fprintf(stderr, "subcarriers file cannot be opened\n");
           close(fd);
       }
       fprintf(stderr, "subcarriers file opened!\n");
       fread_unlocked(sc, sizeof(unsigned char), total_subcarriers, d_fp);
       close(fd);
   }

   int j = 0;
   int k = 0;
   for(int i = 0; i < total_subcarriers; i++)
   {
      if(sc[i] == '1')  // best subcarriers are tagged as '1'
      {
          best_sc_arr[j++] = i;
      }
      else
      {
          rem_sc_arr[k++] = i;
      }
   }

   assert(j == n);
   assert((j+k) == total_subcarriers);

#ifdef DEBUG
   printf("Total subcarriers: (%d)\n", total_subcarriers);
   printf("Best subcarriers: (%d)\n", j);
   for(int i = 0; i < j; i++)
   {
       printf("%d ", best_sc_arr[i]);
   }
   printf("\n");

   printf("Remaining subcarriers: (%d)\n", k);
   for(int i = 0; i < k; i++)
   {
       printf("%d ", rem_sc_arr[i]);
   }
   printf("\n");
#endif
}

/* calculates the delivery rates of all the FEC groups and puts them in 'fec_delivery_rates_arr'
 * order them in the increasing order of their delivery rates
 * selects the FEC groups whose delivery rates are below a certain threshold
 */
void
selectFECGroupsToProtect(std::vector<int> &sel_fec_groups, size_t total_bits,
                         float *fec_delivery_rates_arr, int fec_K, int fec_N,
                         int *interleaved_bit_pos, float *eff_delivery_rate)
{
   int num_fec_groups = ceil((float) total_bits/fec_N);
   int *sorted_fec_groups = (int*) malloc(sizeof(int) * num_fec_groups);
   memset(sorted_fec_groups, 0, sizeof(int) * num_fec_groups);

   printf("selectFECGroupsToProtect, fec_N: %d, fec_K: %d, num_fec_groups: %d, total_bits: %d\n", fec_N, fec_K, num_fec_groups, total_bits); fflush(stdout);

   int symSize = getSymSizeFromN_Bits(fec_N);
   int nsymbols = fec_N/symSize;
   assert(fec_N % symSize == 0);
   assert(fec_K % symSize == 0);

   printf("selectFECGroups, fec_N: %d, fec_K: %d, num_fec_groups: %d, nsymbols: %d\n", fec_N, fec_K, num_fec_groups, nsymbols); fflush(stdout);

   float *t_eff_delivery_rate = (float*) malloc(sizeof(float) * nsymbols);              // for input to getFECGroupDeliveryRate //

   int count = 0;
   for(int i = 0; i < num_fec_groups; i++)
   {
        int nbits = min(fec_N, total_bits - (i*fec_N));                                 // in RS, the last group might have fewer than fec_N bits //
        int ncodewords = nbits/symSize;
        assert(nbits % symSize == 0);

        // 't_eff_delivery_rate' for the entire FEC-group//
        for(int j = 0; j < ncodewords; j++)
        {
            t_eff_delivery_rate[j] = 1.0;
            for(int k = 0; k < symSize; k++)
            {
               int _orig_bit_pos = (i * fec_N) + (j * symSize) + k;
               int _interleaved_bit_pos = interleaved_bit_pos[_orig_bit_pos];
               t_eff_delivery_rate[j] *= eff_delivery_rate[_interleaved_bit_pos];
            }
        }

        float dr = getFECGroupDeliveryRate(t_eff_delivery_rate, ncodewords, fec_N/symSize, fec_K/symSize);
	if(dr == 1.0)
	    continue;

        fec_delivery_rates_arr[count] = dr;
        sorted_fec_groups[count] = i;                                                        // initialize the indices
        printf("fec_group: %d, del_rate: %f\n", i, dr);
	count++;
   }

   free(t_eff_delivery_rate);

   sortFECGroupsByDeliveryRate(fec_delivery_rates_arr, sorted_fec_groups, count);
   sel_fec_groups.assign(sorted_fec_groups, sorted_fec_groups + count);

#ifdef DEBUG_MAC_FEC
   printf("selected FEC groups --- \n");
   for(int i = 0; i < sel_fec_groups.size(); i++)
        printf("FEC_group: %d, delivery_rate: %f\n", sel_fec_groups.at(i), fec_delivery_rates_arr[i]);;
#endif

   free(sorted_fec_groups);
}

/* use Chebyshev's method to estimate the FEC group delivery rate
 *
 * X(i) = probability that bit 'i' is lost
 * Y = sum X(i)
 * alpha = (n - k - mean(Y))/sigma(Y)
 */
float getFECGroupDeliveryRate1(float *t_eff_delivery_rate, int nbits, int fec_N, int fec_K)
{
   float mean_loss_rate = 0.0;
   float var_loss_rate = 0.0;

   for(int i = 0; i < nbits; i++)
   {
        float Pi = 1 - t_eff_delivery_rate[i];
        mean_loss_rate += Pi;
        var_loss_rate += (Pi * (1 - Pi));
   }

   float sigma_loss_rate = sqrt(var_loss_rate);
    
   float ub_loss;
   float threshold = (float) (fec_N - fec_K)/2;
   if(threshold - mean_loss_rate > 0)
   {
	float k = (threshold - mean_loss_rate) / sigma_loss_rate;
	ub_loss = 1 / (1 + pow(k, 2));
   }
   else
 	ub_loss = 1;

   return 1 - ub_loss;
}

/* initializes the effective delivery rates of the bits as their symbol error rates,
 * takes into account the subcarriers on which they are mapped
 */
void initializeEffectiveDeliveryRate(float *eff_delivery_rate, float *csi, int total_bits,
				     int *interleaved_bit_pos, 
                                     std::map<float, float> &snr2ser_map, int bps, int total_subcarriers)
{
   printf("initializeEffectiveDeliveryRate, for bits: %d\n", total_bits); fflush(stdout);

   //setSNR_to_SER_Mapping(snr2ser_map);

   for(int i = 0; i < total_bits; i++)
   {
        int sym_index = i/bps;
        int sc_index = sym_index%total_subcarriers;
        eff_delivery_rate[i] = 1 - getSymbolErrorRate2(csi, sc_index, bps);
	assert(eff_delivery_rate[i] <= 1);
#ifdef DEBUG_MAC_FEC
        printf("eff_del_rate interleaved_pos: %d -- : %f, sym: %d, sc: %d\n", i, eff_delivery_rate[i], sym_index, sc_index); fflush(stdout);
#endif
   }

#ifdef DEBUG_MAC_FEC
   for(int i = 0; i < total_bits; i++)
   {
        int interleaved_pos = interleaved_bit_pos[i];
        printf("orig_pos: %d, interleaved_pos: %d, eff_delivery_rate: %f\n", i, interleaved_pos,
                                                        eff_delivery_rate[interleaved_pos]);
        fflush(stdout);
   }
#endif
}
/* sort the bits by their effective delivery rate (takes into account parity if any already provided)
 */
void sortBitsByEffectiveDeliveryRate(float *sorted_eff_delivery_rates_arr, int *sorted_bit_indices,
                                     float *eff_delivery_rate, int *interleaved_bit_pos,
                                     int fec_group_index, int fec_K, int fec_N)
{
   // initialize the arrays to be sorted //
   for(int i = 0; i < fec_K; i++)
   {
        int _orig_bit_pos = fec_group_index * fec_N + i;
        int _interleaved_bit_pos = interleaved_bit_pos[_orig_bit_pos];
        sorted_eff_delivery_rates_arr[i] = eff_delivery_rate[_interleaved_bit_pos];
        sorted_bit_indices[i] = _interleaved_bit_pos;
   }

   isort(sorted_eff_delivery_rates_arr, sorted_bit_indices, fec_K, true);

#ifdef DEBUG_MAC_FEC
   printf("sortBitsByEffectiveDeliveryRate::::::::: FEC_GROUP: %d\n", fec_group_index);
   for(int i = 0; i < fec_K; i++)
   {
        printf("bit: %d, eff_DR: %f\n", sorted_bit_indices[i], sorted_eff_delivery_rates_arr[i]);
   }
#endif
}

/* gets the Symbol Error Rate from the SNR->SER mapping
 * find the snr of the subcarrier, 'sc_index'
 * to get the ser for the given snr:
 *      - if given snr smaller than the smallest snr in the map, then return the corresponding ser
 *      - if direct match found, then return the corresponding ser
 *      - else, interpolate as
 *              - ser = (prev_ser + (snr - prev_snr)/(curr_snr - prev_snr) * (curr_ser - prev_ser))
 */
float getSymbolErrorRate(float *csi, int sc_index, std::map<float, float> snr2ser_map)
{
   float snr = csi[sc_index];

   float prev_snr = -1;
   float prev_ser = -1.0;

   float curr_snr = 0.0;
   float curr_ser = 0.0;

   std::map<float, float>::iterator it = snr2ser_map.begin();
   while(it != snr2ser_map.end())
   {
        curr_snr = (*it).first;
        curr_ser = (*it).second;

        if(snr <= curr_snr)
                break;
        if(snr == curr_snr)
                // case 2 //
                return curr_ser;

        prev_snr = curr_snr;
        prev_ser = curr_ser;
        it++;
   }

   if(prev_ser == -1.0)
        // case 1 //
        return curr_ser;
   else
        // case 3: interpolation //
        return (prev_ser + ((float) (snr - prev_snr))/(curr_snr - prev_snr) * (curr_ser - prev_ser));
}

void sortFECGroupsByDeliveryRate(float *delivery_rates, int *indices, int n)
{
   printf("sortFECGroupsByDeliveryRate n: %d\n", n); fflush(stdout);
   isort(delivery_rates, indices, n, true);
}

// simple sorting code //
int
fm(float arr[], int b, int n, bool ascending)
{
   int f = b;
   int c;

   for(c = b + 1; c < n; c++)
     if((ascending && arr[c] < arr[f]) || (!ascending && arr[c] > arr[f]))
       f = c;

   return f;
}

void
isort(float arr[], int indices[], int n, bool ascending)
{
   int s, w;
   float sm;
   int ind;
   for(s = 0; s < n - 1; s++) {
     w = fm(arr, s, n, ascending);
     sm = arr[w];
     ind = indices[w];
     arr[w] = arr[s];
     arr[s] = sm;
     indices[w] = indices[s];
     indices[s] = ind;
   }
}

// snr to ser mapping code //

/* reads from the SNR->SER file and sets up the 'snr2ser_map' */
void
setSNR_to_SER_Mapping(std::map<float, float> &snr2ser_map)
{
   const char *filename1 = "snr2ber.qpsk16"; //"snr2ser.qam64.s";

   int fd;
   FILE *d_fp;
   if ((fd = open (filename1,
                   O_RDONLY|OUR_O_LARGEFILE,
                   0664)) < 0){
        perror(filename1);
   }else {
        if((d_fp = fdopen (fd, true ? "r" : "rt")) == NULL) {
           fprintf(stderr, "snr-ser file cannot be opened\n");
           close(fd);
           assert(false);
        }
        fprintf(stderr, "snr-ser file opened!\n");
        float snr;
        float ser;
        char line[100];

        while(fgets(line, 100, d_fp) != NULL)
        {
                sscanf(line, "%f %f", &snr, &ser);
                snr2ser_map[snr] = ser;
        }
        fclose(d_fp);
   }
#ifdef DEBUG_MAC_FEC
   // print the entries //
   std::map<float, float>::iterator it = snr2ser_map.begin();
   while(it != snr2ser_map.end())
   {
        printf(" %f --> %f\n", (*it).first, (*it).second);
        it++;
   }
#endif
}


void readCSISubcarriers(float *csi, const int total_subcarriers)
{
   const char *filename1 = "ber_values"; //"csi.dat"; //"l1.dat_matrix_dB";
   printf("total_subcarriers: %d\n", total_subcarriers); fflush(stdout);
   int fd;
   FILE *d_fp;

   int count = 0;
   if ((fd = open (filename1,
                   O_RDONLY|OUR_O_LARGEFILE,
                   0664)) < 0){
        perror(filename1);
   }else {
        if((d_fp = fdopen (fd, true ? "r" : "rt")) == NULL) {
           fprintf(stderr, "CSI file cannot be opened\n");
           close(fd);
           assert(false);
        }
        fprintf(stderr, "CSI file opened!\n");
        char line[1700];

        float *tmp = (float*) malloc(sizeof(float) * total_subcarriers);
        memset(tmp, 0, sizeof(float) * total_subcarriers);

	int sc_index = 0;
        while(fgets(line, 1700, d_fp) != NULL)
        {
		/*
                sscanf(line, "%f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f", &tmp[0], &tmp[1], &tmp[2], &tmp[3], &tmp[4], &tmp[5], &tmp[6], &tmp[7], &tmp[8], &tmp[9], &tmp[10], &tmp[11], &tmp[12], &tmp[13], &tmp[14], &tmp[15], &tmp[16], &tmp[17], &tmp[18], &tmp[19], &tmp[20], &tmp[21], &tmp[22], &tmp[23], &tmp[24], &tmp[25], &tmp[26], &tmp[27], &tmp[28], &tmp[29], &tmp[30], &tmp[31], &tmp[32], &tmp[33], &tmp[34], &tmp[35], &tmp[36], &tmp[37], &tmp[38], &tmp[39], &tmp[40], &tmp[41], &tmp[42], &tmp[43], &tmp[44], &tmp[45], &tmp[46], &tmp[47], &tmp[48], &tmp[49], &tmp[50], &tmp[51], &tmp[52], &tmp[53], &tmp[54], &tmp[55], &tmp[56], &tmp[57], &tmp[58], &tmp[59], &tmp[60], &tmp[61], &tmp[62], &tmp[63], &tmp[64], &tmp[65], &tmp[66], &tmp[67], &tmp[68], &tmp[69], &tmp[70], &tmp[71]);
		
		
                for(int i = 0; i < total_subcarriers; i++)
                {
                     csi[i] += tmp[i];

                }
                count++;

                if(count == 1)
                        break;
		*/
		
		sscanf(line, "%f", &csi[sc_index]);
		sc_index++;

        }
        fclose(d_fp);
        free(tmp);
   }

   // find the average //
   for(int i = 0; i < total_subcarriers; i++)
   {
        //csi[i] = csi[i]/count - 51;                     //TODO: for now!
	printf("csi: %f, i: %d\n", csi[i], i); fflush(stdout);
   }
}

void sortSubcarriersByCSI(int *subcarrier_indices, float *csi, int total_subcarriers)
{
   // initialize the indices //
   for(int i = 0; i < total_subcarriers; i++)
   {
        subcarrier_indices[i] = i;
   }

   isort(csi, subcarrier_indices, total_subcarriers, true);

#ifdef DEBUG
   printf("sorted subcarrieres and CSI: \n");
   for(int i = 0; i < total_subcarriers; i++)
   {
        printf("sc: %d csi: %f\n", subcarrier_indices[i], csi[i]);
   }
   fflush(stdout);
#endif
}

/* gets the parity bits (if any) into 'paritySymbols' that exist for the given fec group */
void getParityForFECGroup(int fec_group_index, std::string parity_bits_str,
                            std::vector<int> orig_pos_indices, int fec_N, int fec_K,
                            std::vector<int>& fec_parity_indices, short *paritySymbols)
{
   std::string fec_parity_str;

   int start_pos_fec_group = fec_group_index * fec_N;
   int end_pos_fec_group = (fec_group_index + 1) * fec_K;

   for(int i = 0; i < orig_pos_indices.size(); i++)
   {
        if(orig_pos_indices[i] >= start_pos_fec_group && orig_pos_indices[i] <= end_pos_fec_group)
        {
                int pos = orig_pos_indices[i] - fec_group_index * fec_N;
                fec_parity_str.append(1, parity_bits_str[i]);
                fec_parity_indices.push_back(pos);
                assert(pos >= 0 && pos < fec_K);
        }
   }

   int k = 0;
   for(int i = 0; i < fec_parity_str.size(); i+=1)
   {
        std::bitset<sizeof(unsigned char) * 8> bits(fec_parity_str.substr(i, 1));
        paritySymbols[k++] = (short) bits.to_ulong();
   }
   assert(k == fec_parity_indices.size());

#ifdef DEBUG_MAC_FEC
   printf("parity bits for the fec group: %d\n", fec_group_index);
   for(int i = 0; i < k; i++)
   {
        printf("%d[%d] ", fec_parity_indices[i], paritySymbols[i]);
   }
   printf("\n");
#endif
}

int
getSymSizeFromN_Bits(int n)
{
   int symSize = 0;
   switch(n)
   {
	case 6:     symSize = 2; break;
	case 21:    symSize = 3; break;
	case 60:    symSize = 4; break;
	case 155:   symSize = 5; break;
	case 378:   symSize = 6; break;
	case 889:   symSize = 7; break;
	case 2040:  symSize = 8; break;
   }
   return symSize;
}

int
getSymSizeFromN_Syms(int n)
{
   int symSize = 0;
   switch(n)
   {
	case 3:     symSize = 2; break;
	case 7:    symSize = 3; break;
	case 15:    symSize = 4; break;
	case 31:   symSize = 5; break;
	case 63:   symSize = 6; break;
	case 127:   symSize = 7; break;
	case 255:  symSize = 8; break;
   }
   return symSize;
}

/* a more cleaner branch just for the split version - the encoder doesn't need something like this.
 * Just figures out the fec_split_map for (possible) mac-fec decoding later.. 
 */
#ifdef RS_MAC_OPT_VERSION
void
decodeMACLayer_FEC_OPT(std::string &deinterleaved_str,
		    std::string parity_bits_str,
                    int *interleaved_bit_pos,
		    int *orig_bit_pos,
                    int total_subcarriers, int fec_N,
                    int fec_K, int bps, int num_bits_before_deinterleaving,
		    std::map<int, split> &fec_split_map,
		    std::vector<int> &sorted_orig_pos)
{
   //printf("\n\n########################################### decodeMACLayer_FEC ##################################################################### \n"); 
   // get the interleaved string //
   const unsigned char *data = (const unsigned char*) deinterleaved_str.data();
   unsigned int len = deinterleaved_str.size();
   assert(len % 8 == 0);                        // byte boundary

   static bool initialized = false;
   static std::map<float, float> snr2ser_map;                    // (snr, ser) map

   if(!initialized)
   {
	float *csi = (float*) malloc(sizeof(float) * total_subcarriers);
	memset(csi, 0, sizeof(float) * total_subcarriers);
	int parity_added = 0;
        initialized = true;
        initializeMACLayerFEC_OPT(deinterleaved_str.size(),
			          num_bits_before_deinterleaving,
		      	          interleaved_bit_pos,
			          csi,
                                  fec_split_map,
				  sorted_orig_pos,
			          snr2ser_map,
                                  total_subcarriers, fec_N, fec_K, bps,  
				  parity_added,
			          false);
	
	initialized = true;
	// sorted_orig_pos currently contains interleaved pos, should contain orig pos //
	for(int i = 0; i < sorted_orig_pos.size(); i++)
		sorted_orig_pos[i] = orig_bit_pos[sorted_orig_pos[i]];
	free(csi);
   }
}

/* optimized version - returns the number of MAC FEC groups formed, one phy-FEC group *can* have 
 * atmost 2 MAC-FEC groups for it (one for good bits, other for bad) */
void
initializeMACLayerFEC_OPT(int num_bits_before_interleaving,
			  int num_bits_after_interleaving,
                          int *interleaved_bit_pos,
			  float *csi,
			  std::map<int, split>& fec_split_map,
			  std::vector<int>& sorted_orig_pos,
			  std::map<float, float> &snr2ser_map,
			  const int total_subcarriers,
  			  int fec_N, int fec_K, int bps, 
			  int &parity_added, bool encode)
{
   printf("initializeMACLayerFEC_OPT (RS OPT version) \n"); fflush(stdout);

   // original string before the interleaving //
   int num_fec_groups = ceil((float) num_bits_before_interleaving/fec_N);
   printf("num_bits_before_interleaving: %d, fec_N: %d, groups: %d\n", num_bits_before_interleaving, fec_N, num_fec_groups);

   // get the total number of data bits and redundancy bits //
   int n_redundancy_bits = num_fec_groups * (fec_N - fec_K);
   int n_data_bits = num_bits_before_interleaving - n_redundancy_bits;

   // read subcarrier csi values //   
   readCSISubcarriers(csi, total_subcarriers);

   // sort subcarriers by their csi values //
   int *sorted_sc_indices = (int*) malloc(sizeof(int) * total_subcarriers);
   memset(sorted_sc_indices, 0, sizeof(int) * total_subcarriers);
   float *sorted_csi = (float*) malloc(sizeof(float) * total_subcarriers);
   memcpy(sorted_csi, csi, sizeof(float) * total_subcarriers);
   sortSubcarriersByCSI(sorted_sc_indices, sorted_csi, total_subcarriers);

   // eff delivery rate for all interleaved positions (has to account for padded bits as well, since they
   // are part of the interleaved string)
   float *eff_delivery_rate = (float*) malloc(sizeof(float) * num_bits_after_interleaving);
   memset(eff_delivery_rate, 0, sizeof(float) * num_bits_after_interleaving);

   // initialize the effective delivery rates for all the interleaved positions possible //
   initializeEffectiveDeliveryRate(eff_delivery_rate, csi, 
				   num_bits_after_interleaving,
                                   interleaved_bit_pos,
                                   snr2ser_map, bps, total_subcarriers);

   // get the delivery rates for all the FEC groups //
   float *fec_delivery_rates_arr = (float*) malloc(sizeof(float) * num_fec_groups);
   memset(fec_delivery_rates_arr, 0, sizeof(float) * num_fec_groups);

   printf("num_fec_groups: %d\n", num_fec_groups); fflush(stdout);

   // get the FEC groups to protect, sorted by increasing order of delivery rates //
   std::vector<int> fec_groups_to_protect;

   selectFECGroupsToProtect(fec_groups_to_protect, num_bits_before_interleaving, fec_delivery_rates_arr, fec_K, fec_N, interleaved_bit_pos, eff_delivery_rate);
   int num_fec_groups_to_protect = fec_groups_to_protect.size();
   printf("num_fec_groups_to_protect: %d\n", num_fec_groups_to_protect); fflush(stdout);

   // calculate the current throughput ratio //
   float numerator = 0.0;
   for(int i = 0; i < fec_groups_to_protect.size(); i++)
   {
         numerator += fec_delivery_rates_arr[i];
   }
   float new_utility = (float) numerator/num_bits_after_interleaving;

   float old_utility = 0.0;
   int index = 0;
   parity_added = 0;
   int interleaved_parity_pos[bps * total_subcarriers];
   initInterleavedParityPos(interleaved_parity_pos, bps, total_subcarriers);  
   
   const int MAC_FEC_RS_BUDGET = 400;

   /* loop over all the FEC groups as long as the throughput condition is met
    * On each iteration, take the worst fec group; exhaustively search it to get the 
    * split, and finally perform FEC
    */
   std::map<int, split>::iterator it_m = fec_split_map.begin();
   for(int index = 0; index < fec_groups_to_protect.size(); index++)
   {
	old_utility = new_utility;
        int sel_fec_index = -1;
        int sel_index = -1;
        float sel_new_delivery_rate_fec = -1.0;

	int fec_group_index = fec_groups_to_protect[index];
	int num_data_bits_grp = min(fec_K, n_data_bits - fec_group_index * fec_K);
	
	float old_del_rate_fec = fec_delivery_rates_arr[index];
		
	// call lili's code here //
	int d_bad, d_good, r_bad, r_good;
        std::vector<int> t_pos; 
	float new_del_rate_fec = performSplitFEC(num_data_bits_grp, (fec_N - fec_K), MAC_FEC_RS_BUDGET, d_bad, r_bad, d_good, r_good,
		                                 eff_delivery_rate, interleaved_bit_pos, t_pos,
						 parity_added, total_subcarriers, bps,
						 sorted_csi, snr2ser_map, fec_group_index, fec_N, fec_K, interleaved_parity_pos);

	printf("old_del_rate_fec: %f, new_del_rate: %f, index: %d\n", old_del_rate_fec, new_del_rate_fec,
								      index);
	if(new_del_rate_fec < old_del_rate_fec)
	{
		if(index == fec_groups_to_protect.size() - 1)
		    break;
		else
		    continue;
	}

        // utility condition //
        float numerator = 0.0;
        for(int i = 0; i < fec_groups_to_protect.size(); i++)
	{
	    if(i == index)
		numerator += new_del_rate_fec;
	    else
	        numerator += fec_delivery_rates_arr[i];
	}

        int denominator = num_bits_after_interleaving + (parity_added + r_bad + r_good);
        new_utility = (float) numerator/denominator;
	printf("new_utility: %f, old_utility: %f\n", new_utility, old_utility); 
        if(new_utility < old_utility)
	{
	    printf("Utility is too low.. \n");
            continue;
	}

	// phy-FEC group protected now, update its delivery rate //
	fec_delivery_rates_arr[index] = new_del_rate_fec;

	// just store the details - fec_group_index, d_bad, r_bad, for each iteration // 
	//fec_split_map[fec_group_index] = make_split(r_bad, d_bad, r_good, d_good);
	fec_split_map.insert(it_m, std::pair<int, split>(fec_group_index, make_split(r_bad, d_bad, r_good, d_good, sorted_orig_pos.size(), parity_added)));	

	// update the parity added till now //
	parity_added += r_bad + r_good;
				
	// copy 't_pos' into 'sorted_orig_pos' //
	std::vector<int>::iterator it = sorted_orig_pos.begin();
	sorted_orig_pos.insert(it + sorted_orig_pos.size(), t_pos.begin(), t_pos.end());

	printf("fec_group_index: %d\n", fec_group_index);
	printf(" _____________________________ \n");
	
   }		

#ifdef DEBUG
   printf("The FEC groups protected are.. \n");
   std::map<int, split>::iterator it  = fec_split_map.begin();
   while(it != fec_split_map.end())
   {
       printf("%d ", (*it).first);	
       it++;
   }
   printf("\n");   
#endif


   free(fec_delivery_rates_arr);
   free(sorted_sc_indices);
   free(sorted_csi);
}

/* converted by Yi-Chao from 11n/scripts/fec.m 
 * d1: num of data bits in the phy-layer FEC group
 * r1: num of red bits in the phy-layer FEC group
 * r2: mac-layer FEC parity budget 
 */
float performSplitFEC(int d1, int r1, int r2, int &max_bad, int &max_r2_b, int &max_good, int &max_r2_g, 
		         float *eff_delivery_rate, int *interleaved_bit_pos, 
			 std::vector<int>& sorted_orig_indices,
	 	         int parity_mapped_already, int subcarriers, int bps,
	    	         float *sorted_csi, std::map<float, float> snr2ser_map,
	                 int fec_group_index, int fec_N, int fec_K, 
			 int *interleaved_parity_pos)
{
    // initialize these //
    max_bad = 0; 
    max_r2_b = 0; 
    max_good = 0;
    max_r2_g = 0;

    printf("\n\nfec_group: %d, d1: %d, r1: %d, r2: %d, parity_mapped: %d\n", fec_group_index, d1, r1, r2, parity_mapped_already); fflush(stdout);
    float max_delivery = 0.0;
    float max_utility = 0.0;
    // sort the 'data' bits in the FEC group by their eff delivery rates //
    float *sorted_eff_delivery_rate = (float*) malloc(sizeof(float) * d1);

    int *sorted_indices = (int*) malloc(sizeof(int) * d1);
    memset(sorted_indices, 0, sizeof(int) * d1);

    sortBitsByEffectiveDeliveryRate(sorted_eff_delivery_rate, sorted_indices,
		    eff_delivery_rate, interleaved_bit_pos,
		    fec_group_index, d1, fec_N);

    // fit in the 'sorted_orig_indices' //
    for(int i = 0; i < d1; i++)
	sorted_orig_indices.push_back(sorted_indices[i]);
  
    // get the delivery rates for the mac-parity-budget bits //
    float *del_rate_parity_r2 = (float*) malloc(sizeof(float) * r2);
    int mac_parity_mapped_one_round = subcarriers * bps; 
    for(int i = 0; i < r2; i++)
    {
	int parity_mapped_this_round = (parity_mapped_already + i) % mac_parity_mapped_one_round;
	
	int parity_pos = interleaved_parity_pos[parity_mapped_this_round];
	//printf("parity_pos: %d\n", parity_pos); fflush(stdout);	

        int sym_index = parity_pos / bps;
	int sc_index = sym_index % subcarriers;
	del_rate_parity_r2[i] = 1 - getSymbolErrorRate2(sorted_csi, sc_index, bps); 
    }	

    // delivery rate of the phy-layer parity bits //
    float *del_rate_parity_r1 = (float*) malloc(sizeof(float) * r1);
    for(int i = 0; i < r1; i++)
    {
        int _orig_bit_pos = fec_group_index * fec_N + d1 + i;
        int _interleaved_bit_pos = interleaved_bit_pos[_orig_bit_pos];
        del_rate_parity_r1[i] = eff_delivery_rate[_interleaved_bit_pos];
    }

    // P1: delivery probability of the phy-layer FEC group //
    int symSize = getSymSizeFromN_Bits(fec_N);
    assert((d1 + r1) % symSize == 0);

    assert(fec_N % symSize == 0);
    assert(fec_K % symSize == 0);
    int ncodewords = (d1 + r1)/symSize;

    float *t_eff_del_rate = (float*) malloc(sizeof(float) * ncodewords);
    for(int i = 0; i < ncodewords; i++)
    {
        t_eff_del_rate[i] = 1.0;
        for(int j = 0; j < symSize; j++)
        {
            int orig_pos = (fec_group_index * fec_N) + (i * symSize) + j;
            int int_pos = interleaved_bit_pos[orig_pos];
            t_eff_del_rate[i] *= eff_delivery_rate[int_pos];
        }
    }

    // delivery probability of the phy-layer FEC group //	
    float P1 = getFECGroupDeliveryRate(t_eff_del_rate, ncodewords, fec_N/symSize, fec_K/symSize);
    free(t_eff_del_rate);

    // enter all the possible values that 'N' can take!
    for(int sym_b = 2; sym_b <= 8; sym_b++)
    {
      int N_b = pow(2.0, sym_b) - 1;

      for(int bad_sym = 1; bad_sym <= N_b; bad_sym++) {			// symbols
	 
	  int bad = bad_sym * sym_b;					// bits
	
	  //printf("bad_sym: %d, sym_b: %d, bad: %d\n", bad_sym, sym_b, bad); fflush(stdout);
	  if(bad > d1) continue;	  
	 
	  int r2_b_sym = N_b - bad_sym;
	  int r2_b = r2_b_sym * sym_b;
	  if(r2_b > r2) continue;
	  //printf("      r2_b_sym: %d, r2_b: %d\n", r2_b_sym, r2_b); 
	  int good = d1 - bad;						// bits

	  for(int sym_g = 2; sym_g <= 8; sym_g++) 
	  {
		int N_g = pow(2.0, sym_g) - 1;
		if(good > d1) continue;
		int good_sym = ceil((float) good/sym_g);
	
		if(good_sym > N_g) continue;
		int r2_g_sym = N_g - good_sym;
		
		int r2_g = r2_g_sym * sym_g;

		//printf("                N_g: %d, good: %d, r2_g: %d\n", N_g, good, r2_g);	

		if(r2_g + r2_b > r2)
		{
		    r2_g = r2 - r2_b;
		    r2_g_sym = floor((float) r2_g/sym_g);
		    r2_g = r2_g_sym * sym_g;
		    good_sym = N_g - r2_g_sym;
		}

		// printf("             		   r2_g: %d, r2_b: %d\n", r2_g, r2_b);
		//printf("d1: %d, bad: %d, sym_b: %d r2_b_sym: %d, r2_b: %d ", d1, bad, sym_b, r2_b_sym, r2_b);
		//printf(" Ng: %d good: %d sym_g: %d good_sym: %d, r2_g_sym: %d, r2_g: %d\n", N_g, good, sym_g, good_sym, r2_g_sym, r2_g); fflush(stdout);	
		// delivery rates of the mac parity that will protect bad bits //
		float *del_rate_parity_bad = (float*) malloc(sizeof(float) * r2_b);	
		memcpy(del_rate_parity_bad, del_rate_parity_r2, sizeof(float) * r2_b);

		// delivery rates of the mac parity that will protect good bits //
                float *del_rate_parity_good = (float*) malloc(sizeof(float) * r2_g);
                memcpy(del_rate_parity_good, del_rate_parity_r2 + r2_b, sizeof(float) * r2_g);
	
		
		// delivery rates of the bad data bits //	
		float *del_rate_data1_bad = (float*) malloc(sizeof(float) * bad);
		memcpy(del_rate_data1_bad, sorted_eff_delivery_rate, sizeof(float) * bad);
	
			
		// delivery rates of the good data bits //
		float *del_rate_data1_good = (float *)malloc(sizeof(float) * good);
		memcpy(del_rate_data1_good, sorted_eff_delivery_rate + bad, sizeof(float) * good);

		// ---- P2_b: delivery probability of MAC-layer FEC group involving bad data symbols in PHY-layer ---- //
		ncodewords = bad_sym + r2_b_sym;

		float *del_rate_data1_bad_BER_fec2_b = (float *)malloc(sizeof(float) * ncodewords );
		for(int i = 0; i < ncodewords; i++)
		{
		    del_rate_data1_bad_BER_fec2_b[i] = 1.0;
		    for(int j = 0; j < sym_b; j++)
		    {
			int _index = (i * sym_b) + j;
			if(_index < bad)
			    del_rate_data1_bad_BER_fec2_b[i] *= sorted_eff_delivery_rate[_index];		// for data
			else
			    del_rate_data1_bad_BER_fec2_b[i] *= del_rate_parity_bad[_index - bad];		// for parity 
		    }
		}
		float P2_b = getFECGroupDeliveryRate(del_rate_data1_bad_BER_fec2_b, ncodewords, bad_sym + r2_b_sym, bad_sym);	
		// ------------------------------------------------------------------------------------ //


                // ---- P2_g: delivery probability of MAC-layer FEC group involving good data symbols in PHY-layer ---- //
                float *del_rate_data1_good_BER_fec2_g = (float *) malloc(sizeof(float) * (good_sym + r2_g_sym));
                ncodewords = good_sym + r2_g_sym;
                for(int i = 0; i < ncodewords; i++)
                {
                    del_rate_data1_good_BER_fec2_g[i] = 1.0;
                    for(int j = 0; j < sym_g; j++)
                    {
                        int _index = (i * sym_g) + j;
                        if(_index < (good_sym * sym_g))
                        {
                            if(_index < good)
                                del_rate_data1_good_BER_fec2_g[i] *= del_rate_data1_good[_index];
                        }
                        else
                            del_rate_data1_good_BER_fec2_g[i] *= del_rate_parity_good[_index - (good_sym * sym_g)];
                    }
                }

                float P2_g = getFECGroupDeliveryRate(del_rate_data1_good_BER_fec2_g, ncodewords, good_sym + r2_g_sym, good_sym);
                // ------------------------------------------------------------------------------------ //
		

                // -------  P12: delivery probability of PHY-layer FEC group after MAC-layer FEC ----------------------- //
                ncodewords = (d1 + r1)/symSize;
                float *del_rate_data12 = (float *)malloc(sizeof(float) * ncodewords);

                for(int i = 0; i < ncodewords; i++)
                {
                    del_rate_data12[i] = 1.0;
                    for(int j = 0; j < symSize; j++)
                    {
                        int _index = (i * symSize) + j;
                        if(_index < bad)
                            del_rate_data12[i] *= (1 - ((1 - del_rate_data1_bad[_index]) * (1 - P2_b)));
                        else if(_index < d1)
                            del_rate_data12[i] *= (1 - ((1 - del_rate_data1_good[_index - bad]) * (1 - P2_g)));
                        else
                            del_rate_data12[i] *= del_rate_parity_r1[_index - d1];
                    }
                }

                float P12 = getFECGroupDeliveryRate(del_rate_data12, ncodewords, fec_N/symSize, fec_K/symSize);
                // ------------------------------------------------------------------------------------------------------ //

                // expected successful data symbols
                float delivery = P1 * d1 + (1 - P1) * P12 * d1 + (1 - P1) * (1 - P12) * (P2_b * bad + P2_g * good );
		float utility = delivery/(d1 + r2_b + r2_g);

                //printf("bad: %d, r2_b: %d, good: %d, r2_g: %d delivery: %f, max_delivery: %f\n", bad, r2_b, good, r2_g, delivery, max_delivery);

		if(utility > max_utility) {
			max_utility = utility;
			max_delivery = delivery;
			max_bad = bad;
			max_r2_b = r2_b;
			max_r2_g = r2_g;
			//max_good = good + sym_g - (good % sym_g);		// possibly pad max_good
			max_good = good_sym * sym_g;
		}

		free(del_rate_data12);
		free(del_rate_data1_good_BER_fec2_g);
		free(del_rate_data1_bad_BER_fec2_b);
			
		free(del_rate_parity_bad);
		free(del_rate_parity_good);
		free(del_rate_data1_bad);
		free(del_rate_data1_good);
	}
      }
    }	

    if(max_bad == 0)
	max_r2_b = 0;
   
    if(max_good == 0)
	max_r2_g = 0;

    free(sorted_indices);
    free(sorted_eff_delivery_rate);
    free(del_rate_parity_r1);
    free(del_rate_parity_r2);
    printf("bad = %d, r2_b = %d, good = %d, r2_g = %d, delivery = %f\n", max_bad, max_r2_b, max_good, max_r2_g, max_delivery);

    // ensure that the data and red part for each group is a multiple of symSize //
    return max_delivery/d1;
}
#endif

std::string
getBits(const unsigned char *data, int len, const int symSize)
{
  std::string orig_s;
  for(int i = 0; i < len; i++)
  {
      std::bitset<sizeof(unsigned char) * 8> s_bits(data[i]);
      std::string tmp_string = s_bits.to_string<char,std::char_traits<char>,std::allocator<char> >();
      orig_s.append(tmp_string.substr(8 - symSize, symSize));
  }
  return orig_s;
}

/* for reed solomon, converts the s_bits into symbols of 'symSize' bits each. Returns
 * the num symbols created in 'symbols'
 */
int 
convertToSymbols(unsigned char *symbols, int symSize, std::string s_bits)
{
   int k = 0;
   for(int i = 0; i < s_bits.size(); i+=symSize)
   {
        std::bitset<sizeof(unsigned char) * 8> bits(s_bits.substr(i, symSize));
        symbols[k] = (unsigned char) bits.to_ulong();
        k++;
   }

   return k;
}

void 
injectErrors(unsigned char *msg, int msg_length)
{
   printf("injecting errors at .. \n");
   int num_errors = 17;
   for(int i = 0; i < 5; i++)
   {
        int pos = i * 32;
        msg[pos] = msg[pos] ^ 1;
	printf("%d ", pos); 
   }

   for(int i = 0; i < 15; i++)
   {
	int pos = i * 9 + 225;
	msg[pos] = msg[pos] ^ 1;
	printf("%d ", pos);
   }

   printf("\n"); 
}

/* tries to repair the current phy-FEC block, using mac-parity if it exists */
bool
repairUsingMacParity(unsigned char *data, int n, int k, int fec_group_index, int num_decoded,
		     std::string parity_bits_str, std::vector<int> orig_pos_indices, 
		     int mac_fec_N, int mac_fec_K, int symSize, std::vector<struct rs*> rsVector)
{
   printf("trying to repair using MAC parity!!! \n\n"); fflush(stdout);
   int start_pos_fec_group = fec_group_index * n * symSize;
   int end_pos_fec_group = (fec_group_index + 1) * num_decoded * symSize;

   std::string mac_fec_parity_str;
   std::string mac_fec_data_str; 
   std::vector<int> fec_parity_pos;

   int mac_symSize = 0;
   switch(mac_fec_N) {
        case 3:  mac_symSize = 2; break;
        case 15: mac_symSize = 4; break;
        case 255: mac_symSize = 8; break;
        default:
                assert(false);
   }

   // data bits of the phy-layer fec group //
   std::string bits = getBits(data, num_decoded, symSize);

   bool found = false;
   for(int i = 0; i < orig_pos_indices.size(); i++)
   {
	if(orig_pos_indices[i] >= start_pos_fec_group && orig_pos_indices[i] <= end_pos_fec_group)
	{
	    // starting point found for parity, mac_fec_K symbols protected  //
	    for(int j = 0; j < mac_fec_K * mac_symSize; j++)
	    {
		int pos = orig_pos_indices[i+j] - fec_group_index * n;
		mac_fec_data_str.append(1, bits[pos]);
		fec_parity_pos.push_back(pos);
	    }
		
	    // now capture the parity provided //
	    for(int j = 0; j < (mac_fec_N - mac_fec_K) * mac_symSize; j++)
	    {
		mac_fec_parity_str.append(1, parity_bits_str[i+j]);
            } 

	    found = true;
	    break;	
	}
   }

   if(!found)
	return false;

   printf("MAC parity provided --- \n");
   printf("indices protected .. \n");
   for(int i = 0; i < orig_pos_indices.size(); i++)
        printf("%d ", orig_pos_indices[i]);
   printf("\n");

   printf("MAC data bits protected -- \n");
   for(int i = 0; i < mac_fec_data_str.size(); i++)
        printf("%c ", mac_fec_data_str[i]);
   printf("\n");

   printf("parity_bits --- \n");
   for(int i = 0; i < parity_bits_str.size(); i++)
        printf("%c ", parity_bits_str[i]);



   // get all the bits of the mac layer fec group //
   mac_fec_data_str.append(mac_fec_parity_str);
  
   assert(mac_fec_data_str.size() == (mac_fec_N) * mac_symSize);
 
   // convert into the symbols //  
   unsigned char mac_fec_symbols[mac_fec_N]; 
   convertToSymbols(mac_fec_symbols, mac_symSize, mac_fec_data_str); 

   struct rs *rs = find_rs_handle(rsVector, mac_fec_N, mac_fec_K);
   assert(rs);

   int retVal = gr_decode_rs(rs, mac_fec_symbols);

	// whoa.. mac FEC passed //
   if(retVal != -1)
   {
	printf("mac fec passed!!\n"); fflush(stdout);
	// convert the data symbols of the mac-fec-block into bits //
	std::string mac_data_bits = getBits(mac_fec_symbols, mac_fec_K, mac_symSize);
	
	assert(mac_data_bits.size() == fec_parity_pos.size());
	assert(mac_data_bits.size() == mac_fec_K * mac_symSize);

	printf("data bits of the MAC fec group!.. \n");
	for(int i = 0; i < mac_data_bits.size(); i++)
		printf("%c ", mac_data_bits[i]);
	printf("\n");

	// replace the 'bits' in the phy fec group now //
	for(int i = 0; i < fec_parity_pos.size(); i++)
	{
	    bits.replace(fec_parity_pos[i], 1, 1, mac_data_bits[i]); 
	}
   }
   else
   {
	printf("MAC parity fails.. \n"); 
	return false;
   }
   convertToSymbols(data, symSize, bits);
   return true;
}

/* split version */
bool
repairUsingMacParity_split_OPT(unsigned char *data, int n, int k, int fec_group_index, int num_decoded, 
                               std::string parity_bits_str, std::map<int, split> fec_split_map,
			       std::vector<int> sorted_orig_pos,
                               int symSize,
			       std::vector<struct rs*> rsVector)
{
   // data bits in the phy-layer FEC group //
   int n_data_bits = num_decoded * symSize;
   printf("trying to repair FEC-group: %d using MAC parity.. n: %d, k: %d n_data_bits: %d!!! \n\n", fec_group_index, n, k, n_data_bits); 
   fflush(stdout);

   // search in the fec_split_map, if this fec group has an entry //
   std::map<int, split>::iterator it = fec_split_map.find(fec_group_index); 
   if(it == fec_split_map.end())
	return false;

   split split_pair = (*it).second;   
   int r_bad = split_pair.r_bad;
   int d_bad = split_pair.d_bad;
   int r_good = split_pair.r_good;
   int d_good = split_pair.d_good;		// it is possibly padded!
   int d_good_pad = d_good - (n_data_bits - d_bad);
   int d_good_ideal = d_good - d_good_pad;	// actual number of bits!

   printf("bad: %d, r_bad: %d, good: %d, r_good: %d good_ideal: %d \n", d_bad, r_bad, d_good, r_good, d_good_ideal);
   fflush(stdout);

   int offset = split_pair.indices_offset; 

   // extract the total (bad+good) mac parity bits for this fec_group //
   std::string mac_parity_bits = parity_bits_str.substr(split_pair.parity_offset, (r_bad + r_good));
   printf("mac_parity_bits size: %d, actual: %d\n", r_bad + r_good, mac_parity_bits.size());

   // offset found, now extract the actual bad and good data bits //
   std::string bits = getBits(data, num_decoded, symSize);
   bool mac_correction = false;
   if(r_bad > 0)
   {
	printf("mac-fec: bad part, r_bad: %d offset: %d n: %d\n", r_bad, offset, n); fflush(stdout);
	std::string bad_bits;
	for(int i = 0; i < d_bad; i++)
	{
	    int pos = sorted_orig_pos[offset + i] - fec_group_index * n * symSize;
	    bad_bits.append(1, bits[pos]);
	}

	std::string bad_parity = mac_parity_bits.substr(0, r_bad);
	bad_bits.append(bad_parity);
	mac_correction = do_MAC_FEC_decoding(bits, bad_bits, fec_group_index, rsVector,
	 		 d_bad + r_bad, d_bad, offset, sorted_orig_pos, 0, n * symSize);
   }

   if(r_good > 0)
   {
	printf("mac-fec: good part, r_good: %d \n", r_good); fflush(stdout);
	std::string good_bits;
	for(int i = 0; i < d_good_ideal; i++)
	{		
	    int pos = sorted_orig_pos[offset + d_bad + i] - fec_group_index * n * symSize;
	    good_bits.append(1, bits[pos]);
	}
	
	// if d_good_pad > 0, then append '0' to the good_bits
	if(d_good_pad > 0)
		good_bits.append(d_good_pad, '0');	
   
#ifdef DEBUG_MAC_FEC
	printf("good data part : \n");
	for(int i = 0; i < good_bits.size(); i++)
		printf("%c", good_bits[i]);
	printf("\n");
#endif
	   
        assert(good_bits.size() == d_good);
        printf(" d_good: %d\n", d_good); fflush(stdout);
 
	std::string good_parity = mac_parity_bits.substr(r_bad, r_good);
	good_bits.append(good_parity);

#ifdef DEBUG_MAC_FEC
        printf("good parity part : \n");
        for(int i = 0; i < good_parity.size(); i++)
                printf("%c", good_parity[i]);
        printf("\n");
#endif

	int _symSize = 2;
	switch(d_good + r_good)
        {
             case 6: _symSize = 2; break;
	     case 21:_symSize = 3; break;
	     case 60: _symSize = 4; break;
	     case 155: _symSize = 5; break;
	     case 378: _symSize = 6; break;
	     case 889: _symSize = 7; break;
	     case 2040: _symSize = 8; break;
	     default: assert(false);

	}

        unsigned char symbols[n];
        printf("good_bits.size: %d, symSize: %d\n", good_bits.size(), _symSize);
	fflush(stdout);

	mac_correction = do_MAC_FEC_decoding(bits, good_bits, fec_group_index, rsVector,
					     d_good + r_good, d_good, 
		       			     offset + d_bad, sorted_orig_pos, d_good_pad, n * symSize);
   }

   printf("r_good: %d, d_good: %d, r_bad: %d, d_bad: %d, d_good_pad: %d\n", r_good, d_good, r_bad, d_bad, d_good_pad);
   fflush(stdout);

   // convert back the (corrected) phy-fec bits into the FEC-group symbols //
   if(mac_correction)
       convertToSymbols(data, symSize, bits);
   return true; 
}

bool
do_MAC_FEC_decoding(std::string &phy_fec_bits, std::string mac_fec_bits, int fec_group_index,
		    std::vector<struct rs*> rsVector, int n, int k, int offset, 
		    std::vector<int> sorted_orig_pos, int pad_remove, int phy_fec_N)
{
   printf("do_MAC_FEC_decoding, n: %d, k: %d\n", n, k); fflush(stdout);
   int symSize = 2;
   switch(n)
   {
        case 6: symSize = 2; break;
        case 21: symSize = 3; break;
        case 60: symSize = 4; break;
        case 155: symSize = 5; break;
        case 378: symSize = 6; break;
        case 889: symSize = 7; break;
        case 2040: symSize = 8; break;
        default: assert(false);
   }

   int N = n/symSize;
   int K = k/symSize;
   assert(n % symSize == 0);
   assert(k % symSize == 0);
 
   struct rs *rs = find_rs_handle(rsVector, N, K);     
   assert(rs);

   unsigned char symbols[N];
   convertToSymbols(symbols, symSize, mac_fec_bits);

   int retVal = gr_decode_rs(rs, symbols);
   if(retVal >= 0)
   {
      printf("REPAIR partly SUCCESS\n");
      std::string corrected_data_bits = getBits(symbols, K, symSize);
      for(int j = 0; j < corrected_data_bits.size() - pad_remove; j++)
      {
          int pos = sorted_orig_pos[offset + j] - fec_group_index * phy_fec_N;
          phy_fec_bits.replace( pos, 1, 1, corrected_data_bits[j]);
      }
      return true;
   }
   return false;
}


struct rs*
find_rs_handle(std::vector<struct rs*> rs_vec, int n, int k)
{
    printf("finding n: %d, k: %d\n", n, k); fflush(stdout);
    int syms;
    int gfpoly;
    int nn, nroots;

    switch(n){
        case 3:
                syms = 2; gfpoly = 0x7; break;
	case 7:
                syms = 3; gfpoly = 0xb; break;
        case 15:
                syms = 4; gfpoly = 0x13; break;
	case 31:
                syms = 5; gfpoly = 0x25; break;
	case 63:
                syms = 6; gfpoly = 0x43; break;
	case 127:
                syms = 7; gfpoly = 0x89; break;
        case 255:
                syms = 8; gfpoly = 0x11d; break;
        default:
                printf("the length of the block is out of range, FEC cannot be initiated.");
                break;
    }

    std::vector<struct rs*>::iterator it = rs_vec.begin();
    struct rs *rs;

    while(it != rs_vec.end())
    {
        printf("searching!\n"); fflush(stdout);
        rs = *it;
        if(rs->nn == n && (rs->nroots - rs->nn) == k)
        {
		return rs;
        }
        it++;
    }

    nroots = n - k;
    rs = gr_init_rs(syms, gfpoly, 1, 1, nroots);
    rs_vec.push_back(rs);
    return rs;
}

split
make_split(int r_bad, int d_bad, int r_good, int d_good, int d_indices_offset, int d_parity_offset)
{
    struct split s;
    s.r_bad = r_bad;
    s.d_bad = d_bad;
    s.r_good = r_good;
    s.d_good = d_good;
    s.indices_offset = d_indices_offset;
    s.parity_offset = d_parity_offset;
    return s;
}

/* new symbol error rate calculation (from Yi-Chao) */
float getSymbolErrorRate2(float *csi, int sc_index, int mod)
{
   return csi[sc_index];

   //return snr2ber(mod, csi[sc_index]);
}

double snr2ber(int bps, double snr)
{
   double ber, ser;
   int m;

   snr = pow(10, snr / 10);    // dB to power
   snr *= (float)80/52;

   if(bps == 1) {
        m = 2;
        ber = erfc(pow(snr, 0.5) ) / 2;
   }
   else if(bps == 2) {
	m = 4;
	ber = erfc(pow(snr / 2, 0.5) ) / 2;
   }
   else if(bps == 3) {
        m = 8;
        ber = erfc(pow((9 * snr) / 7, 0.5) ) * 4;
   }
   else if(bps == 4) {
	m = 16;
	ber = erfc(pow(snr / 10, 0.5) ) * 3 / 8;
   }
   else if(bps == 6) {
	m = 64;
	ber = erfc(pow(snr / 42, 0.5) ) * 7 / 24;
   }

   return ber;
}

double erf(double x)
{
    static const double rel_error= 1E-12;        //calculate 12 significant figures
    static const double two_sqrtpi=  1.128379167095512574;        // 2/sqrt(pi)
    if (fabs(x) > 2.2) {
        return 1.0 - erfc(x);        //use continued fraction when fabs(x) > 2.2
    }
    double sum= x, term= x, xsqr= x*x;
    int j= 1;
    do {
        term*= xsqr/j;
        sum-= term/(2*j+1);
        ++j;
        term*= xsqr/j;
        sum+= term/(2*j+1);
        ++j;
    } while (fabs(term)/sum > rel_error);
    return two_sqrtpi*sum;
}

double erfc(double x)
{
    static const double rel_error= 1E-12;        //calculate 12 significant figures
    static const double one_sqrtpi=  0.564189583547756287;        // 1/sqrt(pi)
    if (fabs(x) < 2.2) {
        return 1.0 - erf(x);        //use series when fabs(x) < 2.2
    }
    if (x < 0) {               //continued fraction only valid for x>0
        return 2.0 - erfc(-x);
    }
    double a=1, b=x;                //last two convergent numerators
    double c=x, d=x*x+0.5;          //last two convergent denominators
    double q1, q2= b/d;             //last two convergents (a/c and b/d)
    double n= 1.0, t;
    do {
        t= a*n+b*x;
        a= b;
        b= t;
        t= c*n+d*x;
        c= d;
        d= t;
        n+= 0.5;
        q1= q2;
        q2= b/d;
    } while (fabs(q1-q2)/q2 > rel_error);
    return one_sqrtpi*exp(-x*x)*q2;
}

/* complete wrapper that encapsulates the 'deinterleaving' and the 'fec decoding' options. Turn
   the options on and off to achieve the desired effect.
*/
std::string
gr_rx_wrapper(std::string buf, int expectedDataLen, int subcarriers, int bits_per_symbol,
              int deinterleave, int decode_fec, int fec_N, int fec_K, 
	      int mac_parity_added,
	      int& num_fec_grps, int& correct_fec_grps, int& correct_bytes, int& hdr_correct,
	      int mac_fec, int mapping_scheme)
{
   subcarriers -= 8; 
   std::string parity_bits_str;
   //printf("gr_rx_wrapper, buf.size: %d, expectedDataLen: %d, decode_fec: %d mac_parity_added: %d scheme: %d\n", 
   //			buf.size(), expectedDataLen, decode_fec, mac_parity_added, mapping_scheme);  fflush(stdout); 

	/*
   static int count_rx = -1;
   printf("count_rx: %d\n", count_rx);
   count_rx++;
   if(count_rx % 5 != 0)
	return buf.substr(0, expectedDataLen);			// bypass processing of all such packets
	*/
   bool mac_fec_flag = false;
   if(mac_fec == 1)
	mac_fec_flag = true;


   static std::vector<int> sorted_orig_indices;
   static std::map<int, split> fec_split_map;
   static std::vector<int> orig_pos_indices;

#ifdef DEBUG
   printf("before deinterleaving .. \n");
   for(int i = 0; i < buf.size(); i++)
        printf("%d ", buf[i]);
   printf("\n");
#endif

   if(1)
   {
	// log all the bytes (including the phy-crc) being received to be processed offline //
	static bool isLogFileOpen = false;
	static FILE *d_fp1;
	if(!isLogFileOpen)
 	{
	    const char *d_filename = "raw_bits.dat";
	    int fd;
            if ((fd = open (d_filename,
                        O_WRONLY|O_CREAT|O_TRUNC|OUR_O_LARGEFILE|OUR_O_BINARY,
                        0777)) < 0){
                perror(d_filename);
	    }else {
                if((d_fp1 = fdopen (fd, true ? "wb" : "w")) == NULL) {
                        fprintf(stderr, "raw-bytes file cannot be opened\n");
                        close(fd);
                }
                fprintf(stderr, "raw-bytes opened!\n");
            }
            assert(d_fp1);
            isLogFileOpen = true;
	}

	assert(d_fp1);
	int count = ftell(d_fp1);
	std::string d_bits = getBitString((unsigned char*) buf.data(), buf.size());
	count = fwrite((const unsigned char*) d_bits.data(), strlen(d_bits.data()), 1, d_fp1);
	printf("logged %d to the bytes file, size: %d\n", count, ftell(d_fp1));
   }


   std::string deinterleaved_str = buf;	// initialize it
   int hdr_data_size = 24; 
   switch(deinterleave)
   {
       case 0: break;

       case 1: deinterleaved_str = deinterleave_mapping_best(buf, subcarriers, bits_per_symbol, fec_N, fec_K, mapping_scheme, expectedDataLen, 
							     mac_fec_flag, parity_bits_str, mac_parity_added, fec_split_map, sorted_orig_indices);
	       break;

       default: assert(false);
   }

#ifdef DEBUG
   printf("after deinterleaving .. size: %d\n", deinterleaved_str.size());
   for(int i = 0; i < deinterleaved_str.size(); i++)
	printf("%d ", deinterleaved_str[i]);
   printf("\n");  
#endif
 
   std::string decoded_str = deinterleaved_str;
   if(fec_N == 0)
	decode_fec = false;
 
   
   // file parameters //
   int count;
   static bool isFileInitialized = false;
   static FILE *d_fp; 
   int n_fec_grps_this_round = 0;
   int correct_fec_grps_this_round = 0;
   if(!isFileInitialized)
   {
	const char *d_filename = "rx_bits.dat";
	int fd;
	if ((fd = open (d_filename,
			O_WRONLY|O_CREAT|O_TRUNC|OUR_O_LARGEFILE|OUR_O_BINARY,
			0777)) < 0){
		perror(d_filename);
	}else {
		if((d_fp = fdopen (fd, true ? "wb" : "w")) == NULL) {
			fprintf(stderr, "rx-bits file cannot be opened\n");
			close(fd);
		}
		fprintf(stderr, "rx bits opened!\n");
	}
	assert(d_fp);
	isFileInitialized = true;
   }
   switch(decode_fec)
   {
	case 1: if(deinterleave == 1)
		{
#ifdef DEBUG
		    printf("before decoding.. \n");
		    for(int i = 0; i < deinterleaved_str.size(); i++)
			printf("%d ", deinterleaved_str.data()[i]);
		    printf("\n");
#endif
		    decoded_str = gr_decode_rs_fec(deinterleaved_str, fec_N, fec_K, expectedDataLen, 
					           parity_bits_str, mac_fec_flag
#ifdef RS_MAC_OPT_VERSION
				                  , fec_split_map, sorted_orig_indices
#else
				                  , orig_pos_indices
#endif

						  , n_fec_grps_this_round, correct_fec_grps_this_round, correct_bytes);

		    
		    /*
		    // log the data if the hdr is received clearly //
		    // exclude the hdr and the crc //
		    if(correct_fec_grps_this_round > 0)
		    {
			    std::string payload_only = decoded_str.substr(hdr_data_size, decoded_str.size() - hdr_data_size - 4);
			    std::string d_bits = getBitString((unsigned char*) payload_only.data(), payload_only.size());
			    assert(d_fp);
			    count = ftell(d_fp);
			    count = fwrite((const unsigned char*) d_bits.data(), strlen(d_bits.data()), 1, d_fp);
			    printf("bits logged: %d\n", count); 
			    count = ftell(d_fp);
                    }*/
		    num_fec_grps = n_fec_grps_this_round;
		    correct_fec_grps = correct_fec_grps_this_round;
		}
		else
		{
 		    decoded_str = gr_decode_rs_fec(deinterleaved_str, fec_N, fec_K, expectedDataLen, parity_bits_str, mac_fec_flag
#ifdef RS_MAC_OPT_VERSION
                                                  ,fec_split_map, sorted_orig_indices
#else
                                                  ,orig_pos_indices
#endif
			      			  , num_fec_grps, correct_fec_grps, correct_bytes);
		}
 	        break;
	case 2:	assert(false);			       
		break;
   }


   // do the CRC check on the header and log the data if the header is received clearly //
   int hdr_size_wo_crc =  hdr_data_size - 4;
   std::string hdr_w_crc_str = decoded_str.substr(0, hdr_data_size);
   std::string hdr_wo_crc_str = decoded_str.substr(0, hdr_size_wo_crc);
   
   std::string exp_crc = hdr_w_crc_str.substr(hdr_size_wo_crc, 4);
   unsigned int rx_hdr_crc = gr_crc32(hdr_wo_crc_str);   
   char tmp[5];
   snprintf(tmp, sizeof(tmp), "%ld", rx_hdr_crc);
 
   std::string actual_crc(tmp);
   assert(actual_crc.size() == 4);

   std::string payload_only = decoded_str.substr(hdr_data_size, decoded_str.size() - hdr_data_size);
   if(actual_crc.compare(exp_crc) == 0)
   {
	  /*	for partial throughput computation (rx_bits.dat used by partial-throughput.pl) 
	   std::string d_bits = getBitString((unsigned char*) payload_only.data(), payload_only.size() - 4);
	   assert(d_fp);
	   count = ftell(d_fp);
	   count = fwrite((const unsigned char*) d_bits.data(), strlen(d_bits.data()), 1, d_fp);
	   printf("bits logged: %d\n", count);
	   count = ftell(d_fp);
	   printf("HEADER rx CORRECT!\n"); fflush(stdout); */
	
	   hdr_correct = 1;	
   } 
   else
   {
	  // if header is wrong, then discard all fec groups //
	  correct_fec_grps = 0;
	  correct_bytes = 0;
	  hdr_correct = 0;
   }

#ifdef DEBUG
   printf("hdr_wo_crc_str .. \n");
   for(int i = 0; i < hdr_wo_crc_str.size(); i++)
	printf("%d ", hdr_wo_crc_str.data()[i]);  
 
   printf("\npayload ... \n");
   for(int i = 0; i < payload_only.size(); i++)
	printf("%d ", payload_only.data()[i]); 
   printf("\n");
#endif
   decoded_str = hdr_wo_crc_str + payload_only;

#ifdef DEBUG
   printf("after decoding.. \n");
   for(int i = 0; i < decoded_str.size(); i++)
	   printf("%d ", decoded_str.data()[i]);
   printf("\n");
#endif

   return decoded_str;
}


/* --------------------------- for the alternative deinterleaver, which uses the best mapping strategy  ------------------------------- */

std::string
reconstruct_payload_fec(std::string payload_only, std::string fec_only, int fec_N, int fec_K)
{
    if(fec_only.size() == 0)
	return "";

    int len_payload_fec = payload_only.size() + fec_only.size();
    int num_fec_grps = ceil((float) len_payload_fec/fec_N);

    std::string comb_str; 
    for(int i = 0; i < num_fec_grps; i++)
    {
	std::string payload = payload_only.substr(i * fec_K, fec_K);
	comb_str.append(payload);
	std::string fec = fec_only.substr(i*(fec_N - fec_K), fec_N - fec_K);
	comb_str.append(fec);
    }
    return comb_str; 
}

int
deinterleave_bucket(int len, int num_sp_bits, int n_subcarriers, int& bits_demapped_already, int subcarriers, 
		  int& offset, int bps, int rounds, 
		  int* sorted_sc_indices, int* orig_bit_pos,
		  int* final_pos, unsigned char* tracker, int scheme)
{
    int s_len = len * 8;

    //printf("deinterleave_bucket: len:: %d, s_len: %d, bps: %d, num_sp_bits: %d, bits_demapped_already: %d, rounds: %d, offset: %d, n_subcarriers: %d\n", len, s_len, bps, num_sp_bits, bits_demapped_already, rounds, offset, n_subcarriers);

    int Ncbps = n_subcarriers * bps;
    int Nbpsc = bps;
    assert(num_sp_bits % Ncbps == 0);

    int s = (Nbpsc/2 > 1) ? Nbpsc/2 : 1;
    int blocks = ceil((float)num_sp_bits/Ncbps);

    int C = 1;
    for(int i = 6; i >=2; i--)
    {
	if(Ncbps % i == 0)
	{
	   C = Ncbps/i;
	   break;
	}
    }
    //printf("C: %d, blocks: %d\n", C, blocks); 
   
    // random interleaver uses C = 16 //
    if(scheme == 0)
        C = 16;
 
    assert(Ncbps % C == 0);

    int *arr_k = (int*) malloc(sizeof(int) * Ncbps * blocks); 
    memset(arr_k, 0, sizeof(int) * Ncbps * blocks);

    for(int b = 0; b < blocks; b++)
    {
	for(int j = 0; j < Ncbps; j++)
	{
            if(scheme != 0)
                arr_k[b*Ncbps + j] = C * j - (Ncbps - 1) * floor((float) C * j/Ncbps);
            else
            {
                int i = s * floor((float) j/s) + (int)(j + floor((float) C * j/Ncbps)) % s;
                arr_k[b*Ncbps + j] = C * i - (Ncbps - 1) * floor((float) C * i/Ncbps);
            }
	}
    }


    for(int b = 0; b < blocks; b++)
    {
       for(int k = 0; k < Ncbps; k++)
       {
           //int i = (Ncbps/C)*(k%C) + floor((float)k/C);
           //int j = s * floor((float)i/s) + (i + Ncbps - (int) floor((float)C * i/Ncbps))%s;
	   //orig_bit_pos[j+offset] = k+offset; 
	   int interleaved_pos = k + offset;
	   int orig_pos = arr_k[k] + offset;
	
	   orig_bit_pos[interleaved_pos] = orig_pos;
	   
	
	   // will be used to memcpy the interleaved bits into their final mappings //
	   int p = k + b * Ncbps;
	   int sym_index = p/bps;
	   int rem = p%bps;
	   int sc_index = sym_index % n_subcarriers;
	   int actual_sc_index = sorted_sc_indices[sc_index];
	   int pos = (p/Ncbps)*(subcarriers * bps) + actual_sc_index * bps + rem;

	   //printf("p: %d sc_index: %d, actual_sc: %d, sym_index: %d, pos: %d\n", p+ bits_demapped_already, sc_index, actual_sc_index, sym_index, pos);

	   final_pos[pos] = p + bits_demapped_already;
	   //printf("pos: %d\n", pos); 
	   assert(tracker[pos] == 0);
	   tracker[pos] = 1;
	   assert(pos < s_len);
       }
       offset += Ncbps;
    }

    free(arr_k);
    bits_demapped_already += num_sp_bits;
    return num_sp_bits;
}

std::string
deinterleave_mapping_best(std::string buf, int subcarriers, int bps, int fec_N, int fec_K, int scheme, int expectedPayloadWithHdrLen, bool mac_fec_flag,
			  std::string& parity_bits_str, int mac_parity_added, 
			  std::map<int, split>& fec_split_map, std::vector<int>& sorted_orig_pos)
{
    int NBYTES[4];	// 4 priority buckets
    
    int hdr_data_size = 24;
    int hdr_fec_size = fec_N - fec_K;
 
#ifdef DEBUG
    printf("interleave_mapping: size: %d, subcarriers: %d, bps: %d, fec_N: %d, fec_K: %d, scheme: %d\n", 
				buf.size(), subcarriers, bps, fec_N, fec_K, scheme);
#endif

    // form the buckets //
    int num_buckets = 4;
    int num_fec_grp = (fec_K == 0) ? 0:ceil((float)expectedPayloadWithHdrLen/fec_K);
    int payload_fec_size = (num_fec_grp-1) * (fec_N - fec_K);				// first fec goes with hdr
    switch(scheme)
    {
        case 0:   NBYTES[0] = expectedPayloadWithHdrLen + (fec_N - fec_K) +  payload_fec_size;
                  num_buckets = 1;
                  break;

	case 1:	  NBYTES[0] = hdr_data_size;
		  NBYTES[1] = expectedPayloadWithHdrLen - hdr_data_size;
		  NBYTES[2] = (fec_N - fec_K);
		  NBYTES[3] = payload_fec_size;
		  num_buckets = 4;
                  if(payload_fec_size == 0)
                        num_buckets--;
	  
		  break;

        case 2:	  NBYTES[0] = expectedPayloadWithHdrLen - hdr_data_size;
                  NBYTES[1] = hdr_data_size;
                  NBYTES[2] = (fec_N - fec_K);
                  NBYTES[3] = payload_fec_size;
                  num_buckets = 4;

                  if(payload_fec_size == 0)
                        num_buckets--;
		  break;

        case 3:   NBYTES[0] = expectedPayloadWithHdrLen;
                  NBYTES[1] = (fec_N - fec_K); 
                  NBYTES[2] = payload_fec_size;
                  num_buckets = 3;

                  if(payload_fec_size == 0)
                        num_buckets--;
		  
		  break;
	case 100:   NBYTES[0] = hdr_data_size;
		  NBYTES[1] = (expectedPayloadWithHdrLen - hdr_data_size) + (fec_N - fec_K) + payload_fec_size;
		  num_buckets = 2;
		  break;

        case 4:   NBYTES[0] = expectedPayloadWithHdrLen;
                  NBYTES[1] = (fec_N - fec_K) + payload_fec_size;
                  num_buckets = 2;
                  break;

        case 5:   NBYTES[0] = hdr_data_size + (fec_N - fec_K);
                  NBYTES[1] = (expectedPayloadWithHdrLen - hdr_data_size) + payload_fec_size;
                  num_buckets = 2;
                  break;

        case 6:   NBYTES[0] = hdr_data_size;
                  NBYTES[1] = (fec_N - fec_K);
		  NBYTES[2] = expectedPayloadWithHdrLen - hdr_data_size;
                  NBYTES[3] = payload_fec_size;
                  num_buckets = 4;
		  if(payload_fec_size == 0)
			num_buckets--;
		  break;

	case 7:   assert(false);
    }

#ifndef DEBUG 
    printf("hdr size: %d\n", hdr_data_size);
    printf("payload size: %d\n", expectedPayloadWithHdrLen - hdr_data_size);
    printf("hdr-fec size: %d\n", (fec_N - fec_K));
    printf("payload-fec size: %d\n", payload_fec_size);
    printf("expectedPayloadWithHdrLen: %d\n", expectedPayloadWithHdrLen);
    printf("buf.size: %d\n", buf.size());
#endif

    if(mac_parity_added > 0)
    {
	std::string orig_s = getBits((const unsigned char*) buf.data(), buf.size(), 8);
	parity_bits_str = orig_s.substr(orig_s.size() - mac_parity_added, mac_parity_added);

	printf("mac_parity_bits: %d\n", mac_parity_added); 
	buf = buf.substr(0, buf.size() - mac_parity_added/8);
	printf("buf.size: %d\n", buf.size());
    }

    // determine the padding that might be used by the interleaver to fill the subcarriers //
    int padding = abs((int)(buf.size() - (expectedPayloadWithHdrLen + (fec_N - fec_K) + payload_fec_size)));

    // read the csi and sort the subcarriers only once //
    static bool initialized = false;
    static float *csi = (float*) malloc(sizeof(float) * subcarriers);
    static int *sorted_sc_indices = (int*) malloc(sizeof(int) * subcarriers);
    if(!initialized)
    {
	initialized = true;
	readCSISubcarriers(csi, subcarriers);
	if(scheme > 0)
	{
            float *sorted_csi = (float*) malloc(sizeof(float) * subcarriers);
            memcpy(sorted_csi, csi, sizeof(float) * subcarriers);
	    sortSubcarriersByCSI(sorted_sc_indices, sorted_csi, subcarriers);
    	    free(sorted_csi);
	}
	else
	{
		// for scheme=0, we need not sort the subcarriers //
	    for(int i = 0; i < subcarriers; i++)
		sorted_sc_indices[i] = i;	
	}
    }

    size_t len = buf.size();
    int s_len = len * 8;

    int num_symbols = ceil((float) s_len/bps);
    int rounds = ceil((float) num_symbols/subcarriers);
    //printf("num_symbols: %d, subcarriers: %d\n", num_symbols, subcarriers);
    assert(num_symbols % subcarriers == 0);

    // similar to the interleaver, pretend the 'padding' is added to the last bucket, filter it out later //
    NBYTES[num_buckets - 1] += padding;
    
    // tracker helps in debugging //
    unsigned char *tracker = (unsigned char*) malloc(sizeof(unsigned char) * s_len);
    int *orig_bit_pos = (int*) malloc(sizeof(int) * s_len);
    int *final_pos = (int*) malloc(sizeof(int) * s_len);
    bool initialized1 = false;
    if(!initialized1)
    {
	initialized1 = true;
	memset(tracker, 0, sizeof(unsigned char) * s_len);
	memset(orig_bit_pos, 0, sizeof(int) * s_len);
	memset(final_pos, 0, sizeof(int) * s_len);

	int bits_demapped_already = 0;
	int subcarriers_used_already = 0;
	int offset = 0;
	int _pad = 0;
	int num_sp_bits = 0;
	for(int i = 0; i < num_buckets; i++)
        {
            if(NBYTES[i] > 0)
            {
                int rem_bits = s_len - bits_demapped_already;
                if(rem_bits <= NBYTES[i] * 8)
                    num_sp_bits = rem_bits;
		else
			num_sp_bits = NBYTES[i] * 8 - _pad;                     // padding from the previous bucket might have spilled over

		if(num_sp_bits <= 0)
			continue;	

		int num_sp_symbols = ceil((float)num_sp_bits/bps);
		int n_subcarriers = ceil((float)num_sp_symbols/rounds);   

		//printf("rem_bits: %d\n", rem_bits); fflush(stdout); 	

		// (possibly) pad 'num_sp_bits' such that they occupy complete 'rounds' using 'n_best_subcarriers' //
		int n_bits = (n_subcarriers * rounds * bps > num_sp_bits) ? n_subcarriers * rounds * bps : num_sp_bits; 

		if(n_bits > num_sp_bits)
			_pad = n_bits - num_sp_bits;

		assert(n_bits%bps == 0);	
		deinterleave_bucket(len, n_bits, n_subcarriers, 
				bits_demapped_already, subcarriers, 
				offset, bps, rounds, sorted_sc_indices + subcarriers_used_already, 
				orig_bit_pos, final_pos, tracker, scheme);	  

		subcarriers_used_already += n_subcarriers;
	    }
    	}
    }

    int *final_pos_map = (int*) malloc(sizeof(int) * s_len);
    memset(final_pos_map, 0, sizeof(int) * s_len);

    int *orig_pos_map = (int*) malloc(sizeof(int) * s_len);
    memset(orig_pos_map, 0, sizeof(int) * s_len);

        // for mac-fec //
    for(int i = 0; i < s_len; i++)
    {
	int orig_pos = orig_bit_pos[final_pos[i]];
	final_pos_map[orig_pos] = i;
	orig_pos_map[i] = orig_pos;
    }	

    
#ifdef DEBUG
    printf("position map (deinterleaver)\n");
    for(int i = 0; i < s_len; i++)
	printf("%d -> %d -> %d\n", i, final_pos[i], orig_bit_pos[final_pos[i]]);
#endif

    // form the priority string //
    const unsigned char *data = (const unsigned char*) buf.data();
    std::string orig_s = getBitString(data, len);
    unsigned char* priority_bits = (unsigned char*) malloc(sizeof(unsigned char) * s_len);
    for(int i = 0; i < s_len; i++)
	priority_bits[orig_bit_pos[final_pos[i]]] = orig_s[i];

   // filter out the padding now //
   len = len - padding;
   s_len = len * 8; 

   // convert from the priority string back to the desired original string //
   std::string priority_s((const char*) priority_bits, s_len); 
   unsigned char *priority_data = (unsigned char*) malloc(sizeof(unsigned char) * len);
   memset(priority_data, 0, sizeof(unsigned char) * len);
   std::string priority_str = getMsg(priority_s, priority_data);
   NBYTES[num_buckets - 1] -= padding;

   std::string hdr_payload_only = ""; 		// hdr and payload together in 1 bucket
   std::string hdr_only = "";
   std::string hdr_fec_only = "";
   std::string payload_only = "";
   std::string payload_fec_only = "";
   std::string payload_fec = "";
   int pad_hdr = 0;
   std::string others = ""; 			// scheme: 4
   int size_payload_fec = 0;
   std::string hdr_fec_str = "";
   std::string all_fec = "";
   std::string deinterleaved_str = "";
   switch(scheme)
   {
	case 0:	deinterleaved_str = priority_str;
		break;
	
	case 1:	hdr_only = priority_str.substr(0, NBYTES[0]);
		payload_only = priority_str.substr(NBYTES[0], NBYTES[1]);

		// first few bytes of payload should go with the hdr (same fec group)
		pad_hdr = ((fec_K - hdr_data_size) < payload_only.size()) ? fec_K - hdr_data_size: payload_only.size();
		hdr_only.append(payload_only.substr(0, pad_hdr));	
	
		printf("pad_hdr: %d, payload_only: %d\n", pad_hdr, payload_only.size());
		payload_only = payload_only.substr(pad_hdr, payload_only.size() - pad_hdr);

		if(fec_K == 0)
			return hdr_only + payload_only;
		printf("size: %d %d %d %d\n", priority_str.size(), NBYTES[0], NBYTES[1], NBYTES[2]);
		hdr_fec_only = priority_str.substr(NBYTES[0]+ NBYTES[1], NBYTES[2]);
		
		payload_fec_only = priority_str.substr(NBYTES[0] + NBYTES[1] + NBYTES[2], NBYTES[3]);
		printf("payload_fec_only: %d\n", payload_fec_only.size());

		payload_fec = reconstruct_payload_fec(payload_only, payload_fec_only, fec_N, fec_K);
		deinterleaved_str = hdr_only + hdr_fec_only + payload_fec;
		break;

        case 2: payload_only = priority_str.substr(0, NBYTES[0]);
		hdr_only = priority_str.substr(NBYTES[0], NBYTES[1]);

                // first few bytes of payload should go with the hdr (same fec group)
                pad_hdr = ((fec_K - hdr_data_size) < payload_only.size()) ? fec_K - hdr_data_size: payload_only.size();
                hdr_only.append(payload_only.substr(0, pad_hdr));
	        payload_only = payload_only.substr(pad_hdr, payload_only.size() - pad_hdr);

		if(fec_K == 0)
			return hdr_only + payload_only;

                hdr_fec_only = priority_str.substr(NBYTES[0]+ NBYTES[1], NBYTES[2]);
                payload_fec_only = priority_str.substr(NBYTES[0] + NBYTES[1] + NBYTES[2], NBYTES[3]);
                payload_fec = reconstruct_payload_fec(payload_only, payload_fec_only, fec_N, fec_K);
                deinterleaved_str = hdr_only + hdr_fec_only + payload_fec;
                break;

        case 3: hdr_payload_only = priority_str.substr(0, NBYTES[0]);
		hdr_only = hdr_payload_only.substr(0, hdr_data_size);
                payload_only = hdr_payload_only.substr(hdr_data_size, NBYTES[0] - hdr_data_size);

                // first few bytes of payload should go with the hdr (same fec group)
		pad_hdr = ((fec_K - hdr_data_size) < payload_only.size()) ? fec_K - hdr_data_size: payload_only.size();
                hdr_only.append(payload_only.substr(0, pad_hdr));

                payload_only = payload_only.substr(pad_hdr, payload_only.size() - pad_hdr);

                if(fec_K == 0)
                        return hdr_only + payload_only;

                hdr_fec_only = priority_str.substr(NBYTES[0], NBYTES[1]);
                payload_fec_only = priority_str.substr(NBYTES[0] + NBYTES[1], NBYTES[2]);
                payload_fec = reconstruct_payload_fec(payload_only, payload_fec_only, fec_N, fec_K);
                deinterleaved_str = hdr_only + hdr_fec_only + payload_fec;
                break;

        case 100: hdr_only = priority_str.substr(0, NBYTES[0]);
                others = priority_str.substr(NBYTES[0], NBYTES[1]);

                payload_only = others.substr(0, expectedPayloadWithHdrLen - hdr_only.size());
                hdr_fec_only = others.substr(payload_only.size(), (fec_N - fec_K));
                payload_fec_only = others.substr(payload_only.size() + hdr_fec_only.size(), others.size() - (payload_only.size() + hdr_fec_only.size()));

                // first few bytes of payload should go with the hdr (same fec group)
                pad_hdr = ((fec_K - hdr_data_size) < payload_only.size()) ? fec_K - hdr_data_size: payload_only.size();
                hdr_only.append(payload_only.substr(0, pad_hdr));
                payload_only = payload_only.substr(pad_hdr, payload_only.size() - pad_hdr);

                if(fec_K == 0)
                        return hdr_only + payload_only;

                payload_fec = reconstruct_payload_fec(payload_only, payload_fec_only, fec_N, fec_K);
                deinterleaved_str = hdr_only + hdr_fec_only + payload_fec;
                break;

	case 4: hdr_payload_only = priority_str.substr(0, NBYTES[0]);
		hdr_only = hdr_payload_only.substr(0, hdr_data_size);
		payload_only = hdr_payload_only.substr(hdr_data_size, NBYTES[0] - hdr_data_size);

		// first few bytes of payload should go with the hdr (same fec group)
                pad_hdr = ((fec_K - hdr_data_size) < payload_only.size()) ? fec_K - hdr_data_size: payload_only.size();
                hdr_only.append(payload_only.substr(0, pad_hdr));
                payload_only = payload_only.substr(pad_hdr, payload_only.size() - pad_hdr);

                if(fec_K == 0)
                        return hdr_only + payload_only;

                all_fec = priority_str.substr(NBYTES[0], NBYTES[1]);
	        hdr_fec_only = all_fec.substr(0, (fec_N - fec_K));		
	        size_payload_fec = all_fec.size() - hdr_fec_only.size();	
 	
	        payload_fec_only = all_fec.substr((fec_N - fec_K), size_payload_fec);

	        payload_fec = reconstruct_payload_fec(payload_only, payload_fec_only, fec_N, fec_K);
	        deinterleaved_str = hdr_only + hdr_fec_only + payload_fec;
	        break;

	case 5: hdr_fec_str = priority_str.substr(0, NBYTES[0]);
		payload_fec = priority_str.substr(NBYTES[0], NBYTES[1]);
	
		hdr_only = hdr_fec_str.substr(0, hdr_data_size);
		hdr_fec_only = hdr_fec_str.substr(hdr_data_size, fec_N - fec_K);

		payload_only = payload_fec.substr(0, expectedPayloadWithHdrLen - hdr_only.size());
		payload_fec_only = payload_fec.substr(payload_only.size(), payload_fec.size() - payload_only.size());	

		if(fec_K == 0)
			return hdr_only + payload_only;		

		// first few bytes of payload should go with the hdr (same fec group)
		pad_hdr = ((fec_K - hdr_data_size) < payload_only.size()) ? fec_K - hdr_data_size: payload_only.size();
		hdr_only.append(payload_only.substr(0, pad_hdr));
		payload_only = payload_only.substr(pad_hdr, payload_only.size() - pad_hdr);
	
		payload_fec = reconstruct_payload_fec(payload_only, payload_fec_only, fec_N, fec_K);
		deinterleaved_str = hdr_only + hdr_fec_only + payload_fec;
		break;

	case 6: hdr_only = priority_str.substr(0, NBYTES[0]);
		hdr_fec_only = priority_str.substr(NBYTES[0], NBYTES[1]);

		payload_only = priority_str.substr(NBYTES[0] + NBYTES[1], NBYTES[2]);
		payload_fec_only = priority_str.substr(NBYTES[0] + NBYTES[1] + NBYTES[2], NBYTES[3]);	

		// first few bytes of payload should go with the hdr (same fec group)
		pad_hdr = ((fec_K - hdr_data_size) < payload_only.size()) ? fec_K - hdr_data_size: payload_only.size();
		hdr_only.append(payload_only.substr(0, pad_hdr));
		payload_only = payload_only.substr(pad_hdr, payload_only.size() - pad_hdr);

		payload_fec = reconstruct_payload_fec(payload_only, payload_fec_only, fec_N, fec_K);
		deinterleaved_str = hdr_only + hdr_fec_only + payload_fec;
		break;
   }

    if(mac_fec_flag)
    {
#ifdef DEBUG
	printf("rcvr: before deinterleaving.. \n");
	for(int i = 0; i < parity_bits_str.size(); i++)
		printf("%c", parity_bits_str[i]);
	printf("\n");
#endif

	std::string deinterleaved_bits = getBits((const unsigned char*) deinterleaved_str.data(), deinterleaved_str.size(), 8);
	int num_bits_before_deinterleaving = buf.size() * 8;
	int num_parity = parity_bits_str.size();
	num_parity = num_parity - (num_parity % 8);

	parity_bits_str = parity_bits_str.substr(0, num_parity);
	

	// deinterleave the parity and separate the padding from parity (look at interleave_mapping_best) //
	int parity_one_round = subcarriers * bps;
	
	unsigned char *final_parity_bits = (unsigned char*) malloc(sizeof(unsigned char) * mac_parity_added + 1);
	memset(final_parity_bits, 0, sizeof(unsigned char) * mac_parity_added + 1);
	
	int interleaved_parity_pos[subcarriers * bps];
	initInterleavedParityPos(interleaved_parity_pos, bps, subcarriers);
	
	printf("mac_parity_added: %d\n", mac_parity_added); 
	for(int i = 0; i < mac_parity_added; i++)
	{
		int index = i % parity_one_round;
		int blocks_mapped = i / parity_one_round;
		int new_pos = interleaved_parity_pos[index] + blocks_mapped * parity_one_round;
		assert(new_pos < ((blocks_mapped + 1) * parity_one_round));
		final_parity_bits[i] = parity_bits_str[new_pos];
	}

	parity_bits_str.assign((const char*) final_parity_bits);	

#ifdef DEBUG
	printf("rcvr: after deinterleaving.. \n");
	for(int i = 0; i < parity_bits_str.size(); i++)
		printf("%c", parity_bits_str[i]);
	printf("\n");
#endif

	int symSize = 1;      // for LDPC, its immaterial, just keep it at 1
#ifdef RS_MAC_PARITY
        symSize = getSymSizeFromN_Syms(fec_N);
  	fec_N = fec_N * symSize;
  	fec_K = fec_K * symSize;

#endif
        decodeMACLayer_FEC_OPT(deinterleaved_bits,
                               parity_bits_str,
                               final_pos_map,
                               orig_pos_map,
                               subcarriers, fec_N,
                               fec_K, bps, 
			       num_bits_before_deinterleaving,
                               fec_split_map, sorted_orig_pos);
    }

    free(orig_pos_map);
    free(final_pos_map);
    free(tracker);
    free(orig_bit_pos);
    free(final_pos);
    free(priority_bits);
    free(priority_data);
    return deinterleaved_str;	
}

// alternate fec group delivery rate computation - //
/* use normal distribution to estimate the FEC group delivery rate * * X(i) = probability that codeword 'i' is lost * Y = sum X(i) */
double normal_cdf(double z)
{
  if (z > 6.0) { return 1.0; };
  if (z < -6.0) { return 0.0; };
  double b1 = 0.31938153;
  double b2 = -0.356563782;
  double b3 = 1.781477937;
  double b4 = -1.821255978;
  double b5 = 1.330274429;
  double p = 0.2316419;
  double c2 = 0.3989423;

  double a=fabs(z);
  double t = 1.0/(1.0+a*p);
  double b = c2*exp((-z)*(z/2.0));
  double n = (((((b5*t+b4)*t)+b3)*t+b2)*t+b1)*t;
  n = 1.0-b*n;
  if (z < 0.0) n = 1.0-n;
  return n;
}


float getFECGroupDeliveryRate(float *t_eff_delivery_rate, int nsymbols, int fec_N, int fec_K)
{

   float mean_loss_rate = 0.0;
   float var_loss_rate = 0.0;

   for(int i = 0; i < nsymbols; i++)
   {
        float Pi = 1 - t_eff_delivery_rate[i];
        mean_loss_rate += Pi;
        var_loss_rate += (Pi * (1 - Pi));
   }

   float sigma_loss_rate = sqrt(var_loss_rate);

   float threshold = (float) (fec_N - fec_K)/2;

   float normalized_threshold = (threshold-mean_loss_rate)/sigma_loss_rate;

   return(normal_cdf(normalized_threshold));
}

/* returns the interleaved parity bit locations in an interleaver block */
void initInterleavedParityPos(int *interleaved_parity_pos, int bps, int subcarriers)
{
   int C = 16;
   int Ncbps = subcarriers * bps;
   int s = (bps/2 > 1) ? bps/2 : 1;

   for(int k = 0; k < Ncbps; k++)
   {
      int i = (Ncbps/C)*(k%C) + floor((float)k/C);
      interleaved_parity_pos[k] = s * floor((float) i/s) + (int) (i + Ncbps - floor((float) C * i/Ncbps)) % s;
   }
}

static struct rs 
*gr_init_rs(unsigned int symsize,unsigned int gfpoly,unsigned fcr,unsigned prim, unsigned int nroots)
{
  struct rs *rs;
  int i, j, sr,root,iprim;

  printf("coming in......\n");

  if(symsize > 8*sizeof(DTYPE)){
    printf("Need version with ints rather than chars\n");
    exit(0);
    //return NULL; /* Need version with ints rather than chars */
  }
  if(fcr >= (1<<symsize)){
    printf("The fcr value is abnormal.\n");
    exit(0);
    //return NULL;
  }
  if(prim == 0 || prim >= (1<<symsize)){
    printf("The prim value is abnormal...\n");
    exit(0);
    //return NULL;
  }
  if(nroots >= (1<<symsize)){
    printf("Can't have more roots than symbol values!\n");
    exit(0);
    //return NULL; /* Can't have more roots than symbol values! */
  }

  rs = (struct rs *)calloc(1,sizeof(struct rs));
  rs->mm = symsize;
  rs->nn = (1<<symsize)-1;

  rs->alpha_to = (DTYPE *)malloc(sizeof(DTYPE)*(rs->nn+1));
  if(rs->alpha_to == NULL){
    free(rs);
    printf("Cannot set the values of 'alpha_to[]' of reedsolomon...\n");
    exit(0);
    //return NULL;
  }

  rs->index_of = (DTYPE *)malloc(sizeof(DTYPE)*(rs->nn+1));
  if(rs->index_of == NULL){
    free(rs->alpha_to);
    free(rs);
    printf("Cannot set the value of index of reedsolomon...\n");
    exit(0);
    //return NULL;
  }

  /* Generate Galois field lookup tables */
  rs->index_of[0] = A0; /* log(zero) = -inf */
  rs->alpha_to[A0] = 0; /* alpha**-inf = 0 */
  sr = 1;
  for(i=0;i<rs->nn;i++){
    rs->index_of[sr] = i;
    rs->alpha_to[i] = sr;
    sr <<= 1;
    if(sr & (1<<symsize))
      sr ^= gfpoly;
    sr &= rs->nn;
  }
  if(sr != 1){
    /* field generator polynomial is not primitive! */
    free(rs->alpha_to);
    free(rs->index_of);
    free(rs);
    printf("field generator polynomial is not primitive!\n");
    exit(0);
    //return NULL;
  }

  /* Form RS code generator polynomial from its roots */
  rs->genpoly = (DTYPE *)malloc(sizeof(DTYPE)*(nroots+1));
  if(rs->genpoly == NULL){
    free(rs->alpha_to);
    free(rs->index_of);
    free(rs);
    printf("Cannot set the values of generation polynomial of reedsolomon...\n");
    exit(0);
    //return NULL;
  }
  rs->fcr = fcr;
  rs->prim = prim;
  rs->nroots = nroots;

  /* Find prim-th root of 1, used in decoding */
  for(iprim=1;(iprim % prim) != 0;iprim += rs->nn)
    ;
  rs->iprim = iprim / prim;

  rs->genpoly[0] = 1;
  for (i = 0,root=fcr*prim; i < nroots; i++,root += prim) {
    rs->genpoly[i+1] = 1;

    /* Multiply rs->genpoly[] by  @**(root + x) */
    for (j = i; j > 0; j--){
      if (rs->genpoly[j] != 0)
        rs->genpoly[j] = rs->genpoly[j-1] ^ rs->alpha_to[modnn(rs,rs->index_of[rs->genpoly[j]] + root)];
      else
        rs->genpoly[j] = rs->genpoly[j-1];
    }
    /* rs->genpoly[0] can never be zero */
    rs->genpoly[0] = rs->alpha_to[modnn(rs,rs->index_of[rs->genpoly[0]] + root)];
  }
  /* convert rs->genpoly[] to index form for quicker encoding */
  for (i = 0; i <= nroots; i++)
    rs->genpoly[i] = rs->index_of[rs->genpoly[i]];

  return rs;
}

// apurv++ begins //
void gr_encode_rs(void *p, DTYPE *data, DTYPE *bb){

  struct rs *rs = (struct rs *)p;
  int i, j;
  DTYPE feedback;

  memset(bb,0,NROOTS*sizeof(DTYPE));

  for(i=0;i<NN-NROOTS;i++){
    feedback = INDEX_OF[data[i] ^ bb[0]];
    if(feedback != A0){      /* feedback term is non-zero */
#ifdef UNNORMALIZED
      /* This line is unnecessary when GENPOLY[NROOTS] is unity, as it must
       * always be for the polynomials constructed by init_rs()
       */
      feedback = MODNN(NN - GENPOLY[NROOTS] + feedback);
#endif
      for(j=1;j<NROOTS;j++)
        bb[j] ^= ALPHA_TO[MODNN(feedback + GENPOLY[NROOTS-j])];
    }
    /* Shift */
    memmove(&bb[0],&bb[1],sizeof(DTYPE)*(NROOTS-1));
    if(feedback != A0)
      bb[NROOTS-1] = ALPHA_TO[MODNN(feedback + GENPOLY[0])];
    else
      bb[NROOTS-1] = 0;
  }
}

std::string gr_encode_fec(std::string buf, int n, int k)
{
    // qiongwu - added for supporting different % of FEC //
    int blocklength;    // input variable //TODO: make them as inputs
    //float n,k;          // input variables

    int syms;
    int gfpoly;
    int nn, nroots;

    //these three variables just for the use of debug
    blocklength = 255;  // the encode/decode block length
    //k = 1; n = 2;       // (255,223) scheme for reed-solomon

    switch(blocklength){
	/*	
        case 63:
                syms = 6; gfpoly = 0x43; break;
        case 127:
                syms = 7; gfpoly = 0x89; break;
	*/
        case 255:
                syms = 8; gfpoly = 0x11d; break;
	/*
        case 511:
                syms = 9; gfpoly = 0x211; break;
        case 1023:
                syms = 10; gfpoly = 0x409; break;
        case 2047:
                syms = 11; gfpoly = 0x805; break;
        case 4095:
                syms = 12; gfpoly = 0x1053; break;
	*/
        default:
                printf("the length of the block is out of range, FEC cannot be initiated.");
                break;
    }
    nn = (1<<syms)-1;
    nroots = ceil(nn*(1-(float)k/n));
    if((nroots%2)!=0)
    {
        nroots++;
    }
    //printf("nn = %d; nroots=%d, and the percentage is %f\n", nn, nroots, (float)k/n );
    // qiongwu ends //

    static struct rs *rs;
    static unsigned int handle_init = 0;
    // initializes the handle only once //
    if(handle_init == 0)
    {
       printf("entering initial stage......\n");
       //rs = gr_init_rs(8, 0x11d, 1, 1, 32);
       rs = gr_init_rs(syms, gfpoly, 1, 1, nroots);
       printf("complete initial stage, entering gr_update_crc32 #1, handle%p\n", rs);
       handle_init++;
    }

    const unsigned char *data = (const unsigned char*) buf.data();
    size_t len = buf.size();

#ifdef DEBUG_FEC
    printf("raw buf -\n");
    for(int k = 0; k < len; k++)
    {
        printf("%d ", data[k]);
    }
    printf("\n");
#endif

    int num_blocks = ceil(((float)len/(rs->nn-rs->nroots)));
    printf("nn%d nroots%d len%d data+len%p\n", rs->nn, rs->nroots, len, data+len); fflush(stdout);
    int len_w_fec = (num_blocks) * NN;
    int rem = len % (NN-NROOTS);
    if(rem > 0)
    {
	len_w_fec = (num_blocks-1) * NN + rem + NROOTS;
    }

    /* encoded_data: the data that now contains the FEC as well */
    unsigned char *encoded_data = (unsigned char *) malloc(sizeof(unsigned char) * len_w_fec);
    memset(encoded_data, 0, sizeof(unsigned char) * len_w_fec);

#ifdef DEBUG_FEC
    printf("len_w_fec%d, num_blocks%d NN%d NROOTS%d\n", len_w_fec, num_blocks, NN, NROOTS); fflush(stdout);
#endif

    int dataLen = NN-NROOTS;
    int processedDataLen = 0;
    int pad_zeros = 0;
    unsigned char data_block[NN];	//temporary block of data//
    memset(data_block, 0, NN);
    for (int i=0; i<num_blocks; i++)
    {
        printf("starting inner loop, i%d (NN-NROOTS)%d processedDataLen%d\n", i, NN-NROOTS, processedDataLen);
        fflush(stdout);

        int num_bytes_to_encode = 0;

        if(processedDataLen + dataLen > len)
        {
             num_bytes_to_encode = len - processedDataLen;
             pad_zeros = dataLen - num_bytes_to_encode;
        }
        else
        {
             num_bytes_to_encode = dataLen;
        }
        processedDataLen += num_bytes_to_encode;
	
	printf("num_bytes_to_encode%d pad_zeros%d\n", num_bytes_to_encode, pad_zeros); fflush(stdout);
	
        // copy the data portion of this block into the encoded_data //
	memcpy(data_block, data, num_bytes_to_encode);

        data = data + num_bytes_to_encode;

        // use the shortened RS code if necessary, ie pad with 0s //
	memset(data_block + num_bytes_to_encode, 0, pad_zeros);

#ifdef DEBUG_FEC
       // debug apurv start //
       printf("raw data_block - \n");
       for(int k = 0; k < NN; k++)
       {
           printf("%d ", data_block[k]);
       }
       printf("\n\n"); fflush(stdout);
       // debug end //
#endif

        // encode this block using FEC //
	gr_encode_rs(rs, &data_block[0], &data_block[NN-NROOTS]);

#ifdef DEBUG_FEC
       // debug apurv start //
       printf("encoded data_block - \n");
       for(int k = 0; k < NN; k++)
       {
           printf("%d ", data_block[k]);
       }
       printf("\n\n"); fflush(stdout);
       // debug end //
#endif

       // copy the data part //
       memcpy(encoded_data + i * NN, data_block, num_bytes_to_encode);

       // copy the parity part //
       memcpy(encoded_data + i * NN + num_bytes_to_encode, data_block + (NN-NROOTS), NROOTS);
    }

#ifdef DEBUG_FEC
    // debug apurv start //
    printf("encoded data - \n");
    for(int i = 0; i < len_w_fec; i++)
    {
        printf("%d ", encoded_data[i]);
    }
    printf("\n\n"); fflush(stdout);
    // debug end //
#endif

    std::string s1((const char*) encoded_data, len_w_fec);

    printf("making packets complete...... size%d\n", s1.size());
    return s1;
}

int
gr_decode_rs(struct rs *rs, unsigned char *data)
{

  int deg_lambda, el, deg_omega;
  int i, j, r,k;
  DTYPE u,q,tmp,num1,num2,den,discr_r;
  DTYPE lambda[NROOTS+1], s[NROOTS];    /* Err+Eras Locator poly
                                         * and syndrome poly */
  DTYPE b[NROOTS+1], t[NROOTS+1], omega[NROOTS+1];
  DTYPE root[NROOTS], reg[NROOTS+1], loc[NROOTS];
  int syn_error, count;

  /* form the syndromes; i.e., evaluate data(x) at roots of g(x) */
  for(i=0;i<NROOTS;i++)
    s[i] = data[0];

  for(j=1;j<NN;j++){
    for(i=0;i<NROOTS;i++){
      if(s[i] == 0){
        s[i] = data[j];
      } else {
        s[i] = data[j] ^ ALPHA_TO[MODNN(INDEX_OF[s[i]] + (FCR+i)*PRIM)];
      }
    }
  }
  /* Convert syndromes to index form, checking for nonzero condition */
  syn_error = 0;
  for(i=0;i<NROOTS;i++){
    syn_error |= s[i];
    s[i] = INDEX_OF[s[i]];
  }

  if (!syn_error) {
    /* if syndrome is zero, data[] is a codeword and there are no
     * errors to correct. So return data[] unmodified
     */
    count = 0;
    goto finish;
  }
  memset(&lambda[1],0,NROOTS*sizeof(lambda[0]));
  lambda[0] = 1;

  for(i=0;i<NROOTS+1;i++)
    b[i] = INDEX_OF[lambda[i]];

  /*
   * Begin Berlekamp-Massey algorithm to determine error+erasure
   * locator polynomial
   */
  r = 0;
  el = 0;
  while (++r <= NROOTS) {       /* r is the step number */
    /* Compute discrepancy at the r-th step in poly-form */
    discr_r = 0;
    for (i = 0; i < r; i++){
      if ((lambda[i] != 0) && (s[r-i-1] != A0)) {
        discr_r ^= ALPHA_TO[MODNN(INDEX_OF[lambda[i]] + s[r-i-1])];
      }
    }
    discr_r = INDEX_OF[discr_r];        /* Index form */
    if (discr_r == A0) {
      /* 2 lines below: B(x) <-- x*B(x) */
      memmove(&b[1],b,NROOTS*sizeof(b[0]));
      b[0] = A0;
    } else {
      /* 7 lines below: T(x) <-- lambda(x) - discr_r*x*b(x) */
      t[0] = lambda[0];
      for (i = 0 ; i < NROOTS; i++) {
        if(b[i] != A0)
          t[i+1] = lambda[i+1] ^ ALPHA_TO[MODNN(discr_r + b[i])];
        else
          t[i+1] = lambda[i+1];
      }
      if (2 * el <= r - 1) {
        el = r - el;
        /*
         * 2 lines below: B(x) <-- inv(discr_r) *
         * lambda(x)
         */
        for (i = 0; i <= NROOTS; i++)
          b[i] = (lambda[i] == 0) ? A0 : MODNN(INDEX_OF[lambda[i]] - discr_r + NN);
      } else {
        /* 2 lines below: B(x) <-- x*B(x) */
        memmove(&b[1],b,NROOTS*sizeof(b[0]));
        b[0] = A0;
      }
      memcpy(lambda,t,(NROOTS+1)*sizeof(t[0]));
    }
  }


  /* Convert lambda to index form and compute deg(lambda(x)) */
  deg_lambda = 0;
  for(i=0;i<NROOTS+1;i++){
    lambda[i] = INDEX_OF[lambda[i]];
    if(lambda[i] != A0)
      deg_lambda = i;
  }
  /* Find roots of the error+erasure locator polynomial by Chien search */
  memcpy(&reg[1],&lambda[1],NROOTS*sizeof(reg[0]));
  count = 0;            /* Number of roots of lambda(x) */
  for (i = 1,k=IPRIM-1; i <= NN; i++,k = MODNN(k+IPRIM)) {
    q = 1; /* lambda[0] is always 0 */
    for (j = deg_lambda; j > 0; j--){
      if (reg[j] != A0) {
        reg[j] = MODNN(reg[j] + j);
        q ^= ALPHA_TO[reg[j]];
      }
    }
    if (q != 0)
      continue; /* Not a root */
    /* store root (index-form) and error location number */
    root[count] = i;
    loc[count] = k;
    /* If we've already found max possible roots,
     * abort the search to save time
     */
    if(++count == deg_lambda)
      break;
  }
  if (deg_lambda != count) {
    /*
     * deg(lambda) unequal to number of roots => uncorrectable
     * error detected
     */
    count = -1;
    goto finish;
  }
  /*
   * Compute err+eras evaluator poly omega(x) = s(x)*lambda(x) (modulo
   * x**NROOTS). in index form. Also find deg(omega).
   */
  deg_omega = 0;
  for (i = 0; i < NROOTS;i++){
    tmp = 0;
    j = (deg_lambda < i) ? deg_lambda : i;
    for(;j >= 0; j--){
      if ((s[i - j] != A0) && (lambda[j] != A0))
        tmp ^= ALPHA_TO[MODNN(s[i - j] + lambda[j])];
    }
    if(tmp != 0)
      deg_omega = i;
    omega[i] = INDEX_OF[tmp];
  }
  omega[NROOTS] = A0;

  /*
   * Compute error values in poly-form. num1 = omega(inv(X(l))), num2 =
   * inv(X(l))**(FCR-1) and den = lambda_pr(inv(X(l))) all in poly-form
   */
  for (j = count-1; j >=0; j--) {
    num1 = 0;
    for (i = deg_omega; i >= 0; i--) {
      if (omega[i] != A0)
        num1  ^= ALPHA_TO[MODNN(omega[i] + i * root[j])];
    }
    num2 = ALPHA_TO[MODNN(root[j] * (FCR - 1) + NN)];
    den = 0;

    /* lambda[i+1] for i even is the formal derivative lambda_pr of lambda[i] */
    for (i = min(deg_lambda,NROOTS-1) & ~1; i >= 0; i -=2) {
      if(lambda[i+1] != A0)
        den ^= ALPHA_TO[MODNN(lambda[i+1] + i * root[j])];
    }
    if (den == 0) {
      count = -1;
      goto finish;
    }
    /* Apply error to data */
    if (num1 != 0) {
      data[loc[j]] ^= ALPHA_TO[MODNN(INDEX_OF[num1] + INDEX_OF[num2] + NN - INDEX_OF[den])];
    }
  }

  finish:
#ifdef DEBUG_FEC
     printf(" --------------------------------------------\n");
     printf("\nin decode_rs: count%d\n\n", count); fflush(stdout);
     printf(" --------------------------------------------\n");
#endif
     return count;
}

std::string
gr_decode_rs_fec(std::string buf, int n, int k, int dataLen,
		 std::string parity_bits_str, bool mac_fec
#ifdef RS_MAC_OPT_VERSION
		 ,std::map<int, split> fec_split_map,
		 std::vector<int> sorted_orig_pos
#else
		 ,std::vector<int> orig_pos_indices 
#endif
		 , int &total_fec_grps, int &num_correct_fec_grps, int &num_correct_bytes)		 
{	
    bool log_data = false;
  
    static std::vector<struct rs*> rs_vec;
    static struct rs *rs = find_rs_handle(rs_vec, n, k);
    assert(rs);
    int syms = rs->mm;

    //printf("nn = %d; nroots=%d, \n", rs->nn, rs->nroots);
     // get bits from the byte string //
    std::string s_bits = getBits((const unsigned char*) buf.data(), buf.size(), 8);
  
    int num_bits = s_bits.size();
    unsigned char *data = (unsigned char*) malloc(sizeof(unsigned char) * ceil(num_bits/syms));
    memset(data, 0, sizeof(unsigned char) * ceil(num_bits/syms));

    int len = convertToSymbols(data, syms, s_bits);
#ifdef DEBUG_FEC
    printf("in decode_rs_fec: %d---- \n", len);
    for(int i = 0; i < len; i++)
	  printf("%d ", data[i]);
    printf("\n");
#endif

    //injectErrors((unsigned char*) s_bits.data(), s_bits.size());
    convertToSymbols(data, syms, s_bits);

#ifdef DEBUG_FEC
    printf("after ERRORS nsymbols: %d------ \n", len);
    for(int i = 0; i < len; i++)
	  printf("%d ", data[i]);
    printf("\n");

    printf("received data %d symbols (symSize: %d)\n", len, syms);
    for(int i = 0; i < len; i++)
	printf("%d ", data[i]);
    printf("\n\n"); fflush(stdout);
#endif
  
  dataLen = dataLen * ceil((float) 8/syms);
  assert(len > dataLen);

  int num_blocks = ceil(((float)len/NN));
  

  //printf("len%d NN%d num_blocks%d dataLen%d\n", len, NN, num_blocks, dataLen);
  fflush(stdout);

  int processedDataLen = 0;
  unsigned char data_block[NN];
  //memset(data_block, 0, NN);
 
  unsigned char *decoded_data = (unsigned char*) malloc(dataLen);
  memset(decoded_data, 0, dataLen);
  int pad_zeros = 0;

  bool mac_parity_done = false;

  // to count the number of correct fec groups //
  num_correct_fec_grps = 0;
  total_fec_grps += num_blocks;
  num_correct_bytes = 0;

  for(int i=0; i<num_blocks; i++)
  {
     memset(data_block, 0, NN);
     int num_decoded = 0;
     if((processedDataLen + (NN - NROOTS))< dataLen)
     {
	num_decoded = NN-NROOTS;
     }
     else
     {
	if(i != num_blocks - 1)
	{
	   // flag as incorrect reception //
	   return buf;
	}
	assert(i == num_blocks - 1);
	num_decoded = dataLen - processedDataLen;
     }
    
     // perform the FEC decoding on the current <data_block> now //
     memcpy(data_block, data + i*NN, num_decoded);

     // copy the parity (NROOTS) to the data_block (leave space for padding if reqd)//
     memcpy(data_block + (NN-NROOTS), data + i*NN + num_decoded, NROOTS);

     memcpy(decoded_data + processedDataLen, data_block, num_decoded); 

#ifdef DEBUG_FEC
     printf("undecoded block[%d]: \n", i);
     for(int k = 0; k < NN; k++)
        printf("%d ", data_block[k]);
     printf("\n");

     // dump the block into a file //
     if(log_data)
     {
	printf("Logging the data num_decoded%d NROOTS%d-- \n", num_decoded, NROOTS);
        int fd;
	FILE *d_fp;
	char *d_filename = "fec.dat";
	if ((fd = open (d_filename,
			O_WRONLY|O_CREAT|O_TRUNC|OUR_O_LARGEFILE|OUR_O_BINARY,
			0777)) < 0){
	    perror(d_filename);
	}else {
	    if((d_fp = fdopen (fd, true ? "wb" : "w")) == NULL) {
	 	fprintf(stderr, "frame_sink file cannot be opened\n");
		close(fd);
	    }
	    fprintf(stderr, "fec file opened!\n");
	}

	int c = ftell(d_fp);	
        c = fwrite_unlocked(data_block, sizeof(unsigned char), NN, d_fp); 
	printf("logged %d items\n", c); fflush(stdout);
	c = ftell(d_fp);
	log_data = false;
     }
     // end dump //
#endif


     int retVal = gr_decode_rs(rs, data_block);
     //if(!mac_parity_done && mac_fec)
     if(retVal < 0 && !mac_parity_done && mac_fec)
     {
	//printf("retVal: %d before mac_fec\n", retVal); 
	bool ret = false;
#ifdef RS_MAC_OPT_VERSION
	printf("NN: %d, NROOTS: %d, NN-NROOTS: %d\n", NN, NROOTS, NN-NROOTS); fflush(stdout);
	ret = repairUsingMacParity_split_OPT(data + i*NN, NN, (NN-NROOTS), i, num_decoded,
					     parity_bits_str, fec_split_map,
					     sorted_orig_pos,	
			      		     rs->mm, rs_vec);	
#else
	ret = repairUsingMacParity(data + i*NN, NN, (NN-NROOTS), i, num_decoded,
         	 	              parity_bits_str, orig_pos_indices,
                		      0, 0, rs->mm, rs_vec);
#endif
	mac_parity_done = true;
	if(ret)
	{
	    printf("re-decode block: %d\n", i);
	    i--;
            continue;	
        }
     }
     //printf("retVal: %d after mac_fec\n", retVal);
     mac_parity_done = false;

     if(retVal >= 0){
        num_correct_fec_grps++;
        num_correct_bytes += num_decoded * 1;           // if using any other than N=255, then convert into bytes correctly!
        if(i == 0)
                num_correct_bytes -= 24;               // exclude the hdr
        if(i == num_blocks - 1)
                num_correct_bytes -= 4;                 // exclude the pkt crc
	memcpy(decoded_data + processedDataLen, data_block, num_decoded);
     }


#ifdef DEBUG_FEC
     printf("retVal: %d decoded block[%d]::::  \n", retVal, i); 
     for(int k = 0; k < NN; k++) {
	printf("%d ", data_block[k]);
     }
     printf("\n\n"); 
#endif
     
     processedDataLen += num_decoded;
  }
 
  if(processedDataLen != dataLen)
  {
     printf("nn%d nroots%d len%d num_blocks%d processedDataLen%d dataLen%d\n", NN, NROOTS,  
						len, num_blocks, processedDataLen, dataLen);
     assert(false);
  }

#ifdef DEBUG_FEC
  // debug apurv start //
  printf("decoded data - \n");
  for(int i = 0; i < dataLen; i++)
  {
      printf("%d ", decoded_data[i]);
  }
  printf("\n\n"); fflush(stdout);
  // debug end //
#endif

  // convert the symSize symbols back into bytes //
  unsigned char *decoded_bytes = (unsigned char *) malloc(sizeof(unsigned char) * dataLen);
  memset(decoded_bytes, 0, sizeof(unsigned char) * dataLen);

  std::string bits = getBits(decoded_data, dataLen, syms);
  int nbytes = convertToSymbols(decoded_bytes, 8, bits);

  free(decoded_data);

  std::string s1 = std::string((const char*) decoded_bytes, nbytes);
  //printf("processedDataLen%d decodedSize%d\n", processedDataLen, s1.size());
  return s1;
}

void gr_free_rs(void *p){
  struct rs *rs = (struct rs *)p;

  free(rs->alpha_to);
  free(rs->index_of);
  free(rs->genpoly);
  free(rs);
}

// apurv++ ends //

