#include "LDPC.h"
#include "rs.h"
#include <assert.h>
#include <bitset>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h> 
#include <iostream>
#include <vector>
#include <map>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdexcept>
#include <stdio.h>
#include <unistd.h>

using namespace std;

// yichao: global
// #define PRINT_DETAIL
#define CONSERVATIVE_HEADER
// 1. CSI prediction
#define CSI_ORACLE 0
#define CSI_PREVIOUS 1
#define CSI_EWMA 2
#define CSI_HW 3
// 2. header mapping mode
#define HEADER_OUR 0
#define HEADER_RAN 1
#define HEADER_STD 2
#define HEADER_OUR2 3	// put header redundancy behind payload data
#define HEADER_SWITCH 4	// switch symbols between FEC group
// 3. data mapping mode
#define DATA_OUR 0
#define DATA_RAN 1
#define DATA_STD 2
#define DATA_OUR3 3		// mix header and payload data
#define DATA_SWITCH 4	// switch symbols between FEC group
// 4. LDPC hint
#define LDPC_HINT 0
#define NO_LDPC_HINT 1
// 5. general hint
#define GEN_HINT 0
#define NO_GEN_HINT 1
// 6. MAC FEC
#define LDPC_MAC_FEC 0
#define RS_MAC_FEC 1
#define NO_MAC_FEC 2
// ?. coding
#define CODE_LDPC 0
#define CODE_RS 1
#define CODE_OPTIMAL 2
// ?. modulation
#define MOD_BPSK 0
#define MOD_QPSK 1
#define MOD_QAM16 2
#define MOD_QAM64 3
// ?. auto rate
#define FIX_RATE 0
#define AUTO_RATE_SIMULATION 1
#define AUTO_RATE_CHEBYSHEV 2


static const int MAX_PKT_NUM = 500;
// static const int MAX_PKT_NUM = 50;
static const int PKT_SKIP = 0;
static const char *TRACE_PATH = "/v/filer4b/v27q002/ut-wireless/yichao/11n/processed_traces/MonitorExp1/";

static const int HEADER_SIZE = 24 * 8;		// bits

short MODE_CSI_PREDICTION;
short MODE_HEADER_MAPPING;
short MODE_DATA_MAPPING;
short MODE_LDPC_HINT;
short MODE_GEN_HINT;
short MODE_MAC_FEC;
short MODE_PHY_FEC = CODE_RS;
short MODE_MODULATION = MOD_BPSK;
short MODE_RATE = FIX_RATE;

float CSI_SHIFT = 25.4820+2.5;
float ALPHA = 0.7;
short RATE = 12;
short FEC_RATE = 1;
char *TRACE_FILE;		//"l1.dat_matrix_dB"
int CSI_THRESHOLD = -20;
float BER_THRESHOLD = 0.1;	// 0.001
float DR_THRESHOLD = 0.95;
short AUTO_RATE_SIM_CNT = 1;
short PREVIOUS_CNT = 1;
// switch
bool *SW_MAP;		// which symbols are going to be swapped
short SW_N = 10;	// # of symbols per unit	
short MAX_ITR = 5;
// Holt-Winters
float HW_ALPHA = 0.2;
float HW_BETA = 0.1;
float *hw_a_i_1;
float *hw_b_i_1;
int line = 0;
// end yichao: global



#ifdef HAVE_IO_H
#include <io.h>
#endif
#ifdef O_BINARY
#define OUR_O_BINARY O_BINARY
#else
#define OUR_O_BINARY 0
#endif

#ifdef O_LARGEFILE
#define OUR_O_LARGEFILE O_LARGEFILE
#else
#define OUR_O_LARGEFILE 0
#endif

#define RS_MAC_PARITY 1
//#define LDPC_MAC_PARITY 1
#define PAD_EXTRA 1
//#define PERFORM_MAC_FEC 1		-- for RS, the flag is in rs.h
//#define DEBUG_MAC_FEC 1 
//#define DEBUG1 1

char *ldpc_filename;				// for phy layer
char *mac_ldpc_filename;			// for mac layer
bool mac_fec_flag;
float *g_csi;

static const double rel_error= 1E-12;        //calculate 12 significant figures

/* PHY FEC related function declarations */
std::string
	getBitString_LDPC(short *data, int num_symbols, const int bps);

std::string
	gr_FEC(std::string buf, int n, int k, int bps, bool encode, std::string parity_bits_str, 
#ifdef RS_MAC_OPT_VERSION
	std::map<int, split> fec_split_map,
	std::vector<int> sorted_orig_pos,
#endif
	std::vector<int> orig_pos_indices, 
	int mac_fec_N, int mac_fec_K, bool useLDPC, int dataLen, 
	int &total_fec_groups, int &correct_fec_groups, 
	std::string test_orig_string, std::string test_parity_bits_str
#ifdef PARTIAL_TPUT
	, int &num_corr_bits, int *final_pos_map,  int subarriers, float *csi, int hint_flag	
#endif
	);

int
	getSymSizeFromN_Bits(int n);

int
	getSymSizeFromN_Syms(int n);

float performSplitFEC(int d1, int r1, int &max_bad, int &max_r2_b, int &max_good, int &max_r2_g,  
	float *eff_delivery_rate, int *interleaved_bit_pos,
	std::vector<int>& sorted_interleaved_indices, 
	int parity_mapped_already, int subcarriers, int bps,
	float *csi, std::map<float, float> snr2ser_map,
	int fec_group_index, int fec_N, int fec_K, int *interleaved_parity_pos, 
	float *fec_delivery_rates_arr, int n_data_bits, int num_bits_out, float curr_utility, float &hdr_delivery_rate, int hint_flag);
void
	initializeMACLayerFEC_OPT(int num_bits_before_interleaving,
	int num_bits_after_interleaving,
	int *interleaved_bit_pos,
	float *csi,
	std::map<int, split>& fec_split_map,
	std::vector<int>& sorted_orig_pos,
	std::map<float, float> &snr2ser_map,
	const int total_subcarriers,
	int fec_N, int fec_K, int bps, 
	int &parity_added, bool encode, int hint_flag);
std::string
	gr_LDPC(std::string buf, int n, int k, int bps, bool encode, std::string parity_bits_str, std::vector<int> orig_pos_indices,
	int mac_fec_N, int mac_fec_K, bool useMAC_LDPC);
void getSymbols(std::string bits, short *symbols);
void getParityForFECGroup(int fec_group_index, std::string parity_bits_str,
	std::vector<int> orig_pos_indices, int fec_N, int fec_K,
	std::vector<int>& fec_parity_indices, short *paritySymbols);
void 
	injectErrors(short *msg, int msg_length, std::vector<int> indices_protected);

std::string
	injectErrorsProbabilistically(std::string msg, float *csi, int subcarriers, int bps);

void
	encodeMAC_FEC_with_RS(std::string mac_bits, int mac_fec_N, int mac_fec_K,
	int symSize, std::string &parity_bits_str, 
	bool pad_at_byte_boundary);

void
	encodeMAC_FEC_with_RS_split(std::string data_bits,
	split _split,
	std::string &parity_bits_str);


std::string
	interleave_mapping_best(std::string buf, int subcarriers, int bps, int fec_N, int fec_K, int scheme, int& n_mac_parity_added, int hint_flag);
std::string
	deinterleave_mapping_best(std::string buf, int subcarriers, int bps, int fec_N, int fec_K, int scheme, int expectedPayloadLen,
	std::string& parity_bits_str, int mac_parity_bits, 
	std::map<int, split>& fec_split_map, std::vector<int>& sorted_orig_pos, 
	std::string test_buf, std::string& test_parity_bits_str, int hint_flag
#ifdef PARTIAL_TPUT
	, int *final_pos_map
#endif
	);

/* mac fec related function declarations */
void
	encodeMACLayer_FEC( std::string interleaved_str,
	std::string orig_str,
	std::string &parity_bits_str,
	int *interleaved_bit_pos,
	int *orig_bit_pos,
	int total_subcarriers, int fec_N,
	int fec_K, int bps, int &_parity_added, int total_bits, 
	int num_hdr_symbols, int mac_fec_N, int mac_fec_K, int hint_flag);

void
	decodeMACLayer_FEC_OPT(std::string &deinterleaved_str,
	std::string parity_bits_str,
	int *interleaved_bit_pos,
	int *orig_bit_pos,
	int total_subcarriers, int fec_N,
	int fec_K, int bps, int num_bits_before_deinterleaving, 
	std::map<int, split>& fec_split_map, 
	std::vector<int>& sorted_orig_pos,
	int num_hdr_symbols, 
	int mac_fec_N, int mac_fec_K, int hint_flag);

void
	initializeMACLayerFEC(int num_bits_before_interleaving,
	int num_bits_after_interleaving,
	int *interleaved_bit_pos,
	float *csi,
	std::vector<int> &sel_indices,
	std::map<float, float> &snr2ser_map,
	const int total_subcarriers,
	int fec_N, int fec_K, int bps, int num_hdr_symbols, 
	int mac_fec_N, int mac_fec_K);
void
	selectFECGroupsToProtect(std::vector<int> &sel_fec_groups, int total_bits, 
	float *fec_delivery_rates_arr, int fec_K, int fec_N, 
	int *interleaved_bit_pos, float *eff_delivery_rate, 
	float &hdr_delivery_rate, int hint_flag);
float 
	getFECGroupDeliveryRate(float *eff_delivery_rate, int nbits, int fec_N, int fec_K);

float getFECSymbolDeliveryRateBySlope(int fec_N, float avg_ber);

float getFECSymbolDeliveryRateByRoot(float del_rate_bits);

float 
	calculateDeliveryRateIfWorstBitsProtected(int fec_group_index, int *sorted_bits, int parity_mapped, 
	float *eff_delivery_rate, int *interleaved_bit_pos,
	int K_WORST_BITS,  
	float *csi, std::map<float, float> snr2ser_map,
	int fec_N, int fec_K, int total_subcarriers, int bps, 
	float *new_eff_delivery_rate, int mac_fec_N, int mac_fec_K,
	int n_databits);
void 
	protectWorstBits(int *bits_to_protect, float *eff_delivery_rate, float *new_eff_delivery_rate, int mac_fec_K);

void 
	initializeEffectiveDeliveryRate(float *eff_delivery_rate, float *csi, 
	int total_bits,
	int *interleaved_bit_pos, std::map<float, float> &snr2ser_map, 
	int bps, int total_subcarriers, int num_hdr_symbols);
void 
	sortBitsByEffectiveDeliveryRate(float *sorted_eff_delivery_rates_arr, int *sorted_bit_indices,
	float *eff_delivery_rate, int *interleaved_bit_pos,
	int fec_group_index, int ndatabits, int fec_N);

float getSymbolErrorRate(float *csi, int sc_index, std::map<float, float> snr2ser_map);

float getSymbolErrorRate2(float *csi, int sc_index, int bps);

double snr2ber(int mod, double snr);

void sortFECGroupsByDeliveryRate(float *delivery_rates, int *indices, int n);
int fm(float arr[], int b, int n, bool ascending);
void isort(float arr[], int indices[], int n, bool ascending);
void setSNR_to_SER_Mapping(std::map<float, float> &snr2ser_map);
void readCSISubcarriers(float *csi, const int total_subcarriers);
void sortSubcarriersByCSI(int *subcarrier_indices, float *csi, int total_subcarriers);
std::string getBitString(const unsigned char *data, int len);
std::string getMsg(std::string str, unsigned char *data);

unsigned int
	gr_update_crc32(unsigned int crc, const std::string s);

unsigned int
	gr_crc32(const unsigned char *buf, size_t len);

unsigned int
	gr_crc32(const std::string s);

unsigned int
	gr_update_crc32(unsigned int crc, const unsigned char *data, size_t len);

void initInterleavedParityPos(int *interleaved_parity_pos, int bps, int subcarriers);

void run_deinterleaver();

float getHeaderDeliveryRate(float *eff_delivery_rate, float fec_group_delivery_rate, int *interleaved_pos, int hdr_size);
float getUtility(float *fec_delivery_rates, float *eff_delivery_rate, int *interleaved_pos, int n_data_bits, int num_bits_out, int fec_K, int fec_N, float hdr_delivery_rate, int hint_flag);

float getAvgDelRateOfFECGroup(int fec_group_index, float *eff_delivery_rate, int *interleaved_pos, int fec_N, int num_bits);

int dumpActualFECGroupDeliveryRates(std::string fec_encoded_str, std::string fec_decoded_str, int fec_K, int symSize);

int calcBERValuesOnSubcarriers(std::string encoded_str, std::string decoded_str, int modulation, int subcarriers, int *bit_errors, int *total_bits);



double snr2ber3(int mod, double snr)
{
	double ber, ser;
	int m;

	snr = pow(10, snr / 10);	// dB to power
	snr *= (80 / 52.0);

	if(mod == MOD_BPSK) {
		m = 2;
		ber = erfc(pow(snr, 0.5) ) / 2;
	}
	else if(mod == MOD_QPSK) {
		m = 4;
		ber = erfc(pow(snr / 2, 0.5) ) / 2;
	}
	else if(mod == MOD_QAM16) {
		m = 16;
		ber = erfc(pow(snr / 10, 0.5) ) * 3 / 8;
	}
	else if(mod == MOD_QAM64) {
		m = 64;
		ber = erfc(pow(snr / 42, 0.5) ) * 7 / 24;
	}

	return ber;
}

float getSymbolErrorRate3(float *csi, int sc_index)
{
	return snr2ber3(MODE_MODULATION, csi[sc_index]);
}

void set_parameter(int& fec_K, int& bits_per_symbol) {
	RATE = 13;
	// fec_K = 127;
	// bits_per_symbol = 1;
}

void simulate_a_pkt(int fec_N, int fec_K, std::string combined_payload, int hdr_size, int num_subcarriers, int bits_per_symbol, 
	std::string parity_bits_str, 
#ifdef RS_MAC_OPT_VERSION
	std::map<int, split> &fec_split_map_sender,
	std::vector<int> &sorted_orig_pos_sender,
	std::map<int, split> &fec_split_map_rcvr,
	std::vector<int> &sorted_orig_pos_rcvr,
#endif
	std::vector<int> &orig_pos_indices,
	int mac_fec_N, int mac_fec_K, int scheme, bool useLDPC, int& total_bits_delivered, double& used_time, bool sampling, int hint_flag,
	int &mac_parity_bits);
double auto_rate2(int fec_N, int fec_K, std::string combined_payload, int hdr_size, int num_subcarriers, int bits_per_symbol, 
	std::string parity_bits_str, 
#ifdef RS_MAC_OPT_VERSION
	std::map<int, split> &fec_split_map_sender,
	std::vector<int> &sorted_orig_pos_sender,
	std::map<int, split> &fec_split_map_rcvr,
	std::vector<int> &sorted_orig_pos_rcvr,
#endif
	std::vector<int> &orig_pos_indices,
	int mac_fec_N, int mac_fec_K, int &scheme, bool useLDPC, /*int& total_bits_delivered, double& used_time, bool sampling,*/ int hint_flag,
	int &mac_parity_bits);


void
	log_data(std::string buf)
{
	int fd;
	FILE *d_fp;
	const char *d_filename = "rx_bits.dat";
	if ((fd = open (d_filename,
		O_WRONLY|O_CREAT|O_TRUNC|OUR_O_LARGEFILE|OUR_O_BINARY,
	0777)) < 0){
		perror(d_filename);
	}else {
		if((d_fp = fdopen (fd, true ? "wb" : "w")) == NULL) {
			fprintf(stderr, "rx-bits file cannot be opened\n");
			close(fd);
		}
		fprintf(stderr, "rx bits opened!\n");
	}

	int c = ftell(d_fp);
	c = fwrite_unlocked((unsigned char*) buf.data(), sizeof(unsigned char), buf.size(), d_fp);
	// printf("logged %d items\n", c); fflush(stdout);
	c = ftell(d_fp);
}

/* just fiddle with these parameters to get the entire chain running */
void
	initializeParameters(std::string& payload_str, std::string& hdr_str, int& num_subcarriers, int& bits_per_symbol, int& fec_N, 
	int& fec_K, int &hdr_size, int &mac_fec_N, int &mac_fec_K, int &mapping_scheme)
{
	int payload_size = 996;			// bytes			i	616+24, 348+24, 984+24
	payload_str = std::string(payload_size, 'l');

	for(int i = 0; i < payload_size; i++)
		payload_str[i] = (rand() % 26) + 'a';
	//payload_str[i] = 'm';

	num_subcarriers = 32; //192;
	bits_per_symbol = 1;
	fec_N = 255;
	fec_K = 127; 
	ldpc_filename = (char*) "/v/filer4b/v27q002/ut-wireless/Apurv/mac-fec/96.1A3.838.txt";
	mac_ldpc_filename = (char*) "/v/filer4b/v27q002/ut-wireless/Apurv/mac-fec/pchk.n16.k8";
	hdr_size = 20;		// will take an extra 4 bytes for the CRC
	mac_fec_N = 15;
	mac_fec_K = 8;
	hdr_str = std::string(hdr_size, 'a');
	mapping_scheme = 0;
	mac_fec_flag = 1;
}



int main(int argc, char ** argv)
{
	int hdr_size; 				// hdr not protected by the interleaver and LDPC in gnuradio,
						// in simulation results, though it should be! The idea of this code
						// is to verify the code that will run on the gnuradio
	int hint_flag = 1;
	int fail = 0;
	int correct = 0;
	bool isHdrCorrect = false;

	static std::vector<int> orig_pos_indices;
	
#ifdef RS_MAC_OPT_VERSION
	static std::map<int, split> fec_split_map_sender;
	static std::vector<int> sorted_orig_pos_sender;
	static std::map<int, split> fec_split_map_rcvr;
	static std::vector<int> sorted_orig_pos_rcvr;
#endif
	
	
	
		
	std::string payload_str;
	std::string hdr_str;
	int num_subcarriers, bits_per_symbol, fec_N, fec_K, mac_fec_N, mac_fec_K;
	int scheme;
	int t_fec_groups = 0;
	int c_fec_groups = 0;

	initializeParameters(payload_str, hdr_str, num_subcarriers, bits_per_symbol, fec_N, fec_K, hdr_size, mac_fec_N, mac_fec_K, scheme);

	bool useLDPC = false;
	
	
	
	
	
	// yichao: variables
	if(argc < 16) {
		printf("wrong para %d\n", argc);
		return 0;
	}

	// 1. CSI prediction
	if(strcmp(argv[1], "oracle") == 0) {
		MODE_CSI_PREDICTION = CSI_ORACLE;
		// printf("> oracle, ");
	}
	else if(strcmp(argv[1], "previous") == 0) {
		MODE_CSI_PREDICTION = CSI_PREVIOUS;
		// printf("> previous, ");
	}
	else if(strcmp(argv[1], "ewma") == 0) {
		MODE_CSI_PREDICTION = CSI_EWMA;
		// printf("> ewma, ");
	}
	else if(strcmp(argv[1], "hw") == 0) {
		MODE_CSI_PREDICTION = CSI_HW;
	}
	else {
		printf("no such parameter\n");
		assert(false);
	}
	// 2. header mapping mode
	if(strcmp(argv[2], "our") == 0) {
		MODE_HEADER_MAPPING = HEADER_OUR;
		// printf("header: our, ");
	}
	else if(strcmp(argv[2], "random") == 0) {
		MODE_HEADER_MAPPING = HEADER_RAN;
		// printf("header: random, ");
	}
	else if(strcmp(argv[2], "standard") == 0) {
		MODE_HEADER_MAPPING = HEADER_STD;
		// printf("header: standard, ");
	}
	else if(strcmp(argv[2], "our2") == 0) {
		MODE_HEADER_MAPPING = HEADER_OUR2;
		// printf("header: our2, ");
	}
	else if(strcmp(argv[2], "switch") == 0) {
		MODE_HEADER_MAPPING = HEADER_SWITCH;
	}
	else {
		printf("no such parameter\n");
		assert(false);
	}
	// 3. data mapping mode
	if(strcmp(argv[3], "our") == 0) {
		MODE_DATA_MAPPING = DATA_OUR;
		// printf("data: our, ");
	}
	else if(strcmp(argv[3], "random") == 0) {
		MODE_DATA_MAPPING = DATA_RAN;
		// printf("data: random, ");
	}
	else if(strcmp(argv[3], "standard") == 0) {
		MODE_DATA_MAPPING = DATA_STD;
		// printf("data: standard, ");
	}
	else if(strcmp(argv[3], "our3") == 0) {
		MODE_DATA_MAPPING = DATA_OUR3;
		// printf("data: our2, ");
	}
	else if(strcmp(argv[3], "switch") == 0) {
		MODE_DATA_MAPPING = DATA_SWITCH;
	}
	else {
		printf("no such parameter\n");
		assert(false);
	}
	//////////////
	if(MODE_HEADER_MAPPING == HEADER_RAN && MODE_DATA_MAPPING == DATA_RAN) {
		scheme = 0;
	}
	else if(MODE_HEADER_MAPPING == HEADER_OUR && MODE_DATA_MAPPING == DATA_RAN) {
		scheme = 5;
	}
	else if(MODE_HEADER_MAPPING == HEADER_RAN && MODE_DATA_MAPPING == DATA_OUR) {
		scheme = 4;
	}
	else if(MODE_HEADER_MAPPING == HEADER_OUR && MODE_DATA_MAPPING == DATA_OUR) {
		scheme = 1;
	}
	else if(MODE_DATA_MAPPING == DATA_OUR3) {
		scheme = 6;
	}
	else if(MODE_HEADER_MAPPING == HEADER_OUR2 && MODE_DATA_MAPPING == DATA_OUR) {
		scheme = 4;
	}
	else if(MODE_HEADER_MAPPING == HEADER_SWITCH && MODE_DATA_MAPPING == DATA_SWITCH) {
		scheme = 5;
	}
	else {
		printf("error\n");
		assert(false);
	}
	//////////////
	// 4. LDPC hint
	if(strcmp(argv[4], "ldpc-hint") == 0) {
		MODE_LDPC_HINT = LDPC_HINT;
		// printf("with LDPC hint, ");
	}
	else if(strcmp(argv[4], "no-ldpc-hint") == 0) {
		MODE_LDPC_HINT = NO_LDPC_HINT;
		// printf("without LDPC hint, ");
	}
	else {
		printf("no such parameter\n");
		assert(false);
	}
	// 5. general hint
	if(strcmp(argv[5], "gen-hint") == 0) {
		MODE_GEN_HINT = GEN_HINT;
		// hint_flag = 1;
		// printf("with general hint, ");
	}
	else if(strcmp(argv[5], "no-gen-hint") == 0) {
		MODE_GEN_HINT = NO_GEN_HINT;
		// hint_flag = 0;
		// printf("without general hint, ");
	}
	else {
		printf("no such parameter\n");
		assert(false);
	}
	// 6. MAC FEC
	if(strcmp(argv[6], "ldpc-mac-fec") == 0) {
		MODE_MAC_FEC = LDPC_MAC_FEC;
		mac_fec_flag = 0;
		// printf("with LDPC MAC FEC, ");
	}
	else if(strcmp(argv[6], "rs-mac-fec") == 0) {
		MODE_MAC_FEC = RS_MAC_FEC;
		mac_fec_flag = 1;
		// printf("with RS MAC FEC, ");
	}
	else if(strcmp(argv[6], "no-mac-fec") == 0) {
		MODE_MAC_FEC = NO_MAC_FEC;
		mac_fec_flag = 0;
		// printf("without MAC_FEC, ");
	}
	else {
		printf("no such parameter\n");
		assert(false);
	}
	// 7. coding
	if(strcmp(argv[7], "ldpc") == 0) {
		MODE_PHY_FEC = CODE_LDPC;
		// useLDPC = true;
		// fec_N = 96;
		// fec_K = 32;
		// printf("LDPC, ");
	}
	else if(strcmp(argv[7], "rs") == 0) {
		MODE_PHY_FEC = CODE_RS;
		useLDPC = false;
		// printf("RS, ");
	}
	else if(strcmp(argv[7], "optimal") == 0) {
		MODE_PHY_FEC = CODE_OPTIMAL;
		useLDPC = false;
		// printf("optimal, ");
	}
	else {
		printf("no such parameter\n");
		assert(false);
	}
	// 8. modulation
	if(strcmp(argv[8], "BPSK") == 0) {
		MODE_MODULATION = MOD_BPSK;
		// printf("BPSK, ");
	}
	else if(strcmp(argv[8], "QPSK") == 0) {
		MODE_MODULATION = MOD_QPSK;
		// printf("QPSK, ");
	}
	else if(strcmp(argv[8], "QAM16") == 0) {
		MODE_MODULATION = MOD_QAM16;
		// printf("QAM16, ");
	}
	else if(strcmp(argv[8], "QAM64") == 0) {
		MODE_MODULATION = MOD_QAM64;
		// printf("QAM64, ");
	}
	else {
		printf("no such parameter\n");
		assert(false);
	}
	// 9. csi shift
	CSI_SHIFT = atof(argv[9]);
	// printf("CSI shift = %d, ", CSI_SHIFT);
	// 10. sending rate
	RATE = atoi(argv[10]);
	// printf("sending rate = %d, ", RATE);
	// 11. FEC rate
	FEC_RATE = atoi(argv[11]);
	// 12. trace file name
	TRACE_FILE = argv[12];		//"l1.dat_matrix_dB"
	// printf("trace file = %s", TRACE_FILE);
	// 13. alpha of EWMA
	ALPHA = atof(argv[13]);
	// printf(", EWMA alpha = %f", ALPHA);
	// 14. auto rate	
	if(strcmp(argv[14], "fixed-rate") == 0) {
		MODE_RATE = FIX_RATE;
	}
	else if(strcmp(argv[14], "auto-rate-simulation") == 0) {
		MODE_RATE = AUTO_RATE_SIMULATION;
	}
	else if(strcmp(argv[14], "auto-rate-chebyshev") == 0) {
		MODE_RATE = AUTO_RATE_CHEBYSHEV;
	}
	else {
		printf("no such parameter\n");
		assert(false);
	}
	// 15. count for auto rate simulation
	AUTO_RATE_SIM_CNT = atoi(argv[15]);
	// 16. count for using previous "x" CSI for sender
	PREVIOUS_CNT = atoi(argv[16]);
	if(PREVIOUS_CNT == 0) {
		PREVIOUS_CNT = 1;
	}
	SW_MAP = (bool *)malloc(sizeof(bool) * MAX_ITR);
	for(int i = 0; i < MAX_ITR; i ++) {
		SW_MAP[i] = false;
	}

	set_parameter(fec_K, bits_per_symbol);
	
	double used_time = 0;
	
	
	

	// for injecting errors 
	int nn = 32;
	g_csi = (float*) malloc(sizeof(float) * nn);
	memset(g_csi, 0, sizeof(float) * nn);
	// readCSISubcarriers(g_csi, nn);

	int total_fec_groups = 0;
	int correct_fec_groups = 0;
	int hdr_correct = 0;
	int total_bits_delivered = 0;

#ifdef PARTIAL_TPUT
	if(1)
	{
		run_deinterleaver();
		exit(0);
	}
#else
	int loop_index = 0;
	int start = 0;

	int *bit_errors = (int*)malloc(sizeof(int) * 192);
	memset(bit_errors, 0, sizeof(int) * 192);

	int *total_bits = (int*) malloc(sizeof(int) * 192);
	memset(total_bits, 0, sizeof(int) * 192);


	// yichao: read csi
	int fd;
	FILE *d_fp;
	if ((fd = open(TRACE_FILE, O_RDONLY|OUR_O_LARGEFILE, 0664)) < 0){
		perror(TRACE_FILE);
	}else {
		if((d_fp = fdopen (fd, true ? "r" : "rt")) == NULL) {
			fprintf(stderr, "CSI file cannot be opened\n");
			close(fd);
			assert(false);
		}
	}
	// end yichao
	
	// do crc on the header //
	unsigned int hdr_crc = gr_crc32(hdr_str);
	char hdr_crc_str1[5];
	snprintf(hdr_crc_str1, sizeof(hdr_crc_str1), "%ld", hdr_crc);
	std::string hdr_crc_str(hdr_crc_str1);
	hdr_str.append(hdr_crc_str);
	// crc complete on the header //
	
	

	// for(loop_index = start; loop_index < 2; loop_index++)
	while(line < 100)
	{
		line ++;
	
		/**
		* 1. get subcarrier CSI
		**/
		// readCSISubcarriers2(curr_csi, num_subcarriers, line);
		char aline[1700];
		fgets(aline, 1700, d_fp);
		float *curr_csi = (float*) malloc(sizeof(float) * num_subcarriers);
		sscanf(aline, "%f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f", 
			&curr_csi[0], &curr_csi[1], &curr_csi[2], &curr_csi[3], &curr_csi[4], &curr_csi[5], &curr_csi[6], 
			&curr_csi[7], &curr_csi[8], &curr_csi[9], &curr_csi[10], &curr_csi[11], &curr_csi[12], &curr_csi[13], 
			&curr_csi[14], &curr_csi[15], &curr_csi[16], &curr_csi[17], &curr_csi[18], &curr_csi[19], &curr_csi[20], 
			&curr_csi[21], &curr_csi[22], &curr_csi[23], &curr_csi[24], &curr_csi[25], &curr_csi[26], &curr_csi[27],
			&curr_csi[28], &curr_csi[29]);
		curr_csi[30] = curr_csi[28];
		curr_csi[31] = curr_csi[29];
		
		for(int i = 0; i < 32; i++) {
			curr_csi[i] = curr_csi[i] - CSI_SHIFT;                     //TODO: for now!
			// printf("csi: %f, i: %d\n", curr_csi[i], i); fflush(stdout);
			g_csi[i] = getSymbolErrorRate3(curr_csi, i);
			// printf("%f, ", g_csi[i]); fflush(stdout);	
		}
		free(curr_csi);
		// exit(0);
		
	
		
		

		const unsigned char *data = (const unsigned char*) payload_str.data();
		size_t len = payload_str.size();		// bytes

#ifdef DEBUG1
		printf("Original payload:: \n");
		for(int i =0; i < len; i++)
			printf("%d ", data[i]);
		printf("\n");
#endif
		srand(loop_index);
		int mac_parity_bits = 0;

		// record the parity and indices that are protected by that parity, might be needed after the phy-fec-decoding //
		std::string parity_bits_str;

		
// #ifdef DEBUG1
// 		printf(" <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< INPUT, lt: %d >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> \n", len);
// 		for(int i = 0; i < len; i++)
// 			printf("%d ", data[i]);
// 		printf("\n");    
// #endif
		std::string combined_payload = hdr_str + payload_str;	

		// printf("combined payload in bits.. %d\n", combined_payload.size()*8);
// #ifdef DEBUG10
// 		std::string encoded_bits = getBitString((const unsigned char*) combined_payload.data(), combined_payload.size());	
// 		for(int i= 0; i < encoded_bits.size(); i++)
// 			printf("%c", encoded_bits[i]);
// 		printf("\n");
// #endif


		
		

		if(MODE_HEADER_MAPPING == HEADER_SWITCH) {
			
			auto_rate2(fec_N, fec_K, combined_payload, hdr_size, num_subcarriers, bits_per_symbol, parity_bits_str, 
#ifdef RS_MAC_OPT_VERSION
				fec_split_map_sender,
				sorted_orig_pos_sender,
				fec_split_map_rcvr,
				sorted_orig_pos_rcvr,
#endif
				orig_pos_indices,
				mac_fec_N, mac_fec_K, scheme, useLDPC, hint_flag, mac_parity_bits);
			
		}
		

		simulate_a_pkt(fec_N, fec_K, combined_payload, hdr_size, num_subcarriers, bits_per_symbol, parity_bits_str, 
#ifdef RS_MAC_OPT_VERSION
			fec_split_map_sender,
			sorted_orig_pos_sender,
			fec_split_map_rcvr,
			sorted_orig_pos_rcvr,
#endif
			orig_pos_indices,
			mac_fec_N, mac_fec_K, scheme, useLDPC, total_bits_delivered, used_time, false, hint_flag, mac_parity_bits);
		
		
		
	}

	// printf("\n\nTotal Pkts: %d Failed Pkts: %d, Correct Pkts: %d, Total FEC Groups: %d, Correct FEC Groups: %d, Hdr Correct: %d, total_bits_delivered: %d\n", loop_index-start, fail, correct, total_fec_groups, correct_fec_groups, hdr_correct, total_bits_delivered);
	
	
	// printf(", %f(%d)", total_bits_delivered / used_time, total_bits_delivered);
	printf(", %f", total_bits_delivered / used_time);
	free(g_csi);
	free(hw_a_i_1);
	free(hw_b_i_1);
	free(SW_MAP);
	/*
	printf("dump BER values of different subcarriers -- \n");
	for(int i = 0; i < 192; i++)
	printf("sc_index; %d, BER: %f\n", i, ((float)bit_errors[i])/total_bits[i]);
	*/
	free(bit_errors);
	free(total_bits);

	#endif
	return 0;

}


std::string
	gr_FEC(std::string buf, int n, int k, int bps, bool encode, std::string parity_bits_str, 
#ifdef RS_MAC_OPT_VERSION
	std::map<int, split> fec_split_map,
	std::vector<int> sorted_orig_indices,
#endif
	std::vector<int> orig_pos_indices, 
	int mac_fec_N, int mac_fec_K, bool useLDPC, int dataLen, 
	int &total_fec_groups, int &correct_fec_groups, 
	std::string test_orig_str, std::string test_parity_bits_str
#ifdef PARTIAL_TPUT
	, int &num_corr_bits, int *final_pos_map, int subcarriers, float *csi, int hint_flag
#endif
	)
{
	if(useLDPC)
		assert(false);
	else
	{
		if(encode)
		{
			std::string buf_bits = getBitString((const unsigned char*) buf.data(), buf.size());
			return gr_encode_rs_fec(buf_bits, n, k);
		}
		else
		{

			return gr_decode_rs_fec(buf, n, k, dataLen, parity_bits_str, 
#ifdef RS_MAC_OPT_VERSION
				fec_split_map,
				sorted_orig_indices,
#else
				orig_pos_indices, 
#endif
				mac_fec_N, mac_fec_K, 
				total_fec_groups, correct_fec_groups, mac_fec_flag, test_orig_str, test_parity_bits_str
#ifdef PARTIAL_TPUT
				, num_corr_bits, final_pos_map,  bps, subcarriers, csi, hint_flag
#endif
				);
		}
	} 
	assert(false);
}

// from yi-chao or snr2ber computation //
float getSymbolErrorRate2(float *csi, int sc_index, int mod)
{
	return csi[sc_index];
	/*
	float ber = snr2ber(mod, csi[sc_index]);
	//printf("csi: %f, sc_index: %d, ber: %f\n", csi[sc_index], sc_index, ber);
	return ber; */
	}

	double snr2ber(int bps, double snr)
	{
		double ber, ser;
		int m;

		snr = pow(10, snr / 10);    // dB to power
		snr *= (float)80/52;

		if(bps == 1) {
			m = 2;
			ber = erfc(pow(snr, 0.5) ) / 2;
		}
		else if(bps == 2) {
			m = 4;
			ber = erfc(pow(snr / 2, 0.5) ) / 2;
		}
		else if(bps == 3) {
			m = 8;
			ber = erfc(pow((3 * snr) / 7, 0.5) ) * 4;
		}
		else if(bps == 4) {
			m = 16;
			ber = erfc(pow(snr / 10, 0.5) ) * 3 / 8;
		}
		else if(bps == 6) {
			m = 64;
			ber = erfc(pow(snr / 42, 0.5) ) * 7 / 24;
		}
		if(ber < 0.0 || ber > 1.0)
		{
			printf("ber: %f snr: %f\n", ber, snr); 
			assert(false);
		}
		return ber;
	}

	double erf(double x)
	{
		static const double two_sqrtpi=  1.128379167095512574;        // 2/sqrt(pi)
		if (fabs(x) > 2.2) {
			return 1.0 - erfc(x);        //use continued fraction when fabs(x) > 2.2
		}
		double sum= x, term= x, xsqr= x*x;
		int j= 1;
		do {
			term*= xsqr/j;
			sum-= term/(2*j+1);
			++j;
			term*= xsqr/j;
			sum+= term/(2*j+1);
			++j;
		} while (fabs(term)/sum > rel_error);
		return two_sqrtpi*sum;
	}

	double erfc(double x)
	{
		static const double one_sqrtpi=  0.564189583547756287;        // 1/sqrt(pi)
		if (fabs(x) < 2.2) {
			return 1.0 - erf(x);        //use series when fabs(x) < 2.2
		}
		if (signbit(x)) {               //continued fraction only valid for x>0
			return 2.0 - erfc(-x);
		}
		double a=1, b=x;                //last two convergent numerators
		double c=x, d=x*x+0.5;          //last two convergent denominators
		double q1, q2= b/d;             //last two convergents (a/c and b/d)
		double n= 1.0, t;
		do {
			t= a*n+b*x;
			a= b;
			b= t;
			t= c*n+d*x;
			c= d;
			d= t;
			n+= 0.5;
			q1= q2;
			q2= b/d;
		} while (fabs(q1-q2)/q2 > rel_error);
		return one_sqrtpi*exp(-x*x)*q2;
	}

// crc related stuff //
	unsigned int
		gr_update_crc32(unsigned int crc, const std::string s)
	{
		return gr_update_crc32(crc, (const unsigned char *) s.data(), s.size());
	}

	unsigned int
		gr_crc32(const unsigned char *buf, size_t len)
	{
		return gr_update_crc32(0xffffffff, buf, len) ^ 0xffffffff;
	}

	unsigned int
		gr_crc32(const std::string s)
	{
		return gr_crc32((const unsigned char *) s.data(), s.size());
	}

// Automatically generated CRC function
// polynomial: 0x104C11DB7
	unsigned int
		gr_update_crc32(unsigned int crc, const unsigned char *data, size_t len)
	{
		static const unsigned int table[256] = {
			0x00000000U,0x04C11DB7U,0x09823B6EU,0x0D4326D9U,
				0x130476DCU,0x17C56B6BU,0x1A864DB2U,0x1E475005U,
				0x2608EDB8U,0x22C9F00FU,0x2F8AD6D6U,0x2B4BCB61U,
				0x350C9B64U,0x31CD86D3U,0x3C8EA00AU,0x384FBDBDU,
				0x4C11DB70U,0x48D0C6C7U,0x4593E01EU,0x4152FDA9U,
				0x5F15ADACU,0x5BD4B01BU,0x569796C2U,0x52568B75U,
				0x6A1936C8U,0x6ED82B7FU,0x639B0DA6U,0x675A1011U,
				0x791D4014U,0x7DDC5DA3U,0x709F7B7AU,0x745E66CDU,
				0x9823B6E0U,0x9CE2AB57U,0x91A18D8EU,0x95609039U,
				0x8B27C03CU,0x8FE6DD8BU,0x82A5FB52U,0x8664E6E5U,
				0xBE2B5B58U,0xBAEA46EFU,0xB7A96036U,0xB3687D81U,
				0xAD2F2D84U,0xA9EE3033U,0xA4AD16EAU,0xA06C0B5DU,
				0xD4326D90U,0xD0F37027U,0xDDB056FEU,0xD9714B49U,
				0xC7361B4CU,0xC3F706FBU,0xCEB42022U,0xCA753D95U,
				0xF23A8028U,0xF6FB9D9FU,0xFBB8BB46U,0xFF79A6F1U,
				0xE13EF6F4U,0xE5FFEB43U,0xE8BCCD9AU,0xEC7DD02DU,
				0x34867077U,0x30476DC0U,0x3D044B19U,0x39C556AEU,
				0x278206ABU,0x23431B1CU,0x2E003DC5U,0x2AC12072U,
				0x128E9DCFU,0x164F8078U,0x1B0CA6A1U,0x1FCDBB16U,
				0x018AEB13U,0x054BF6A4U,0x0808D07DU,0x0CC9CDCAU,
				0x7897AB07U,0x7C56B6B0U,0x71159069U,0x75D48DDEU,
				0x6B93DDDBU,0x6F52C06CU,0x6211E6B5U,0x66D0FB02U,
				0x5E9F46BFU,0x5A5E5B08U,0x571D7DD1U,0x53DC6066U,
				0x4D9B3063U,0x495A2DD4U,0x44190B0DU,0x40D816BAU,
				0xACA5C697U,0xA864DB20U,0xA527FDF9U,0xA1E6E04EU,
				0xBFA1B04BU,0xBB60ADFCU,0xB6238B25U,0xB2E29692U,
				0x8AAD2B2FU,0x8E6C3698U,0x832F1041U,0x87EE0DF6U,
				0x99A95DF3U,0x9D684044U,0x902B669DU,0x94EA7B2AU,
				0xE0B41DE7U,0xE4750050U,0xE9362689U,0xEDF73B3EU,
				0xF3B06B3BU,0xF771768CU,0xFA325055U,0xFEF34DE2U,
				0xC6BCF05FU,0xC27DEDE8U,0xCF3ECB31U,0xCBFFD686U,
				0xD5B88683U,0xD1799B34U,0xDC3ABDEDU,0xD8FBA05AU,
				0x690CE0EEU,0x6DCDFD59U,0x608EDB80U,0x644FC637U,
				0x7A089632U,0x7EC98B85U,0x738AAD5CU,0x774BB0EBU,
				0x4F040D56U,0x4BC510E1U,0x46863638U,0x42472B8FU,
				0x5C007B8AU,0x58C1663DU,0x558240E4U,0x51435D53U,
				0x251D3B9EU,0x21DC2629U,0x2C9F00F0U,0x285E1D47U,
				0x36194D42U,0x32D850F5U,0x3F9B762CU,0x3B5A6B9BU,
				0x0315D626U,0x07D4CB91U,0x0A97ED48U,0x0E56F0FFU,
				0x1011A0FAU,0x14D0BD4DU,0x19939B94U,0x1D528623U,
				0xF12F560EU,0xF5EE4BB9U,0xF8AD6D60U,0xFC6C70D7U,
				0xE22B20D2U,0xE6EA3D65U,0xEBA91BBCU,0xEF68060BU,
				0xD727BBB6U,0xD3E6A601U,0xDEA580D8U,0xDA649D6FU,
				0xC423CD6AU,0xC0E2D0DDU,0xCDA1F604U,0xC960EBB3U,
				0xBD3E8D7EU,0xB9FF90C9U,0xB4BCB610U,0xB07DABA7U,
				0xAE3AFBA2U,0xAAFBE615U,0xA7B8C0CCU,0xA379DD7BU,
				0x9B3660C6U,0x9FF77D71U,0x92B45BA8U,0x9675461FU,
				0x8832161AU,0x8CF30BADU,0x81B02D74U,0x857130C3U,
				0x5D8A9099U,0x594B8D2EU,0x5408ABF7U,0x50C9B640U,
				0x4E8EE645U,0x4A4FFBF2U,0x470CDD2BU,0x43CDC09CU,
				0x7B827D21U,0x7F436096U,0x7200464FU,0x76C15BF8U,
				0x68860BFDU,0x6C47164AU,0x61043093U,0x65C52D24U,
				0x119B4BE9U,0x155A565EU,0x18197087U,0x1CD86D30U,
				0x029F3D35U,0x065E2082U,0x0B1D065BU,0x0FDC1BECU,
				0x3793A651U,0x3352BBE6U,0x3E119D3FU,0x3AD08088U,
				0x2497D08DU,0x2056CD3AU,0x2D15EBE3U,0x29D4F654U,
				0xC5A92679U,0xC1683BCEU,0xCC2B1D17U,0xC8EA00A0U,
				0xD6AD50A5U,0xD26C4D12U,0xDF2F6BCBU,0xDBEE767CU,
				0xE3A1CBC1U,0xE760D676U,0xEA23F0AFU,0xEEE2ED18U,
				0xF0A5BD1DU,0xF464A0AAU,0xF9278673U,0xFDE69BC4U,
				0x89B8FD09U,0x8D79E0BEU,0x803AC667U,0x84FBDBD0U,
				0x9ABC8BD5U,0x9E7D9662U,0x933EB0BBU,0x97FFAD0CU,
				0xAFB010B1U,0xAB710D06U,0xA6322BDFU,0xA2F33668U,
				0xBCB4666DU,0xB8757BDAU,0xB5365D03U,0xB1F740B4U,
			};

			while (len > 0)
			{
				crc = table[*data ^ ((crc >> 24) & 0xff)] ^ (crc << 8);
				data++;
				len--;
			}
			return crc;
		}

#ifdef PERFORM_MAC_FEC
// MAC FEC related functions //
void
	encodeMACLayer_FEC( std::string interleaved_str,
	std::string orig_str,
	std::string &parity_bits_str,
	int *interleaved_bit_pos,
	int *orig_bit_pos,
	int total_subcarriers, int fec_N,
	int fec_K, int bps, int &parity_added, int orig_len, 
	int num_hdr_symbols, int mac_fec_N, int mac_fec_K, int hint_flag)
{
// get the interleaved string //
	const unsigned char *data = (const unsigned char*) interleaved_str.data();
	unsigned int len = interleaved_str.size();
	assert(len % 8 == 0);                        // byte boundary

#ifdef DEBUG_MAC_FEC
	printf("data ::: \n");
	for(int i = 0; i < interleaved_str.size(); i++)
	{
		printf("%c", data[i]);
	}
	printf("\n");
#endif

	static bool initialized = false;
	static std::vector<int> sel_indices;
	std::map<float, float> snr2ser_map;                    // (snr, ser) map
#ifdef RS_MAC_OPT_VERSION
	static std::map<int, split> fec_split_map;
	static std::vector<int> sorted_orig_pos;
#endif

	int symSize = 1;      // for LDPC, its immaterial, just keep it at 1
#ifdef RS_MAC_PARITY
	symSize = getSymSizeFromN_Syms(fec_N);
	fec_N = fec_N * symSize;
	fec_K = fec_K * symSize;

	symSize = getSymSizeFromN_Syms(mac_fec_N);
	mac_fec_N = mac_fec_N * symSize;
	mac_fec_K = mac_fec_K * symSize;
#endif

	if(!initialized)
	{

		float *csi = (float*) malloc(sizeof(float) * total_subcarriers);
		memset(csi, 0, sizeof(float) * total_subcarriers);

		initialized = true;
		initializeMACLayerFEC_OPT(orig_len,
			interleaved_str.size(),
			interleaved_bit_pos,
			csi,
			fec_split_map,
			sorted_orig_pos,
			snr2ser_map,
			total_subcarriers, fec_N, fec_K, bps, 
			parity_added,
			true, hint_flag);	

// update the 'sorted_orig_pos' with the 'orig' indices rather than the interleaved ones //
//printf("encodeMACLayer_FEC, parity_added: %d\n", parity_added); 
		for(int i = 0; i < sorted_orig_pos.size(); i++)
		{
			int int_index = sorted_orig_pos[i];
			sorted_orig_pos[i] = orig_bit_pos[int_index];
//printf("%d ", int_index);
		}
		free(csi);
	}
	static int parity_offset1 = parity_added; 		// should not be used by Yi-Chao, this is coz CSI read just once!



	int num_fec_groups = ceil((float) orig_len/fec_N);		// phy layer fec
	int n_red_bits = num_fec_groups * (fec_N - fec_K);
	int n_data_bits = orig_len - n_red_bits;
//printf("\n\nnum_fec_groups: %d, n_red_bits: %d, n_data_bits: %d fec_split_map.size: %d\n", num_fec_groups, n_red_bits, n_data_bits, fec_split_map.size()); fflush(stdout);
	std::string mac_data_bits;
	std::map<int, split>::iterator it = fec_split_map.begin();

	
	
	
	
	
	parity_bits_str.append(parity_offset1, '0');		// should not be used by Yi-Chao, this is coz CSI read just once!
	while(it != fec_split_map.end())
	{	
		int fec_group_index = (*it).first;
		split _split = (*it).second;

		int num_data_bits_grp = min(fec_K, n_data_bits - fec_group_index * fec_K);
		// printf("num_data_bits_grp: %d, sorted_orig_indices.size: %d, orig_str size: %d\n", num_data_bits_grp, sorted_orig_pos.size(), orig_str.size()); fflush(stdout);
		mac_data_bits = orig_str.substr(fec_group_index * fec_N, num_data_bits_grp);

//printf("num_data_bits_grp: %d, fec_group_index: %d, mac_data_bits.size: %d sorted_orig_pos.size: %d\n", num_data_bits_grp, fec_group_index, mac_data_bits.size(), sorted_orig_pos.size()); fflush(stdout);
// construct the bad/white groups //
		std::string sorted_mac_data_bits;
		for(int i = 0; i < num_data_bits_grp; i++)
		{
//printf("%d ", sorted_orig_pos[_split.indices_offset + i]);
			int pos = sorted_orig_pos[_split.indices_offset + i] - fec_group_index * fec_N;
//printf("%d(%c) ", pos, mac_data_bits[pos]);
			sorted_mac_data_bits.append(1, mac_data_bits[pos]);
		}
//printf("============================================================== \n");
//printf("\nencodeMAC_FEC_with_RS_split for fec_group: %d", fec_group_index);
		std::string t_parity_str; 
		encodeMAC_FEC_with_RS_split(sorted_mac_data_bits, 
			_split, t_parity_str); 
//printf("============================================================== \n\n\n");

		int parity_offset = _split.parity_offset;
		// printf("parity_offset.. %d, size: %d \n", parity_offset, t_parity_str.size()); fflush(stdout);
		parity_bits_str.replace(parity_offset, t_parity_str.size(), t_parity_str);

		it++;
	}
	
	

// now interleave the mac parity as well into 'final_parity_bits_str' ! //
	parity_added = parity_bits_str.size();
}

/* do reed-solomon coding over the mac bytes provided resulting in 'parity_bit_str' (non-split version) */
		void
			encodeMAC_FEC_with_RS(std::string mac_bits, int mac_fec_N, int mac_fec_K, 
			int symSize, std::string &parity_bits_str, bool pad_at_byte_boundary)
		{
//printf("\nencodeMAC_FEC_with_RS, n_data: %d, n: %d, k: %d, symSize: %d\n", mac_bits.size(), mac_fec_N, mac_fec_K, symSize);
//fflush(stdout);

			assert(mac_fec_N % symSize == 0);
#ifdef RS_MAC_OPT_VERSION
	// will only be executed for the 'good' part, since thd 'bad' data part is always on the boundary (performSplit) //
//if(mac_fec_K % symSize != 0)
			if(mac_bits.size() % mac_fec_K != 0)
			{
	//int pad = symSize - (mac_fec_K % symSize);
				int pad = mac_fec_K - (mac_bits.size() % mac_fec_K);
				mac_bits.append(pad, '0');
	//printf("had to pad %d bits to the mac_bits\n", pad); fflush(stdout);

#ifdef DEBUG_MAC_FEC
				printf("good data part at the encoder::: \n");
				for(int i = 0; i < mac_bits.size(); i++)
					printf("%c", mac_bits[i]);
				printf("\n\n");
#endif
			}
#endif
			int num_data_bits_last_blk = mac_bits.size();
			assert(mac_fec_K % symSize == 0);

//std::string rs_encoded_mac_bytes = gr_encode_rs_fec(mac_bits, mac_fec_N/symSize, mac_fec_K/symSize);

			std::string rs_encoded_symbols = gr_encode_rs_fec(mac_bits, mac_fec_N/symSize, mac_fec_K/symSize);

#ifdef DEBUG
			printf("rs_encoded_symbols at sender for mac-fec: %d symbols\n", rs_encoded_symbols.size());
			for(int i = 0; i < rs_encoded_symbols.size(); i++)
				printf("%d ", ((const unsigned char*) rs_encoded_symbols.data())[i]);
			printf("\n");   
#endif

//std::string rs_bits = getBitString((const unsigned char*) rs_encoded_mac_bytes.data(), rs_encoded_mac_bytes.size());
			std::string rs_bits = getBits((const unsigned char*) rs_encoded_symbols.data(), mac_fec_N/symSize, symSize);
			unsigned char mac_coded_block[mac_fec_N/symSize];
			int llen = convertToSymbols(mac_coded_block, symSize, rs_bits);

// copy the parity part of each fec blk //
			int parity_bits_blk = (mac_fec_N - mac_fec_K);
//printf("mac_fec_N: %d, mac_fec_K: %d\n", mac_fec_N, mac_fec_K); fflush(stdout);
//printf("num_mac_fec_blks: %d, rs_encoded_mac_bytes: %d, rs_bits: %d, parity_bits_blk: %d\n", num_mac_fec_blks, rs_encoded_mac_bytes.size(), rs_bits.size(), parity_bits_blk);

			parity_bits_str.append(rs_bits.substr(num_data_bits_last_blk, parity_bits_blk));

//printf("parity_added: %d\n", parity_bits_str.size()); fflush(stdout);

// make the parity at the byte border as well //
			if(pad_at_byte_boundary)
			{
				while(parity_bits_str.size() % 8 != 0)
				{
					parity_bits_str.append(1, '0');
				}
			}
		}

/* split version */
		void
			encodeMAC_FEC_with_RS_split(std::string data_bits,
			split _split,
			std::string &parity_bits_str)
		{
			int total_mac_bits = data_bits.size();

			int r_bad = _split.r_bad;
			int d_bad = _split.d_bad;
			int r_good = _split.r_good;
			int d_good = _split.d_good;

// just for info //
			int d_good_ideal = (total_mac_bits - d_bad);
			int d_good_pad = d_good - d_good_ideal;

			int n_bad = d_bad + r_bad;
			int n_good = d_good + r_good;

//printf("\nencodeMAC_FEC_with_RS_split Total: %d, d_good: %d, r_good: %d, n_good: %d, d_bad: %d, r_bad: %d, n_bad: %d, d_good_ideal: %d, d_good_pad: %d, data_bits: %d\n", total_mac_bits, d_good, r_good, n_good, d_bad, r_bad, n_bad, d_good_ideal, d_good_pad, data_bits.size());
//fflush(stdout);

			if(r_bad > 0)
			{
				int symSize = getSymSizeFromN_Bits(n_bad);
	//printf("symSize-bad: %d\n", symSize); fflush(stdout);
				encodeMAC_FEC_with_RS(data_bits.substr(0, d_bad), n_bad, d_bad, symSize, parity_bits_str, false);
				std::string bad_bits = data_bits.substr(0, d_bad);
#ifdef DEBUG
				printf("sender: bad data part:: \n");
				for(int i = 0; i < bad_bits.size(); i++)
					printf("%c", bad_bits[i]);
				printf("\n");

				printf("bad parity:::\n");
				for(int i = 0; i < parity_bits_str.size(); i++)
					printf("%c", parity_bits_str[i]);
				printf("\n");
				int jj = (bad_bits.size() + parity_bits_str.size())/symSize;
				unsigned char b_blk[jj];
				int _num = convertToSymbols(b_blk, symSize, bad_bits + parity_bits_str);  
				printf("sender: mac-fec block(bad): %d symbols\n", _num);
				for(int ii = 0; ii < _num; ii++)
					printf("%d ", b_blk[ii]);
				printf("\n");
#endif
			}

			if(r_good > 0)
			{
				int debug_parity_size = parity_bits_str.size();
				int symSize = getSymSizeFromN_Bits(n_good);
	//printf("symSize-good: %d\n", symSize); fflush(stdout);
				encodeMAC_FEC_with_RS(data_bits.substr(d_bad, d_good_ideal), n_good, d_good, symSize, parity_bits_str, false);
				std::string good_bits = data_bits.substr(d_bad, d_good_ideal);
#ifdef DEBUG_MAC_FEC
	/*
	printf("good-bits : \n");
	for(int i = 0; i < good_bits.size(); i++)
	printf("%c", good_bits[i]);

printf("good-parity: \n");*/
	std::string d_parity = parity_bits_str.substr(debug_parity_size, parity_bits_str.size() - debug_parity_size);
	/*
	for(int i = 0; i < d_parity.size(); i++)
	printf("%c", d_parity[i]);

printf("\n"); */

	printf("sender: mac-fec block(good-data symbols): %d bits\n", good_bits.size());
std::string _good_bits = data_bits.substr(d_bad, d_good);
_good_bits.append(d_good_pad, '0');
_good_bits.append(d_parity);
int jj = (_good_bits.size())/symSize;
unsigned char b_blk[jj];
int _num = convertToSymbols(b_blk, symSize, _good_bits);
printf("sender: mac-fec block(good-data symbols): %d bits %d symbols\n", good_bits.size(), _num);
for(int ii = 0; ii < _num; ii++)
	printf("%d ", b_blk[ii]);
printf("\n");
#endif
}
}

#endif //PERFORM_MAC_FEC

int
	getSymSizeFromN_Bits(int n)
{
	int symSize = 0;
	switch(n)
	{
		case 6:     symSize = 2; break;
		case 21:    symSize = 3; break;
		case 60:    symSize = 4; break;
		case 155:   symSize = 5; break;
		case 378:   symSize = 6; break;
		case 889:   symSize = 7; break;
		case 2040:  symSize = 8; break;
	}
	return symSize;
}

int
	getSymSizeFromN_Syms(int n)
{
	int symSize = 0;
	switch(n)
	{
		case 3:     symSize = 2; break;
		case 7:    symSize = 3; break;
		case 15:    symSize = 4; break;
		case 31:   symSize = 5; break;
		case 63:   symSize = 6; break;
		case 127:   symSize = 7; break;
		case 255:  symSize = 8; break;
	}
	return symSize;
}

/* a more cleaner branch just for the split version - the encoder doesn't need something like this.
* Just figures out the fec_split_map for (possible) mac-fec decoding later.. 
*/
#ifdef RS_MAC_OPT_VERSION
	void
	decodeMACLayer_FEC_OPT(std::string &deinterleaved_str,
	std::string parity_bits_str,
	int *interleaved_bit_pos,
	int *orig_bit_pos,
	int total_subcarriers, int fec_N,
	int fec_K, int bps, int num_bits_before_deinterleaving,
	std::map<int, split> &fec_split_map,
	std::vector<int> &sorted_orig_pos,
	int num_hdr_symbols, 
	int mac_fec_N, int mac_fec_K, int hint_flag)
{
	// printf("\n\n########################################### decodeMACLayer_FEC_OPT ################################################################### \n"); 
// get the interleaved string //
	const unsigned char *data = (const unsigned char*) deinterleaved_str.data();
	unsigned int len = deinterleaved_str.size();
	assert(len % 8 == 0);                        // byte boundary

	static bool initialized = false;
	static std::map<float, float> snr2ser_map;                    // (snr, ser) map


	if(!initialized)
	{
		float *csi = (float*) malloc(sizeof(float) * total_subcarriers);
		memset(csi, 0, sizeof(float) * total_subcarriers);

		initialized = true;
		int parity_added = 0;
		initializeMACLayerFEC_OPT(deinterleaved_str.size(),
			num_bits_before_deinterleaving,
			interleaved_bit_pos,
			csi,
			fec_split_map,
			sorted_orig_pos,
			snr2ser_map,
			total_subcarriers, fec_N, fec_K, bps, 
			parity_added,
			false, hint_flag);
		initialized = true;
	// sorted_orig_pos currently contains interleaved pos, should contain orig pos //

		for(int i = 0; i < sorted_orig_pos.size(); i++)
			sorted_orig_pos[i] = orig_bit_pos[sorted_orig_pos[i]];

		free(csi);
	}
}
#endif

/* calculates the delivery rates of all the FEC groups and puts them in 'fec_delivery_rates_arr'
* order them in the increasing order of their delivery rates
* selects the FEC groups whose delivery rates are below a certain threshold
*/
void
selectFECGroupsToProtect(std::vector<int> &sel_fec_groups, int total_bits, 
float *fec_delivery_rates_arr, int fec_K, int fec_N, 
int *interleaved_bit_pos, float *eff_delivery_rate, 
float &hdr_delivery_rate, int hint_flag)
{
	int num_fec_groups = ceil((float) total_bits/fec_N);
	int *fec_groups = (int*) malloc(sizeof(int) * num_fec_groups);
	memset(fec_groups, 0, sizeof(int) * num_fec_groups);

	int symSize = getSymSizeFromN_Bits(fec_N);
	int nsymbols = fec_N/symSize;
	assert(fec_N % symSize == 0);
	assert(fec_K % symSize == 0);

	// printf("selectFECGroups, fec_N: %d, fec_K: %d, num_fec_groups: %d, nsymbols: %d\n", fec_N, fec_K, num_fec_groups, nsymbols); fflush(stdout);

	float *t_eff_delivery_rate = (float*) malloc(sizeof(float) * nsymbols);		// for input to getFECGroupDeliveryRate //

	int count = 0;
	for(int i = 0; i < num_fec_groups; i++)
	{
		int nbits = min(fec_N, total_bits - (i*fec_N));					// in RS, the last group might have fewer than fec_N bits //
		int ndatabits = nbits - (fec_N - fec_K);
		int ncodewords = nbits/symSize;	
		assert(nbits % symSize == 0);	

	// 't_eff_delivery_rate' for the entire FEC-group//
		float avg_dr = 0.0;
		float hdr_bits_dr = 1.0;
		int bit_count = 0;

		for(int j = 0; j < ncodewords; j++)		
		{
			t_eff_delivery_rate[j] = 1.0;
			for(int k = 0; k < symSize; k++)	
			{
				int _orig_bit_pos = (i * fec_N) + (j * symSize) + k;	
				int _interleaved_bit_pos = interleaved_bit_pos[_orig_bit_pos];
				t_eff_delivery_rate[j] *= eff_delivery_rate[_interleaved_bit_pos];

		// only for the data part of the fec-group //
				if(bit_count < ndatabits)
					avg_dr += eff_delivery_rate[_interleaved_bit_pos];

		// fec-group 0, so use it for calculating the header delivery rate //
				if(i == 0 && bit_count < 192)
					hdr_bits_dr *= eff_delivery_rate[_interleaved_bit_pos];		   
				bit_count++;
			}
		}

		float dr = getFECGroupDeliveryRate(t_eff_delivery_rate, ncodewords, fec_N/symSize, fec_K/symSize);

	// calculate the hdr delivery probability, if it is fec-group 0 //
		if(i == 0)
			hdr_delivery_rate = 1 - (1 - dr) * (1 - hdr_bits_dr);

	// estimate of bits delivered in this fec-group //
		float est_bits_delivered = dr * ndatabits;
		if(hint_flag == 1)
		{
			avg_dr = avg_dr/ndatabits;
			est_bits_delivered = est_bits_delivered + (1 - dr) * avg_dr * ndatabits;
		}

		// printf("fec_group: %d, est_bits_delivered: %f, ndatabits: %d, dr: %f, avg_dr: %f\n", i, est_bits_delivered, ndatabits, dr, avg_dr);

		if(dr == ndatabits)
			continue;	

		fec_delivery_rates_arr[count] = est_bits_delivered;
		fec_groups[count] = i;              						 // initialize the indices
		count++;
	}

	free(t_eff_delivery_rate);

//sortFECGroupsByDeliveryRate(fec_delivery_rates_arr, fec_groups, count);		// remove the sorting
	sel_fec_groups.assign(fec_groups, fec_groups + count);

#ifdef DEBUG_MAC_FEC
	printf("selected FEC groups --- \n");
	for(int i = 0; i < sel_fec_groups.size(); i++)
		printf("FEC_group: %d, delivery_rate: %f\n", sel_fec_groups.at(i), fec_delivery_rates_arr[i]);;
#endif

	free(fec_groups);
}

/* initializes the effective delivery rates of the bits as their symbol error rates,
* takes into account the subcarriers on which they are mapped
*/
	void initializeEffectiveDeliveryRate(float *eff_delivery_rate, float *csi,
	int total_bits,  int *interleaved_bit_pos, 
	std::map<float, float> &snr2ser_map, int bps, int total_subcarriers, 
	int num_hdr_symbols)
{
	for(int i = 0; i < total_bits; i++)
	{
		int sym_index = i/bps + num_hdr_symbols;
		int sc_index = sym_index%total_subcarriers;
		eff_delivery_rate[i] = 1 - getSymbolErrorRate2(csi, sc_index, bps);
		assert(eff_delivery_rate[i] <= 1);
#ifdef DEBUG_MAC_FEC
		printf("eff_del_rate interleaved_pos: %d -- : %f, sym: %d, sc: %d csi: %f\n", i, eff_delivery_rate[i], sym_index, sc_index, csi[sc_index]); fflush(stdout);
#endif
	}

#ifdef DEBUG_MAC_FEC
	for(int i = 0; i < total_bits; i++)
	{
		int interleaved_pos = interleaved_bit_pos[i];
		printf("orig_pos: %d, interleaved_pos: %d, eff_delivery_rate: %f\n", i, interleaved_pos, eff_delivery_rate[interleaved_pos]);
	}
#endif
}

/* sort the bits by their effective delivery rate 
*/
void sortBitsByEffectiveDeliveryRate(float *sorted_eff_delivery_rates_arr, 
	int *sorted_bit_indices,
	float *eff_delivery_rate, int *interleaved_bit_pos,
	int fec_group_index, int n_databits, int fec_N)
{
// initialize the arrays to be sorted //
	for(int i = 0; i < n_databits; i++)
	{
		int _orig_bit_pos = fec_group_index * fec_N + i;
		int _interleaved_bit_pos = interleaved_bit_pos[_orig_bit_pos];				// interleaved position for this bit
		sorted_eff_delivery_rates_arr[i] = eff_delivery_rate[_interleaved_bit_pos];
		sorted_bit_indices[i] = _interleaved_bit_pos;
	}

	isort(sorted_eff_delivery_rates_arr, sorted_bit_indices, n_databits, true);

#ifdef DEBUG_MAC_FEC
	printf("sortBitsByEffectiveDeliveryRate of data_bits::::::::: FEC_GROUP: %d\n", fec_group_index);
	for(int i = 0; i < n_databits; i++)
	{
		printf("bit: %d, eff_DR: %f\n", sorted_bit_indices[i], sorted_eff_delivery_rates_arr[i]);
	}
#endif
}

/* gets the Symbol Error Rate from the SNR->SER mapping
* find the snr of the subcarrier, 'sc_index'
* to get the ser for the given snr:
*      - if given snr smaller than the smallest snr in the map, then return the corresponding ser
*      - if direct match found, then return the corresponding ser
*      - else, interpolate as
*              - ser = (prev_ser + (snr - prev_snr)/(curr_snr - prev_snr) * (curr_ser - prev_ser))
*/
float getSymbolErrorRate(float *csi, int sc_index, std::map<float, float> snr2ser_map)
{


	float snr = csi[sc_index];

	float prev_snr = -1;
	float prev_ser = -1.0;

	float curr_snr = 0.0;
	float curr_ser = 0.0;

	std::map<float, float>::iterator it = snr2ser_map.begin();
	while(it != snr2ser_map.end())
	{
		curr_snr = (*it).first;
		curr_ser = (*it).second;

		if(snr <= curr_snr)
			break;
		if(snr == curr_snr)
				// case 2 //
			return curr_ser;

		prev_snr = curr_snr;
		prev_ser = curr_ser;
		it++;
	}

	if(prev_ser == -1.0)
		// case 1 //
		return curr_ser;
	else
		// case 3: interpolation //
		return (prev_ser + ((float) (snr - prev_snr))/(curr_snr - prev_snr) * (curr_ser - prev_ser));
}

void sortFECGroupsByDeliveryRate(float *delivery_rates, int *indices, int n)
{
//printf("sortFECGroupsByDeliveryRate n: %d\n", n); fflush(stdout);
	isort(delivery_rates, indices, n, true);
}

// simple sorting code //
int
	fm(float arr[], int b, int n, bool ascending)
{
	int f = b;
	int c;

	for(c = b + 1; c < n; c++)
		if((ascending && arr[c] < arr[f]) || (!ascending && arr[c] > arr[f]))
		f = c;

	return f;
}

void
	isort(float arr[], int indices[], int n, bool ascending)
{
	int s, w;
	float sm;
	int ind;

	for(s = 0; s < n - 1; s++) {
		w = fm(arr, s, n, ascending);
		sm = arr[w];
		ind = indices[w];
		arr[w] = arr[s];
		arr[s] = sm;
		indices[w] = indices[s];
		indices[s] = ind;
	}
}

// snr to ser mapping code //

/* reads from the SNR->SER file and sets up the 'snr2ser_map' */
void
	setSNR_to_SER_Mapping(std::map<float, float> &snr2ser_map)
{
	const char *filename1 = "snr2ser.qam64.s";

	int fd;
	FILE *d_fp;
	if ((fd = open (filename1,
		O_RDONLY|OUR_O_LARGEFILE,
	0664)) < 0){
		perror(filename1);
	}else {
		if((d_fp = fdopen (fd, true ? "r" : "rt")) == NULL) {
			fprintf(stderr, "snr-ser file cannot be opened\n");
			close(fd);
			assert(false);
		}
		//fprintf(stderr, "snr-ser file opened!\n");
		float snr;
		float ser;
		char line[100];

		while(fgets(line, 100, d_fp) != NULL)
		{
			sscanf(line, "%f %f", &snr, &ser);
			snr2ser_map[snr] = ser;
		}
		fclose(d_fp);
	}
#ifdef DEBUG_MAC_FEC
// print the entries //
	std::map<float, float>::iterator it = snr2ser_map.begin();
	while(it != snr2ser_map.end())
	{
		// printf(" %f --> %f\n", (*it).first, (*it).second);
		it++;
	}
#endif
}


// fake - only for debugging // 
void readCSISubcarriers1(float *csi, const int total_subcarriers)
{
//srand(1);
// make 4 subcarriers really bad //
	for(int i = 0; i < total_subcarriers/12; i++)
		csi[i] = ((float)(rand()))/RAND_MAX + 10;


	for(int i = total_subcarriers/12; i < total_subcarriers; i++)
		csi[i] = ((float)(rand()))/RAND_MAX + 10;

#ifdef DEBUG_FEC
	printf("CSI of subcarriers;: \n"); 
	for(int i = 0; i < total_subcarriers; i++)
		printf("%d %lf\n", i, csi[i]);
#endif
}

/* reads the BER instead */
void readCSISubcarriers(float *csi, const int total_subcarriers)
{
	const char *filename1 = "bers.txt";
	printf("total_subcarriers: %d\n", total_subcarriers); fflush(stdout);
	int fd;
	FILE *d_fp;

	int count = 0;
	if ((fd = open (filename1,
		O_RDONLY|OUR_O_LARGEFILE,
	0664)) < 0){
		perror(filename1);
	}else {
		if((d_fp = fdopen (fd, true ? "r" : "rt")) == NULL) {
			fprintf(stderr, "CSI file cannot be opened\n");
			close(fd);
			assert(false);
		}
		char line[1700];

		float *tmp = (float*) malloc(sizeof(float) * total_subcarriers);
		memset(tmp, 0, sizeof(float) * total_subcarriers);

		int sc_index = 0;
		while(fgets(line, 1700, d_fp) != NULL)
		{
			sscanf(line, "%f", &csi[sc_index]);
			sc_index++;
		}
		fclose(d_fp);
		free(tmp);
	}
}


void readCSISubcarriers11(float *csi, const int total_subcarriers)
{
	const char *filename1 = "csi.dat"; //"l1.dat_matrix_dB";
	printf("total_subcarriers: %d\n", total_subcarriers); fflush(stdout);
	int fd;
	FILE *d_fp;

	int count = 0;
	if ((fd = open (filename1,
		O_RDONLY|OUR_O_LARGEFILE,
	0664)) < 0){
		perror(filename1);
	}else {
		if((d_fp = fdopen (fd, true ? "r" : "rt")) == NULL) {
			fprintf(stderr, "CSI file cannot be opened\n");
			close(fd);
			assert(false);
		}
		char line[1700];

		float *tmp = (float*) malloc(sizeof(float) * total_subcarriers);
		memset(tmp, 0, sizeof(float) * total_subcarriers); 

		int sc_index = 0;
		while(fgets(line, 1700, d_fp) != NULL)
		{
			sscanf(line, "%f", &csi[sc_index]);
			sc_index++;
		}
		fclose(d_fp);
		free(tmp);
	}

	for(int i = 0; i < total_subcarriers/12; i++)
		csi[i] = csi[i] - 3;
}

void sortSubcarriersByCSI(int *subcarrier_indices, float *csi, int total_subcarriers)
{
// initialize the indices //
	for(int i = 0; i < total_subcarriers; i++)
		subcarrier_indices[i] = i;

	isort(csi, subcarrier_indices, total_subcarriers, true);

#ifdef DEBUG_MAC_FEC
	printf("sorted subcarriers and CSI: \n");
	for(int i = 0; i < total_subcarriers; i++)
		printf("sc: %d csi: %f\n", subcarrier_indices[i], csi[i]);
	fflush(stdout);
#endif
}


// utility functions //
// convert the entire data into bits //
std::string
	getBitString(const unsigned char *data, int len)
{
	std::string orig_s;
	for(int i = 0; i < len; i++)
	{
		std::bitset<sizeof(unsigned char) * 8> s_bits(data[i]);
		std::string tmp_string = s_bits.to_string<char,std::char_traits<char>,std::allocator<char> >();
		orig_s.append(tmp_string);
	}
	return orig_s;
}

// converts a bitstring back into a proper msg //
std::string
	getMsg(std::string str, unsigned char *data)
{
#ifdef DEBUG
	if(str.size() % 8 != 0)
	{
		printf("%d ", str.size()); fflush(stdout);
	}
#endif

	assert(str.size() % 8 == 0);
#ifdef DEBUG
	for(int i = 0; i < str.size(); i++)
		printf("%d", str[i]);
	printf("\n");
#endif
	int s_len = str.length();
	int k = 0;
	for(int i = 0; i < s_len; i=i+8)
	{
		std::bitset<sizeof(unsigned char)*8> s_bits(str.substr(i, 8));
		data[k] = (unsigned char) s_bits.to_ulong();
		k++;
	}

#ifdef DEBUG
	printf("Final Msg: \n");
	for(int i = 0; i < k; i++)
	{
		printf("%d ", data[i]);
	}
	printf("\n");
#endif

	std::string ret_s((const char*) data, k);
	return ret_s;
}

std::string injectErrorsProbabilistically(std::string buf, float *csi, int subcarriers, int bps)
{
	// printf("injectErrorsProbabilistically\n");
	std::string bits = getBitString((unsigned char*) buf.data(), buf.size()); 
//srand(1);

	int n_errors = 0;
	for(int i = 0; i < bits.size(); i++)
	{
		int sym_index = i/bps;
		int sc_index = sym_index % subcarriers;

		float ber = getSymbolErrorRate2(csi, sc_index, bps);	
	//printf("%lf ", ber); 	
		float _rand = ((float)rand())/RAND_MAX; 
	//printf("%lf---%lf ", _rand, ber);	
		if(_rand <= ber)
		{
			bits.replace(i, 1, 1, bits[i] ^ 1);
		//printf("%d         ", i); fflush(stdout);
			n_errors++;
			// if(i < 192)
			// 	printf("bit: %d on sc: %d (%lf, %lf)\n", i, sc_index, ber, _rand); fflush(stdout);
		}
	}
	// printf("errors: %d\n", n_errors);
	unsigned char *err_data = (unsigned char*) malloc(sizeof(unsigned char) * buf.size());
	memset(err_data, 0, sizeof(unsigned char) * buf.size());
	std::string t = getMsg(bits, err_data);
	return t;
}

#ifdef RS_MAC_OPT_VERSION
/* optimized version - returns the number of MAC FEC groups formed, one phy-FEC group *can* have 
* atmost 2 MAC-FEC groups for it (one for good bits, other for bad) */
	void
	initializeMACLayerFEC_OPT(int num_bits_before_interleaving,
	int num_bits_after_interleaving,
	int *interleaved_bit_pos,
	float *csi,
	std::map<int, split>& fec_split_map,
	std::vector<int>& sorted_orig_pos,
	std::map<float, float> &snr2ser_map,
	const int total_subcarriers,
	int fec_N, int fec_K, int bps, 
	int &parity_added, bool encode, int hint_flag)
{
	// printf("initializeMACLayerFEC_OPT (RS OPT version) \n"); fflush(stdout);

// original string before the interleaving //
	int num_fec_groups = ceil((float) num_bits_before_interleaving/fec_N);
	// printf("num_bits_before_interleaving: %d, fec_N: %d, groups: %d\n", num_bits_before_interleaving, fec_N, num_fec_groups);

// get the total number of data bits and redundancy bits //
	int n_redundancy_bits = num_fec_groups * (fec_N - fec_K);
	int n_data_bits = num_bits_before_interleaving - n_redundancy_bits;

// read subcarrier csi values //   
//readCSISubcarriers(csi, total_subcarriers);
	memcpy(csi, g_csi, total_subcarriers * sizeof(float));  

// eff delivery rate for all interleaved positions (has to account for padded bits as well, since they
// are part of the interleaved string)
	float *eff_delivery_rate = (float*) malloc(sizeof(float) * num_bits_after_interleaving);
	memset(eff_delivery_rate, 0, sizeof(float) * num_bits_after_interleaving);

// initialize the effective delivery rates for all the interleaved positions possible //
	initializeEffectiveDeliveryRate(eff_delivery_rate, csi, 
		num_bits_after_interleaving,
		interleaved_bit_pos,
		snr2ser_map, bps, total_subcarriers, 0);

// get the delivery rates for all the FEC groups //
	float *fec_delivery_rates_arr = (float*) malloc(sizeof(float) * num_fec_groups);
	memset(fec_delivery_rates_arr, 0, sizeof(float) * num_fec_groups);

	// printf("num_fec_groups: %d\n", num_fec_groups); fflush(stdout);

// get the FEC groups to protect //
	std::vector<int> fec_groups_to_protect;
	float hdr_delivery_rate;

	selectFECGroupsToProtect(fec_groups_to_protect, num_bits_before_interleaving, fec_delivery_rates_arr, fec_K, fec_N, interleaved_bit_pos, eff_delivery_rate, hdr_delivery_rate, hint_flag);

	int num_fec_groups_to_protect = fec_groups_to_protect.size();

	// printf("num_fec_groups_to_protect: %d\n", num_fec_groups_to_protect); fflush(stdout);
	// printf("The current utility ::: \n");
	float curr_utility = getUtility(fec_delivery_rates_arr, eff_delivery_rate, interleaved_bit_pos, n_data_bits, num_bits_after_interleaving, fec_K, fec_N, hdr_delivery_rate, hint_flag); 

	float old_utility = curr_utility;
	int index = 0;
	parity_added = 0;
	int interleaved_parity_pos[bps * total_subcarriers];
	initInterleavedParityPos(interleaved_parity_pos, bps, total_subcarriers); 

/* loop over all the FEC groups as long as the throughput condition is met
* On each iteration, take the fec group; exhaustively search it to get the 
* split, and finally perform FEC
*/

// printf("start, utility: %f\n", curr_utility); 
std::map<int, split>::iterator it_m = fec_split_map.begin();
for(int index = 0; index < fec_groups_to_protect.size(); index++)
{
	int sel_fec_index = -1;
	int sel_index = -1;
	float sel_new_delivery_rate_fec = -1.0;

	int fec_group_index = fec_groups_to_protect[index];

	int num_data_bits_grp = min(fec_K, n_data_bits - fec_group_index * fec_K);

	float old_del_rate_fec = fec_delivery_rates_arr[index];

	// call lili's code here //
	int d_bad, d_good, r_bad, r_good;
	std::vector<int> t_pos; 
	//float new_del_rate_fec = 
	curr_utility = performSplitFEC(num_data_bits_grp, (fec_N - fec_K), d_bad, r_bad, d_good, r_good,
		eff_delivery_rate, interleaved_bit_pos, t_pos,
		parity_added, total_subcarriers, bps,
		csi, snr2ser_map, fec_group_index, fec_N, fec_K, interleaved_parity_pos, 
		fec_delivery_rates_arr, n_data_bits, num_bits_after_interleaving, old_utility, hdr_delivery_rate, hint_flag);

	if(curr_utility <= old_utility)
	{
		// printf("Utility is too low .. \n");
		continue;	
	}
	old_utility = curr_utility;

	/*	
	printf("old_del_rate_fec: %f, new_del_rate: %f, index: %d\n", old_del_rate_fec, new_del_rate_fec,
	index);
	if(new_del_rate_fec < old_del_rate_fec)
	{
	if(index == (fec_groups_to_protect.size() - 1))
	break;
	else
	continue;
	}


// utility condition //
fec_delivery_rates_arr[index] = new_del_rate_fec;	// update it for now //
new_utility = getUtility(fec_delivery_rates_arr, eff_delivery_rate, interleaved_bit_pos, n_data_bits, num_bits_after_interleaving + (parity_added + r_bad + r_good), fec_K);

printf("new_utility: %f, old_utility: %f\n", new_utility, old_utility); 
if(new_utility < old_utility)
{
printf("Utility is too low.. \n");
fec_delivery_rates_arr[index] = old_del_rate_fec;	// rejected, so restore it //
continue;
}
*/

	// just store the details - fec_group_index, d_bad, r_bad, for each iteration // 
// printf("updating the fec_split_map::: , fec_group: %d, parity_offset: %d\n", fec_group_index, parity_added);
fec_split_map.insert(it_m, pair<int, split>(fec_group_index, make_split(r_bad, d_bad, r_good, d_good, sorted_orig_pos.size(), parity_added)));	

	// update the parity added till now //
parity_added += r_bad + r_good;

	// copy 't_pos' into 'sorted_orig_pos' //
std::vector<int>::iterator it = sorted_orig_pos.begin();
sorted_orig_pos.insert(it + sorted_orig_pos.size(), t_pos.begin(), t_pos.end());

// printf("fec_group_index: %d, parity_added_till_now: %d\n", fec_group_index, parity_added);
// printf(" _____________________________ \n");
}		
#ifdef DEBUG
printf("sorted_orig_pos in initializeM::: \n");
for(int i = 0; i < sorted_orig_pos.size(); i++)
	printf("%d ", sorted_orig_pos[i]);
printf("\n");
#endif

free(fec_delivery_rates_arr);
}

/* returns the interleaved parity bit locations in an interleaver block */
void initInterleavedParityPos(int *interleaved_parity_pos, int bps, int subcarriers)
{
	int C = 16;
	int Ncbps = subcarriers * bps;
	int s = (bps/2 > 1) ? bps/2 : 1;

	for(int k = 0; k < Ncbps; k++)
	{
		int i = (Ncbps/C)*(k%C) + floor((float)k/C);
		interleaved_parity_pos[k] = s * floor((float) i/s) + (int) (i + Ncbps - floor((float) C * i/Ncbps)) % s;
	}
} 


/* converted by Yi-Chao from 11n/scripts/fec.m 
* d1: num of data bits in the phy-layer FEC group
* r1: num of red bits in the phy-layer FEC group
* r2: mac-layer FEC parity budget 
*/
float performSplitFEC(int d1, int r1, int &max_bad, int &max_r2_b, int &max_good, int &max_r2_g, 
float *eff_delivery_rate, int *interleaved_bit_pos, 
std::vector<int>& sorted_interleaved_indices,
int parity_mapped_already, int subcarriers, int bps,
float *csi, std::map<float, float> snr2ser_map,
int fec_group_index, int fec_N, int fec_K, 
int *interleaved_parity_pos, 
float *fec_delivery_rates_arr, int n_data_bits, int num_bits_out, float curr_utility, float &hdr_delivery_rate, int hint_flag)
{
	// initialize these //
	max_bad = 0; 
	max_r2_b = 0; 
	max_good = 0;
	max_r2_g = 0;

	// printf("\n\nperformSplitFEC, fec_group: %d, d1: %d, r1: %d, parity_mapped: %d\n", fec_group_index, d1, r1, parity_mapped_already); fflush(stdout);
	float max_delivery = 0.0;

	// sort the 'data' bits in this FEC group by their eff delivery rates //
	float *sorted_eff_delivery_rate = (float*) malloc(sizeof(float) * d1);
	int *sorted_indices = (int*) malloc(sizeof(int) * d1);
	memset(sorted_indices, 0, sizeof(int) * d1);

	sortBitsByEffectiveDeliveryRate(sorted_eff_delivery_rate, sorted_indices,
		eff_delivery_rate, interleaved_bit_pos,
		fec_group_index, d1, fec_N);


	// fit in the 'sorted_interleaved_indices' //
	for(int i = 0; i < d1; i++)
		sorted_interleaved_indices.push_back(sorted_indices[i]);

#ifdef DEBUG
	printf("performSplitFEC, sorted_interleaved_indices.. fec_group: %d\n", fec_group_index);
	for(int i = 0; i < sorted_interleaved_indices.size(); i++)
		printf("%d ", sorted_interleaved_indices[i]);
	printf("\n");
#endif 

	// delivery rate of the phy-layer parity bits //
	float *del_rate_parity_r1 = (float*) malloc(sizeof(float) * r1);
	for(int i = 0; i < r1; i++)
	{
		int _orig_bit_pos = fec_group_index * fec_N + d1 + i;
		int _interleaved_bit_pos = interleaved_bit_pos[_orig_bit_pos];
		del_rate_parity_r1[i] = eff_delivery_rate[_interleaved_bit_pos];
	}

	// P1: delivery probability of the phy-layer FEC group //
	int symSize = getSymSizeFromN_Bits(fec_N);				// symbol size of this FEC group
	assert((d1 + r1) % symSize == 0);

	assert(fec_N % symSize == 0);					// some sanitary checks
	assert(fec_K % symSize == 0);
	int ncodewords = (d1 + r1)/symSize;					// total symbols (codewords) in this FEC group

	/* --------------- P1: FEC-group delivery of the phy-layer FEC group ------------------ */

	float *t_eff_del_rate = (float*) malloc(sizeof(float) * ncodewords);
	for(int i = 0; i < ncodewords; i++)
	{
		t_eff_del_rate[i] = 1.0;
		for(int j = 0; j < symSize; j++)
		{
			int orig_pos = (fec_group_index * fec_N) + (i * symSize) + j;
			int int_pos = interleaved_bit_pos[orig_pos];
			t_eff_del_rate[i] *= eff_delivery_rate[int_pos];
		}
	}
	float P1 = getFECGroupDeliveryRate(t_eff_del_rate, ncodewords, fec_N/symSize, fec_K/symSize);
	free(t_eff_del_rate);
	/* ----------------------------------- P1 done  ---------------------------------------- */

	float orig_hdr_delivery_rate = hdr_delivery_rate;
	// enter all the possible values that 'N' can take!
	for(int sym_b = 0; sym_b <= 8; sym_b++) {

		int N_b = pow(2.0, sym_b) - 1;					// N: 3, 7, 15 ... 255 

		if(N_b == 1)							// bypasses sym_b 1
			continue;
		int step = sym_b;

		for(int bad = 0; bad <= d1; bad+=step) {			// go over all bits, while ensuring symbol boundary

			int bad_sym, r2_b_sym;
			if(sym_b > 0)
			{
				bad_sym = bad/sym_b; 					// bad-data codewords
				r2_b_sym = N_b - bad_sym;				// bad-redundancy codewords
			}
			else
			{
				bad_sym = bad;						// helps calculating the avg_del_rate_bad
				assert(N_b == 0);
				r2_b_sym = 0;
				step = 1;
			}

			if(r2_b_sym < 0)
				break;

			if(bad == 0)
				r2_b_sym = 0;

			int r2_b = r2_b_sym * sym_b;					// bad-redundancy bits

			for(int sym_g = 2; sym_g <= 8; sym_g++) { 
				int N_g = pow(2.0, sym_g) - 1;

		// potential padding required to ensure good-symbol-size boundary. 
		// this padding of data symbols is only logical, it is *not* sent out on air 
				int good = d1 - bad;
				int good_pad = 0; 
				if(good % sym_g != 0)
				{
					good_pad = sym_g - (good%sym_g);
					good += good_pad;	
				}

		//printf("N_g: %d, sym_g: %d, good: %d, good_pad: %d\n", N_g, sym_g, good, good_pad);
				int good_sym = good/sym_g;				// good-data codewords

				if(good_sym > N_g)
					continue;

				int r2_g_sym = N_g - good_sym;				// good-redundancy codewords
				int r2_g = r2_g_sym * sym_g;				// good-redundancy bits

				// delivery rates of the bad data bits //
				float *del_rate_data1_bad = (float*) malloc(sizeof(float) * bad);
				memcpy(del_rate_data1_bad, sorted_eff_delivery_rate, sizeof(float) * bad);

				// delivery rates of the good data bits //
				float *del_rate_data1_good = (float *)malloc(sizeof(float) * (good - good_pad));		// padding is logical!
				memcpy(del_rate_data1_good, sorted_eff_delivery_rate + bad, sizeof(float) * (good - good_pad));

		// average delivery probability of the good and bad bits //
				float avg_del_rate_data1_good = 0.0;
				float avg_del_rate_data1_bad = 0.0;	

				// ---- P2_b: delivery probability of MAC-layer FEC group involving bad data symbols in PHY-layer ---- //
				ncodewords = bad_sym + r2_b_sym;

				float *del_rate_data1_bad_BER_fec2_b = (float *)malloc(sizeof(float) * ncodewords);
				int mac_parity_mapped_one_round = subcarriers * bps;
				int bad_parity = 0;

				bool isHdr_bad_mac_group = false;
				float del_rate_bad_hdr_bits = 1.0;

				for(int i = 0; i < ncodewords; i++)
				{
					del_rate_data1_bad_BER_fec2_b[i] = 1.0;
					for(int j = 0; j < step; j++)
					{
						int _index = (i * step) + j;
						if(_index < bad)
						{
							del_rate_data1_bad_BER_fec2_b[i] *= sorted_eff_delivery_rate[_index];               // for data
							avg_del_rate_data1_bad += sorted_eff_delivery_rate[_index];

							if(sorted_indices[_index] < 192 && fec_group_index == 0)	
							{
								isHdr_bad_mac_group = true;
								del_rate_bad_hdr_bits *= sorted_eff_delivery_rate[_index];
							}
						}
						else
						{
							bad_parity++;
							int parity_mapped_this_round = (parity_mapped_already + bad_parity) % mac_parity_mapped_one_round;
							int parity_pos = interleaved_parity_pos[parity_mapped_this_round];

							int p_sym_index = parity_pos / bps;
							int p_sc_index = p_sym_index % subcarriers;
							float del_rate_parity_bad = 1 - getSymbolErrorRate2(csi, p_sc_index, bps);

							del_rate_data1_bad_BER_fec2_b[i] *= del_rate_parity_bad;                               // for parity
							assert(del_rate_parity_bad >= -0.0);
							assert(del_rate_parity_bad <= 1.0);
						}
					}
				}

				if(bad > 0)
					avg_del_rate_data1_bad = avg_del_rate_data1_bad/bad;

				float P2_b = getFECGroupDeliveryRate(del_rate_data1_bad_BER_fec2_b, ncodewords, bad_sym + r2_b_sym, bad_sym);
				if(bad_sym == 0)
					P2_b = 1;
				if(r2_b_sym == 0)
					P2_b = 0;
				// ------------------------------------------------------------------------------------ //

				// ---- P2_g: delivery probability of MAC-layer FEC group involving good data symbols in PHY-layer ---- //
				float *del_rate_data1_good_BER_fec2_g = (float *) malloc(sizeof(float) * (good_sym + r2_g_sym));
				int good_parity = 0;

				float del_rate_good_hdr_bits = 1.0;
				bool isHdr_good_mac_group = false;
				ncodewords = good_sym + r2_g_sym;
				for(int i = 0; i < ncodewords; i++)
				{
					del_rate_data1_good_BER_fec2_g[i] = 1.0;
					for(int j = 0; j < sym_g; j++)
					{
						int _index = (i * sym_g) + j;								// bit index
						if(_index < good)									// good-data bits
						{
							if(_index < (good - good_pad))							// 'good_pad' is logical -> ignore!
							{
								del_rate_data1_good_BER_fec2_g[i] *= del_rate_data1_good[_index];	
								avg_del_rate_data1_good += del_rate_data1_good[_index];

								if(sorted_indices[_index] < 192 && fec_group_index == 0)
								{
									isHdr_good_mac_group = true;
									del_rate_good_hdr_bits *= del_rate_data1_good[_index];			
								}
							}
						}
						else											// good-parity bits
						{
							good_parity++;
							int parity_mapped_this_round = (parity_mapped_already + bad_parity + good_parity) % mac_parity_mapped_one_round;
							int parity_pos = interleaved_parity_pos[parity_mapped_this_round];

							int p_sym_index = parity_pos / bps;
							int p_sc_index = p_sym_index % subcarriers;

							float del_rate_parity_good = 1 - getSymbolErrorRate2(csi, p_sc_index, bps);
							del_rate_data1_good_BER_fec2_g[i] *= del_rate_parity_good;			       
						}
					}
				}

				if((good - good_pad) > 0)
					avg_del_rate_data1_good = avg_del_rate_data1_good/(good - good_pad);
				float P2_g = getFECGroupDeliveryRate(del_rate_data1_good_BER_fec2_g, ncodewords, good_sym + r2_g_sym, good_sym);

				if(r2_g_sym == 0)
					P2_g = 0;

				// ------------------------------------------------------------------------------------ //

		// expected delivered data bits (ignore padding on the good bits)
				float delivery = P1*d1 + (1 - P1)*P2_g*(good-good_pad) + (1 - P1)*P2_b*bad;
				if(hint_flag == 1)
					delivery += (1 - P1)*(1 - P2_g)*avg_del_rate_data1_good*(good - good_pad) + (1 - P1)*(1 - P2_b)*avg_del_rate_data1_bad*bad; 

				float old_hdr_delivery_rate = hdr_delivery_rate;	
		// update the header delivery probability //
				if(fec_group_index == 0)
				{
					float prob = 1.0;
					bool prob_changed = false;

					if(r2_b_sym > 0)
					{
						if(isHdr_bad_mac_group)
						{
							prob *= P2_b;
							prob_changed = true;
						}

						if(r2_g_sym == 0 && isHdr_good_mac_group)
						{
							prob *= del_rate_good_hdr_bits;
							prob_changed = true;
						}
						else if(r2_g_sym  > 0 && isHdr_good_mac_group)
						{
							prob *= P2_g;
							prob_changed = true;
						}
					}	

					else if(r2_g_sym > 0)
					{
						if(isHdr_good_mac_group)
						{
							prob *= P2_g;
							prob_changed = true;
						}

						if(isHdr_bad_mac_group)
						{
							prob *= del_rate_bad_hdr_bits;
							prob_changed = true;
						}
					}


					if(!prob_changed)
						prob = 0.0;

					float rhs = 1 - orig_hdr_delivery_rate;
					rhs *= (1 - prob);
					hdr_delivery_rate = 1 - rhs;
				}

		// global utility //
				float old_delivery_rate = fec_delivery_rates_arr[fec_group_index];
				fec_delivery_rates_arr[fec_group_index] = delivery;
				float utility = getUtility(fec_delivery_rates_arr, eff_delivery_rate, interleaved_bit_pos, n_data_bits, num_bits_out + parity_mapped_already + bad_parity + good_parity, fec_K, fec_N, hdr_delivery_rate, hint_flag); 


				//printf("bad: %d, r2_b: %d, sym_b: %d, good: %d, good_pad: %d, r2_g: %d sym_g: %d delivery: %f, utility: %f curr_utility: %f\n", bad, r2_b, sym_b, good, good_pad, r2_g, sym_g, delivery, utility, curr_utility);
		//printf("\n");
		//printf("P1: %f, P2_b: %f, P2_g: %f, avg_del_rate_data1_bad: %f, avg_del_rate_data1_good: %f\n\n", P1, P2_b, P2_g, avg_del_rate_data1_bad, avg_del_rate_data1_good); fflush(stdout);

		//printf("bad: %d, r2_b: %d, good: %d r2_g: %d\n", bad, r2_b, good-good_pad, r2_g); fflush(stdout);

				assert(delivery <= d1);
				if(utility > curr_utility)
				{
					max_delivery = delivery;
					curr_utility = utility;
					max_bad = bad;
					max_r2_b = r2_b;
					max_r2_g = r2_g;
					max_good = good_sym * sym_g;
					old_hdr_delivery_rate = hdr_delivery_rate;
				}
				else
				{
					fec_delivery_rates_arr[fec_group_index] = old_delivery_rate;
					hdr_delivery_rate = old_hdr_delivery_rate;
				}

				free(del_rate_data1_good_BER_fec2_g);
				free(del_rate_data1_bad_BER_fec2_b);

				free(del_rate_data1_bad);
				free(del_rate_data1_good);
			}
		}
	}

	free(sorted_indices);
	free(sorted_eff_delivery_rate);
	free(del_rate_parity_r1);

	if(max_bad == 0)
		max_r2_b = 0;

	if(max_good == 0)
		max_r2_g = 0;

	// printf("bad = %d, r2_b = %d, good = %d, r2_g = %d, delivery = %f, curr_utility = %f\n", max_bad, max_r2_b, max_good, max_r2_g, max_delivery, curr_utility);

	return curr_utility;
	//return max_delivery/d1;
}
#endif

std::string
	getPayloadOnly(std::string buf, int fec_N, int fec_K, int hdr_data_size)
{
	int len = buf.size();
	int num_fec_grps = ceil((float) len/fec_N);

	int total_fec = num_fec_grps * (fec_N - fec_K);
	int remaining_bytes = buf.size() - total_fec;

	int payload_only_size = remaining_bytes - hdr_data_size;

	if(fec_N == 0)
		return buf.substr(hdr_data_size, len - hdr_data_size);

	std::string payload_only_str;
	int num_symbols_last_fec_grp = 0;

	// if just one fec grp, then return the payload from that fec (excluding the hdr) //
	if(num_fec_grps == 1)
		return buf.substr(hdr_data_size, payload_only_size);

	int payload_first_fec_grp = fec_K - hdr_data_size;

	// figure out the symbols in the last fec grp //
	if(len % fec_N != 0)
		num_symbols_last_fec_grp = (payload_only_size - payload_first_fec_grp) % fec_K;

	// printf("num_fec_grps: %d, num_symbols_last_fec_grp: %d\n", num_fec_grps, num_symbols_last_fec_grp);

	int num_data_bytes = fec_K;
	for(int i = 0; i < num_fec_grps; i++)
	{
		if(i == 0)
		{
			payload_only_str.append(buf.substr(hdr_data_size, fec_K - hdr_data_size));
			// printf("i: %d, size: %d\n", i, payload_only_str.size());
			continue;
		}
		else if(i == num_fec_grps - 1 && num_symbols_last_fec_grp != 0)
			num_data_bytes = num_symbols_last_fec_grp;

		std::string payload = buf.substr(i * fec_N, num_data_bytes);
		payload_only_str.append(payload);	
	}

	return payload_only_str; 
}

std::string
	getPayloadFECOnly(std::string buf, int fec_N, int fec_K, int hdr_data_size)
{
	if(fec_N == 0)
		return "";

	int len = buf.size();
	if(len <= fec_N)			// just one fec group, and the hdr-fec is combined with payload-fec thr.. 
		return "";

	int start_offset = fec_N; 		// since the hdr is combined with the first few payload bytes to make up 'fec_K' size

	int len_payload_fec = len - fec_N;

	int num_fec_grps = ceil((float) len_payload_fec/fec_N);

	int num_symbols_last_fec_grp = fec_K;
	if(len_payload_fec % fec_N != 0)
		num_symbols_last_fec_grp = len_payload_fec%fec_N - (fec_N - fec_K);

	std::string payload_fec_only_str; 
	for(int i = 0; i < num_fec_grps; i++)
	{
		int num_data_bytes = fec_K;
		if(i == num_fec_grps - 1)
			num_data_bytes = num_symbols_last_fec_grp;

		int offset = start_offset + i * fec_N + num_data_bytes; 
	//printf("i: %d, offset: %d num_data_bytes: %d\n", i, offset, num_data_bytes);
		std::string payload = buf.substr(offset, (fec_N - fec_K));
		payload_fec_only_str.append(payload);
	}

	return payload_fec_only_str; 
}

std::string
	getHeaderOnly(std::string buf, int fec_N, int fec_K, int hdr_data_size)
{
	return buf.substr(0, hdr_data_size);
}

std::string
	getHeaderFECOnly(std::string buf, int fec_N, int fec_K, int hdr_data_size)
{
	// hdr-data is combined with the payload (first fec group) and then encoded with FEC //
	int fec_only_size = fec_N - fec_K;
	if(buf.size() <= fec_N)
		return buf.substr(buf.size() - fec_only_size, fec_only_size);
	return buf.substr(fec_K, fec_only_size);
}

int
	interleave_bucket(int len, int num_sp_bits, int n_subcarriers, int& bits_mapped_already, int subcarriers, 
	int& offset, int bps, int rounds, 
	int* sorted_sc_indices, int* interleaved_bit_pos,
	int* final_pos, unsigned char* tracker, int scheme)
{
	int s_len = len * 8;

	//printf("interleave_bucket: len:: %d, s_len: %d, bps: %d, num_sp_bits: %d, bits_mapped_already: %d, rounds: %d, offset: %d, n_subcarriers: %d\n", len, s_len, bps, num_sp_bits, bits_mapped_already, rounds, offset, n_subcarriers);

	int Ncbps = n_subcarriers * bps;
	int Nbpsc = bps;
	assert(num_sp_bits % Ncbps == 0);

	int s = (Nbpsc/2 > 1) ? Nbpsc/2 : 1;
	int blocks = ceil((float)num_sp_bits/Ncbps);

	// a lame ad-hoc attempt to keep C near 16 (as in standard interleaver) // 
	int C = 1;
	for(int i = 6; i >=2; i--)
	{
		if(Ncbps % i == 0)
		{
			C = Ncbps/i;
			break;
		}
	}

	// random interleaver uses C = 16 //    
	if(scheme == 0)
		C = 16;

	//printf("Ncbps: %d C: %d, blocks: %d, s: %d\n", Ncbps, C, blocks, s); 
	assert(Ncbps % C == 0);


	int *arr_j = (int*) malloc(sizeof(int) * Ncbps * blocks); 
	memset(arr_j, 0, Ncbps * sizeof(int) * blocks);

	// only does the first step of the permuation //
	for(int b = 0; b < blocks; b++)
	{
		for(int k = 0; k < Ncbps; k++)
		{
			int i = (Ncbps/C)*(k%C) + floor((float)k/C);
			if(scheme != 0)
				arr_j[b*Ncbps + k] = i;
			else
				arr_j[b*Ncbps + k] = s * floor((float) i/s) + (int) (i + Ncbps - floor((float) C * i/Ncbps)) % s;
		}
	}
	/*
	int *arr_i = (int*) malloc(sizeof(int) * Ncbps * blocks);
	memset(arr_i, 0, Ncbps * sizeof(int) * blocks);

// only does the first step of the permuation //
for(int b = 0; b < blocks; b++)
for(int k = 0; k < Ncbps; k++)
arr_i[b*Ncbps + k] = (Ncbps/C)*(k%C) + floor((float)k/C);
*/

for(int b = 0; b < blocks; b++)
{
	for(int k = 0; k < Ncbps; k++)
	{
		int orig_pos = k + offset;
		int interleaved_pos = arr_j[k] + offset;
		interleaved_bit_pos[orig_pos] = interleaved_pos;

		// will be used to memcpy the interleaved bits into their final mappings //	
		int p = k + b * Ncbps;
		int sym_index = p/bps;
		int rem = p%bps;
		int sc_index = sym_index % n_subcarriers;
		int actual_sc_index = sorted_sc_indices[sc_index];
		int pos = (p/Ncbps)*(subcarriers * bps) + actual_sc_index * bps + rem;


		final_pos[p + bits_mapped_already] = pos;
		//printf("orig: %d int: %d index: %d pos: %d\n", orig_pos, interleaved_pos, p+bits_mapped_already, pos);
		assert(tracker[pos] == 0);
		tracker[pos] = 1;
		assert(pos < s_len);
	}
	offset += Ncbps;
}

free(arr_j);
bits_mapped_already += num_sp_bits;

return num_sp_bits;
}

/* used only for debug, prints out where the header and payload were sent out */
void
	verifySubcarrierMapping(int *interleaved_pos, int *final_pos, int len, int subcarriers, int bps, int fec_N, int fec_K)
{
	// printf("verifySubcarrierMapping ... \n");
	int hdr_size = 24;

	int *sc_mapping = (int*) malloc(sizeof(int) * subcarriers);
	memset(sc_mapping, 0, sizeof(int) * subcarriers);

	// find subcarriers used by the header //
	for(int i = 0; i < hdr_size * 8; i++)
	{
		int mapped_pos = final_pos[interleaved_pos[i]];
		int sym_index = mapped_pos/bps;
		int sc_index = sym_index % subcarriers; 

		sc_mapping[sc_index] = 1;
	}

	int count = 0;
	// printf("subcarriers for hdr used are:: \n");
	for(int i = 0; i < subcarriers; i++)
	{
		if(sc_mapping[i] == 1)
		{
			count++;
			// printf("%d ", i);
		}
	}
	// printf("\n");

	// printf("total used: %d\n", count); 

	/*
	// find subcarriers used by the payload //
	int offset = hdr_size * 8;
	int payload_size = 980;

	memset(sc_mapping, 0, subcarriers);
	for(int i = 0; i <  payload_size * 8; i++)
	{
	int mapped_pos = final_pos[interleaved_pos[i + offset]];
	int sym_index = mapped_pos/bps;
	int sc_index = sym_index % subcarriers;

	sc_mapping[sc_index] = 1;
	}verifySubcarrierMapping

	count = 0;

	printf("subcarriers for payload used are:: \n");
	for(int i = 0; i < subcarriers; i++)
	{
	if(sc_mapping[i] == 1)
	{
	count++;
	printf("%d ", i);
	}
	}
	printf("\n");

*/
/*
	// subcarriers used by the payload-fec //
	int payload_size = 980;
	int total_data = payload_size + hdr_size;
	int num_fec_grps = ceil((float) total_data/fec_K);
	memset(sc_mapping, 0, subcarriers);

	int fec_N_bits = fec_N * 8;
	int fec_K_bits = fec_K * 8;

	for(int i = 0; i < subcarriers; i++)
	sc_mapping[i] = 0;

	int num_data_bits_last_grp = fec_K_bits;
	if(len % fec_N != 0)
	num_data_bits_last_grp = (len % fec_N - (fec_N - fec_K)) * 8;

	printf("num_fec_grps: %d\n", num_fec_grps);

	for(int i = 0; i < num_fec_grps; i++)
	{
	int num_data_bits_grp = fec_K_bits;
	if(i == num_fec_grps - 1)
	num_data_bits_grp = num_data_bits_last_grp;


	printf("for fec-group: %d :::::::::: \n", i);
	int s_count = 0;
	for(int j = 0; j < (fec_N_bits - fec_K_bits); j++)
	{
	int pos = (i * fec_N_bits) + num_data_bits_grp + j;
	int mapped_pos = final_pos[interleaved_pos[pos]];
	int sym_index = mapped_pos/bps;
	int sc_index = sym_index % subcarriers;

	sc_mapping[sc_index]++; 
	if(sc_index >=136 && sc_index <=184)
	{
	printf("bit %d -> sc %d\n", pos, sc_index);
	s_count++;
	}

	}
	printf("fec_grp: %d has %d bits mapped to bad ones!\n", i, s_count);
	}


	printf("subcarriers for fec are:: \n");
	for(int i = 0; i < subcarriers; i++)
	{
	if(sc_mapping[i] > 0)
	{
	count++;
	printf("sc: %d contains %d bits\n", i, sc_mapping[i]);
	}
	}
*/
	// printf("\n");
	free(sc_mapping);

}


/* classify 4 priorities, the bytes in the priority buckets will depend on the desired scheme */
std::string
	interleave_mapping_best(std::string buf, int subcarriers, int bps, int fec_N, int fec_K, int scheme, int& n_mac_parity_bits, int hint_flag)
{
	int NBYTES[4];	// 4 priority buckets

	std::string priority_str;
	int num_buckets = 4;
	

	int hdr_data_size = 24;
	std::string hdr = getHeaderOnly(buf, fec_N, fec_K, hdr_data_size);
	std::string payload = getPayloadOnly(buf, fec_N, fec_K, hdr_data_size);
	std::string hdr_fec = getHeaderFECOnly(buf, fec_N, fec_K, hdr_data_size);			// will be padded by few payload-fec bytes
	std::string payload_fec = getPayloadFECOnly(buf, fec_N, fec_K, hdr_data_size);		// remainder of the payload-fec fec
	

#ifdef DEBUG

	printf("hdr-fec size: %d\n", hdr_fec.size()); fflush(stdout);
	printf("hdr: %d\n", hdr.size());
	printf("payload: %d\n", payload.size());
	printf("payload-fec: %d\n", payload_fec.size());
#endif

#ifdef DEBUG
	printf("interleave_mapping: size: %d, subcarriers: %d, bps: %d, fec_N: %d, fec_K: %d, scheme: %d\n", 
		buf.size(), subcarriers, bps, fec_N, fec_K, scheme);


	printf("\nbuf :: lt: %d\n", buf.size());
	for(int i = 0; i < buf.size(); i++)
		printf("%d ", buf[i]);

	printf("\n\nhdr only:: lt: %d\n", hdr.size());
	for(int i = 0; i < hdr.size(); i++)
		printf("%d ", hdr[i]);

	printf("\n\npayload only:: lt: %d\n", payload.size());
	for(int i = 0; i < payload.size(); i++)
		printf("%d ", payload[i]);

	printf("\n\nhdr-fec only:: lt: %d\n", hdr_fec.size());
	for(int i = 0; i < hdr_fec.size(); i++)
		printf("%d ", hdr_fec[i]);

	printf("\n\npayload-fec only:: lt: %d\n", payload_fec.size());
	for(int i = 0; i < payload_fec.size(); i++)
		printf("%d ", payload_fec[i]);
#endif
	switch(scheme)
	{
		case 0:   priority_str = buf; 						// random 
		NBYTES[0] = priority_str.size();
		num_buckets = 1;
		break;

		case 1:	  priority_str = hdr + payload + hdr_fec + payload_fec;		 // header-data mapping (acc to the writeup)
		NBYTES[0] = hdr.size();					 // hdr-data //
		NBYTES[1] = payload.size();					 // payload-data //
		NBYTES[2] = hdr_fec.size();					 // hdr-fec //
		NBYTES[3] = payload_fec.size();				 // payload-fec //
		num_buckets = 4;

		if(payload_fec.size() == 0)
			num_buckets--;
		break;

		case 2:	  priority_str = payload + hdr + hdr_fec + payload_fec;       
		NBYTES[0] = payload.size();
		NBYTES[1] = hdr.size();
		NBYTES[2] = hdr_fec.size();
		NBYTES[3] = payload_fec.size();
		num_buckets = 4;

		if(payload_fec.size() == 0)
			num_buckets--;
		break;

		case 3:   priority_str = hdr + payload + hdr_fec + payload_fec;         
		NBYTES[0] = hdr.size() + payload.size();			 // hdr-data + payload-data //
		NBYTES[1] = hdr_fec.size();					 // hdr-fec //
		NBYTES[2] = payload_fec.size();				 // payload-fec //
		num_buckets = 3;

		if(payload_fec.size() == 0)
			num_buckets--;
		break;
		case 4:   priority_str = hdr + payload + hdr_fec + payload_fec;         // XXX: data
		NBYTES[0] = hdr.size() + payload.size();			// hdr-data + payload-data //
		NBYTES[1] = hdr_fec.size() + payload_fec.size();		// hdr-fec + payload-fec //
		num_buckets = 2;
		break;

		case 5:   priority_str = hdr + hdr_fec + payload + payload_fec;         // XXX: header
		NBYTES[0] = hdr.size() + hdr_fec.size();			// header-data + hdr-fec first //
		NBYTES[1] = payload.size() + payload_fec.size();		// payload-data + payload-fec //
		num_buckets = 2;
		break;
		case 6:   priority_str = hdr + hdr_fec + payload + payload_fec;         // hdr data first //
		NBYTES[0] = hdr.size();					// hdr-data //
		NBYTES[1] = hdr_fec.size();					// hdr-fec //
		NBYTES[2] = payload.size();					// payload-data //
		NBYTES[3] = payload_fec.size();				// payload-fec //
		num_buckets = 4;

		if(payload_fec.size() == 0)
			num_buckets--;
		break;
		case 7:   assert(false);
	}

	//assert(hdr.size() + payload.size() + hdr_fec.size() + payload_fec.size() == buf.size()); 

#ifdef DEBUG
	printf("bucket sizes:: \n");
	for(int i = 0; i < num_buckets; i++)
	{
		printf("%d ", NBYTES[i]);
	}
	printf("\n");
#endif

	// read the csi and sort the subcarriers only once //
	static bool initialized = false;
	static float *csi = (float*) malloc(sizeof(float) * subcarriers);
	static int *sorted_sc_indices = (int*) malloc(sizeof(int) * subcarriers);
	if(!initialized)
	{
		initialized = true;
		// readCSISubcarriers(csi, subcarriers);
		memcpy(csi, g_csi, subcarriers * sizeof(float));
		
		
		if(scheme > 0)
		{
			float *sorted_csi = (float*) malloc(sizeof(float) * subcarriers);
			memcpy(sorted_csi, csi, sizeof(float) * subcarriers);
			sortSubcarriersByCSI(sorted_sc_indices, sorted_csi, subcarriers);
			free(sorted_csi);
		}
		else
		{
				// for scheme=0, we need not sort the subcarriers //
			for(int i = 0; i < subcarriers; i++)
				sorted_sc_indices[i] = i;
		}
	}

	const unsigned char *data = (const unsigned char*) priority_str.data();
	size_t len = priority_str.size();
	int s_len = len * 8;

	int num_symbols = ceil((float) s_len/bps);
	int rounds = ceil((float) num_symbols/subcarriers);
	
	
	/* CASE A: possible padding to fill the subcarriers completely */
	// printf("len: %d num_symbols: %d, rounds: %d\n", len, num_symbols, rounds);  
	if(num_symbols % subcarriers != 0)
	{
		int pad_symbols = subcarriers - (num_symbols % subcarriers);		// pad symbols //
		num_symbols += pad_symbols;
		s_len = num_symbols *bps;
		rounds = num_symbols/subcarriers;
		len = s_len/8;

		int pad_bytes = len - priority_str.size();				// pad bytes //
		priority_str.append(pad_bytes, '0');
		NBYTES[num_buckets - 1] += pad_bytes;					// put the extra pad bytes to the last bucket //
		printf("padding: %d, pad_bytes: %d\n", pad_symbols, pad_bytes); 
	}

	assert(num_symbols % subcarriers == 0);
	
	unsigned char tracker[sizeof(unsigned char) * s_len];
	memset(tracker, 0, sizeof(unsigned char) * s_len);

	int *interleaved_bit_pos = (int*) malloc(sizeof(int) * s_len);
	memset(interleaved_bit_pos, 0, sizeof(int) * s_len);

	int *final_pos = (int*) malloc(sizeof(int) * s_len);
	memset(final_pos, 0, sizeof(int) * s_len);
	
	int bits_mapped_already = 0;
	int _pad = 0;
	int subcarriers_used_already = 0;
	int offset = 0;
	int num_sp_bits = 0;
	for(int i = 0; i < num_buckets; i++)
	{
		if(NBYTES[i] > 0)
		{
			int rem_bits = s_len - bits_mapped_already;
			// printf("rem_bits: %d, NBYTES[%d](bits): %d\n", rem_bits, i, NBYTES[i] * 8); 
			if(rem_bits <= NBYTES[i] * 8)
				num_sp_bits = rem_bits;
			else
				num_sp_bits = NBYTES[i] * 8 - _pad;			// padding from the previous bucket might have spilled over

			if(num_sp_bits <= 0)
				continue;


		//printf("-- num_sp_bits: %d\n", num_sp_bits);
			int num_sp_symbols = ceil((float)num_sp_bits/bps);
			int n_subcarriers = ceil((float)num_sp_symbols/rounds);   	

		// (possibly) pad 'num_sp_bits' such that they occupy complete 'rounds' using 'n_best_subcarriers' //
			int n_bits = (n_subcarriers * rounds * bps > num_sp_bits) ? n_subcarriers * rounds * bps: num_sp_bits; 

			if(n_bits > num_sp_bits)
				_pad = n_bits - num_sp_bits;

			assert(n_bits%bps == 0);
			// printf("\nn_bits: %d rem_bits: %d n_subcarriers: %d num_sp_symbols: %d NBYTES: %d num_sp_bits: %d, pad: %d\n", n_bits, rem_bits, n_subcarriers, num_sp_symbols, NBYTES[i], num_sp_bits, _pad); fflush(stdout);	
			interleave_bucket(len, n_bits, n_subcarriers, 
				bits_mapped_already, subcarriers, 
				offset, bps, rounds, sorted_sc_indices + subcarriers_used_already, 
				interleaved_bit_pos, final_pos, tracker, scheme);	  

			subcarriers_used_already += n_subcarriers;
		}
	}

	// printf("buf.size: %d, s_len: %d, total_symbols: %d, bps: %d, subcarriers: %d\n", len, s_len, num_symbols, bps, subcarriers);

#ifdef DEBUG
	printf("interleaver position map\n");
	for(int i = 0; i < s_len; i++)
		printf("%d -> %d -> %d\n", i, interleaved_bit_pos[i], final_pos[interleaved_bit_pos[i]]);
#endif

	std::string priority_s = getBitString((unsigned char*) priority_str.data(), len);
	unsigned char *final_bits = (unsigned char*) malloc(sizeof(unsigned char) * s_len);
	
	
	/**
	 * do switching
	 **/
	if(MODE_HEADER_MAPPING == HEADER_SWITCH) {
		int pos_a_st;
		// if(scheme != 1) pos_a_st = fec_N * 8;
		// else pos_a_st = 0;
		pos_a_st = fec_N * 8;
		// int pos_b_st = s_len - MAX_ITR * SW_N * 8;
		for(int i = 0; i < MAX_ITR * SW_N * 8; i ++) {
			if(!SW_MAP[i/8/SW_N]) continue;
			char tmp = priority_s[pos_a_st + i];
			// priority_s[pos_a_st + i] = priority_s[pos_b_st + i];
			priority_s[pos_a_st + i] = priority_s[priority_s.size() - 1 - i];
			priority_s[priority_s.size() - 1 - i] = tmp;
		}
	}
	///////////
	

	// the orig->final map and final->orig map (used only for mac-fec //
	int *orig_map = (int*) malloc(sizeof(int) * s_len);
	memset(orig_map, 0, sizeof(int) * s_len);

	int *final_map = (int*) malloc(sizeof(int) * s_len);
	memset(final_map, 0, sizeof(int) * s_len);

	for(int i = 0; i < s_len; i++)
	{
		int pos = final_pos[interleaved_bit_pos[i]];
		final_bits[pos] = priority_s[i];
		//printf("%d ", pos); fflush(stdout);

		orig_map[pos] = i;
		final_map[i] = pos;
	}

#ifdef DEBUG
	printf("final_s: \n");
	for(int i = 0; i < s_len; i++)
		printf("%c-%d", final_bits[i], i);
	printf("\n");

#endif


	verifySubcarrierMapping(interleaved_bit_pos, final_pos, len, subcarriers, bps, fec_N, fec_K);
	
	std::string tmp_s((const char*) final_bits, s_len);    

	if(mac_fec_flag)		// mac-fec flag
	{
		std::string parity_bits_str;
		std::string orig_s = getBitString((const unsigned char*) buf.data(), buf.size());

		encodeMACLayer_FEC(tmp_s, orig_s, parity_bits_str, final_map, orig_map, subcarriers, 
			fec_N, fec_K, bps, n_mac_parity_bits, orig_s.size(),
			0, 0, 0, hint_flag);


#ifdef DEBUG
		printf("sender: parity (before padding) .. \n");
		for(int i = 0; i < n_mac_parity_bits; i++)
			printf("%c", parity_bits_str[i]);
		printf("\n");
#endif

		// pad and interleave parity //
		int parity_one_round = subcarriers * bps;
		int pad_parity = 0;
		if(n_mac_parity_bits % parity_one_round != 0)
		{
			pad_parity = parity_one_round - (n_mac_parity_bits % parity_one_round);
			n_mac_parity_bits += pad_parity;
		}

		parity_bits_str.append(pad_parity, '0');
		
		

#ifdef DEBUG

		printf("sender: parity (after padding) .. \n");
		for(int i = 0; i < n_mac_parity_bits; i++)
			printf("%c", parity_bits_str[i]);
		printf("\n");
#endif

		unsigned char *final_parity_bits = (unsigned char*) malloc(sizeof(unsigned char) * n_mac_parity_bits + 1);
		memset(final_parity_bits, 0, sizeof(unsigned char) * n_mac_parity_bits + 1);

		int interleaved_parity_pos[subcarriers * bps];
		initInterleavedParityPos(interleaved_parity_pos, bps, subcarriers);

		for(int i = 0; i < n_mac_parity_bits; i++)
		{
			int index = i % parity_one_round;
			int blocks_mapped = i / parity_one_round;
			int new_pos = interleaved_parity_pos[index] + blocks_mapped * parity_one_round;
			assert(new_pos < ((blocks_mapped + 1) * parity_one_round));
			final_parity_bits[new_pos] = parity_bits_str[i];
		}

		parity_bits_str.assign((const char*) final_parity_bits);	
		n_mac_parity_bits = parity_bits_str.size();
#ifdef DEBUG
		printf("sender: interleaved parity .. \n");
		for(int i = 0; i < n_mac_parity_bits; i++)
			printf("%c", parity_bits_str[i]); 
#endif
		// printf("n_mac_parity_bits: %d, pad_parity: %d\n", n_mac_parity_bits, pad_parity); 

		tmp_s.append(parity_bits_str);
	// update the s_len and the len params //	
		s_len = tmp_s.size();	
		len = s_len/8;
	}
	

	unsigned char *final_data = (unsigned char*) malloc(sizeof(unsigned char) * len);
	memset(final_data, 0, sizeof(unsigned char) * len);
	std::string final_mapped_str = getMsg(tmp_s, final_data);
	return final_mapped_str;
}



std::string
	reconstruct_payload_fec(std::string payload_only, std::string fec_only, int fec_N, int fec_K)
{
	if(fec_only.size() == 0)
		return "";

	int len_payload_fec = payload_only.size() + fec_only.size();
	int num_fec_grps = ceil((float) len_payload_fec/fec_N);

	// printf("payload_only.size: %d, fec_only.size: %d, num_fec_grps: %d\n",
	// 	payload_only.size(), fec_only.size(), num_fec_grps);

	std::string comb_str; 
	for(int i = 0; i < num_fec_grps; i++)
	{
		std::string payload = payload_only.substr(i * fec_K, fec_K);
		comb_str.append(payload);
		std::string fec = fec_only.substr(i*(fec_N - fec_K), fec_N - fec_K);
		comb_str.append(fec);
	}
	return comb_str; 
}

int
	deinterleave_bucket(int len, int num_sp_bits, int n_subcarriers, int& bits_demapped_already, int subcarriers, 
	int& offset, int bps, int rounds, 
	int* sorted_sc_indices, int* orig_bit_pos,
	int* final_pos, unsigned char* tracker, int scheme)
{
	int s_len = len * 8;

	//printf("deinterleave_bucket: len:: %d, s_len: %d, bps: %d, num_sp_bits: %d, bits_demapped_already: %d, rounds: %d, offset: %d, n_subcarriers: %d\n", len, s_len, bps, num_sp_bits, bits_demapped_already, rounds, offset, n_subcarriers);

	int Ncbps = n_subcarriers * bps;
	int Nbpsc = bps;
	assert(num_sp_bits % Ncbps == 0);

	int s = (Nbpsc/2 > 1) ? Nbpsc/2 : 1;
	int blocks = ceil((float)num_sp_bits/Ncbps);

	int C = 1;
	for(int i = 6; i >=2; i--)
	{
		if(Ncbps % i == 0)
		{
			C = Ncbps/i;
			break;
		}
	}
	// printf("C: %d, blocks: %d\n", C, blocks); 

	assert(Ncbps % C == 0);

	if(scheme == 0)
		C = 16;

	int *arr_k = (int*) malloc(sizeof(int) * Ncbps * blocks); 
	memset(arr_k, 0, sizeof(int) * Ncbps * blocks);

	for(int b = 0; b < blocks; b++)
	{
		for(int j = 0; j < Ncbps; j++)
		{
			if(scheme != 0)
				arr_k[b*Ncbps + j] = C * j - (Ncbps - 1) * floor((float) C * j/Ncbps);
			else
			{
				int i = s * floor((float) j/s) + (int)(j + floor((float) C * j/Ncbps)) % s;
				arr_k[b*Ncbps + j] = C * i - (Ncbps - 1) * floor((float) C * i/Ncbps);
			}
		}
	}

	for(int b = 0; b < blocks; b++)
	{
		for(int k = 0; k < Ncbps; k++)
		{
			int interleaved_pos = k + offset;
			int orig_pos = arr_k[k] + offset;

			orig_bit_pos[interleaved_pos] = orig_pos;


	// will be used to memcpy the interleaved bits into their final mappings //
			int p = k + b * Ncbps;
			int sym_index = p/bps;
			int rem = p%bps;
			int sc_index = sym_index % n_subcarriers;
			int actual_sc_index = sorted_sc_indices[sc_index];
			int pos = (p/Ncbps)*(subcarriers * bps) + actual_sc_index * bps + rem;

	//printf("p: %d sc_index: %d, actual_sc: %d, sym_index: %d, pos: %d\n", p+ bits_demapped_already, sc_index, actual_sc_index, sym_index, pos);

			final_pos[pos] = p + bits_demapped_already;
	//printf("pos: %d\n", pos); 
			assert(tracker[pos] == 0);
			tracker[pos] = 1;
			assert(pos < s_len);
		}
		offset += Ncbps;
	}

	free(arr_k);
	bits_demapped_already += num_sp_bits;
	return num_sp_bits;
}

std::string
	deinterleave_mapping_best(std::string buf, int subcarriers, int bps, int fec_N, int fec_K, int scheme, int expectedPayloadWithHdrLen, 
	std::string& parity_bits_str, int mac_parity_added, 
	std::map<int, split>& fec_split_map, std::vector<int>& sorted_orig_pos, 
	std::string test_buf, std::string& test_parity_bits_str, int hint_flag
#ifdef PARTIAL_TPUT
	, int *g_final_pos_map
#endif
	)
{
	int NBYTES[4];	// 4 priority buckets
	int num_buckets = 4;
	int hdr_data_size = 24;
	int hdr_fec = fec_N - fec_K;
	// printf("deinterleave_mapping: size: %d, subcarriers: %d, bps: %d, fec_N: %d, fec_K: %d, scheme: %d\n", 
	// 	buf.size(), subcarriers, bps, fec_N, fec_K, scheme); 
	// form the buckets //
	int num_fec_grp = (fec_K == 0) ? 0:ceil((float)expectedPayloadWithHdrLen/fec_K);
	int payload_fec_size = (num_fec_grp-1) * (fec_N - fec_K);				// first fec goes with hdr
	switch(scheme)
	{
		case 0:   NBYTES[0] = expectedPayloadWithHdrLen + (fec_N - fec_K) +  payload_fec_size;
		num_buckets = 1;
		break;

		case 1:	  NBYTES[0] = hdr_data_size;
		NBYTES[1] = expectedPayloadWithHdrLen - hdr_data_size;
		NBYTES[2] = (fec_N - fec_K);
		NBYTES[3] = payload_fec_size;
		num_buckets = 4;
		if(payload_fec_size == 0)
			num_buckets--;

		break;

		case 2:	  NBYTES[0] = expectedPayloadWithHdrLen - hdr_data_size;
		NBYTES[1] = hdr_data_size;
		NBYTES[2] = (fec_N - fec_K);
		NBYTES[3] = payload_fec_size;
		num_buckets = 4;

		if(payload_fec_size == 0)
			num_buckets--;
		break;

		case 3:   NBYTES[0] = expectedPayloadWithHdrLen;
		NBYTES[1] = (fec_N - fec_K); 
		NBYTES[2] = payload_fec_size;
		num_buckets = 3;

		if(payload_fec_size == 0)
			num_buckets--;
		break;

		case 100:   NBYTES[0] = hdr_data_size;				// old one!
		NBYTES[1] = (expectedPayloadWithHdrLen - hdr_data_size) + (fec_N - fec_K) + payload_fec_size;
		num_buckets = 2;
		break;

		case 4:   NBYTES[0] = expectedPayloadWithHdrLen;
		NBYTES[1] = (fec_N - fec_K) + payload_fec_size;
		num_buckets = 2;
		break;

		case 5:   NBYTES[0] = hdr_data_size + (fec_N - fec_K);
		NBYTES[1] = (expectedPayloadWithHdrLen - hdr_data_size) + payload_fec_size;
		num_buckets = 2;
		break;

		case 6:   NBYTES[0] = hdr_data_size;
		NBYTES[1] = (fec_N - fec_K);
		NBYTES[2] = expectedPayloadWithHdrLen - hdr_data_size;
		NBYTES[3] = payload_fec_size;
		num_buckets = 4;
		if(payload_fec_size == 0)
			num_buckets--;
		break;

		case 7:   assert(false);
	}

#ifdef DEBUG 
	printf("hdr size: %d\n", hdr_data_size);
	printf("payload size: %d\n", expectedPayloadWithHdrLen - hdr_data_size);
	printf("hdr-fec size: %d\n", (fec_N - fec_K));
	printf("payload-fec size: %d\n", payload_fec_size);
	printf("expectedPayloadWithHdrLen: %d\n", expectedPayloadWithHdrLen);
	printf("buf.size: %d\n", buf.size());
	printf("num_buckets: %d\n", num_buckets); 
#endif

	if(mac_fec_flag)
	{
		std::string orig_s = getBits((const unsigned char*) buf.data(), buf.size(), 8);
		parity_bits_str = orig_s.substr(orig_s.size() - mac_parity_added, mac_parity_added);

		//std::string test_orig_s = getBits((const unsigned char*) test_buf.data(), test_buf.size(), 8);
		//test_parity_bits_str = test_orig_s.substr(orig_s.size() - mac_parity_added, mac_parity_added);

		// printf("mac_parity_bits: %d\n", mac_parity_added); 
		buf = buf.substr(0, buf.size() - mac_parity_added/8);
		// printf("buf.size: %d\n", buf.size());
	}

	// determine the padding that might be used by the interleaver to fill the subcarriers //
	int padding = abs((int)(buf.size() - (expectedPayloadWithHdrLen + (fec_N - fec_K) + payload_fec_size)));

	// read the csi and sort the subcarriers only once //
	static bool initialized = false;
	static float *csi = (float*) malloc(sizeof(float) * subcarriers);
	static int *sorted_sc_indices = (int*) malloc(sizeof(int) * subcarriers);
	if(!initialized)
	{
		initialized = true;
		// readCSISubcarriers(csi, subcarriers);
		memcpy(csi, g_csi, subcarriers * sizeof(float));
		
		if(scheme > 0)
		{
			float *sorted_csi = (float*) malloc(sizeof(float) * subcarriers);
			memcpy(sorted_csi, csi, sizeof(float) * subcarriers);
			sortSubcarriersByCSI(sorted_sc_indices, sorted_csi, subcarriers);
			free(sorted_csi);
		}
		else
		{
				// for scheme=0, we need not sort the subcarriers //
			for(int i = 0; i < subcarriers; i++)
				sorted_sc_indices[i] = i;
		}
	}

	size_t len = buf.size();
	int s_len = len * 8;
	// printf("s_len: %d, len: %d\n", s_len, len); 

	int num_symbols = ceil((float) s_len/bps);
	int rounds = ceil((float) num_symbols/subcarriers);
	assert(num_symbols % subcarriers == 0);

	// similar to the interleaver, pretend the 'padding' is added to the last bucket, filter it out later //
	NBYTES[num_buckets - 1] += padding;

	// tracker helps in debugging //
	unsigned char tracker[sizeof(unsigned char) * s_len];
	memset(tracker, 0, sizeof(unsigned char) * s_len);

	int *orig_bit_pos = (int*) malloc(sizeof(int) * s_len);
	memset(orig_bit_pos, 0, sizeof(int) * s_len);

	int *final_pos = (int*) malloc(sizeof(int) * s_len);
	memset(final_pos, 0, sizeof(int) * s_len);

	assert(num_symbols % subcarriers == 0);

	int bits_demapped_already = 0;
	int subcarriers_used_already = 0;
	int offset = 0;
	int _pad = 0; 
	int num_sp_bits = 0;
	for(int i = 0; i < num_buckets; i++)
	{
		if(NBYTES[i] > 0)
		{
			int rem_bits = s_len - bits_demapped_already;
			if(rem_bits <= NBYTES[i] * 8)
				num_sp_bits = rem_bits;
			else
				num_sp_bits = NBYTES[i] * 8 - _pad;                     // padding from the previous bucket might have spilled over

			if(num_sp_bits <= 0)
				continue;

			int num_sp_symbols = ceil((float)num_sp_bits/bps);
			int n_subcarriers = ceil((float)num_sp_symbols/rounds);   

		//printf("rem_bits: %d\n", rem_bits); fflush(stdout); 	

		// (possibly) pad 'num_sp_bits' such that they occupy complete 'rounds' using 'n_best_subcarriers' //
			int n_bits = (n_subcarriers * rounds * bps > num_sp_bits) ? n_subcarriers * rounds * bps : num_sp_bits; 

			if(n_bits > num_sp_bits)
				_pad = n_bits - num_sp_bits;

			assert(n_bits%bps == 0);	
			deinterleave_bucket(len, n_bits, n_subcarriers, 
				bits_demapped_already, subcarriers, 
				offset, bps, rounds, sorted_sc_indices + subcarriers_used_already, 
				orig_bit_pos, final_pos, tracker, scheme);	  

			subcarriers_used_already += n_subcarriers;
		}
	}

	int *final_pos_map = (int*) malloc(sizeof(int) * s_len);
	memset(final_pos_map, 0, sizeof(int) * s_len);

	int *orig_pos_map = (int*) malloc(sizeof(int) * s_len);
	memset(orig_pos_map, 0, sizeof(int) * s_len);

	// for mac-fec //
	for(int i = 0; i < s_len; i++)
	{
		int orig_pos = orig_bit_pos[final_pos[i]];
		final_pos_map[orig_pos] = i;
		orig_pos_map[i] = orig_pos;
	}	

#ifdef PARTIAL_TPUT
	memcpy(g_final_pos_map, final_pos_map, sizeof(int) * s_len);
#endif

#ifdef DEBUG
	printf("position map (deinterleaver)\n");
	for(int i = 0; i < s_len; i++)
		printf("%d -> %d -> %d\n", i, final_pos[i], orig_bit_pos[final_pos[i]]);
#endif

	// form the priority string //
	const unsigned char *data = (const unsigned char*) buf.data();
	std::string orig_s = getBitString(data, len);
	unsigned char* priority_bits = (unsigned char*) malloc(sizeof(unsigned char) * s_len);
	for(int i = 0; i < s_len; i++)
		priority_bits[orig_bit_pos[final_pos[i]]] = orig_s[i];

// filter out the padding now //
	len = len - padding;
	s_len = len * 8; 
	
	
	
	
	/**
	 * do switching
	 **/
	if(MODE_HEADER_MAPPING == HEADER_SWITCH) {
		int pos_a_st;
		// if(scheme != 1) pos_a_st = fec_N * 8;
		// else pos_a_st = 0;
		pos_a_st = fec_N * 8;
		// int pos_b_st = s_len - MAX_ITR * SW_N * 8;
		// for(int i = 0; i < MAX_ITR * SW_N * 8; i ++) {
		// 	if(!SW_MAP[i]) continue;
		// 	char tmp = priority_bits[pos_a_st + i];
		// 	priority_bits[pos_a_st + i] = priority_bits[pos_b_st + i];
		// 	priority_bits[pos_b_st + i] = tmp;
		// }
		for(int i = 0; i < MAX_ITR * SW_N * 8; i ++) {
			if(!SW_MAP[i/8/SW_N]) continue;
			char tmp = priority_bits[pos_a_st + i];
			priority_bits[pos_a_st + i] = priority_bits[s_len - 1 - i];
			priority_bits[s_len - 1 - i] = tmp;
		}
	}
	///////////
	
	

	// convert from the priority string back to the desired original string //
	std::string priority_s((const char*) priority_bits, s_len); 
	unsigned char *priority_data = (unsigned char*) malloc(sizeof(unsigned char) * len);
	memset(priority_data, 0, sizeof(unsigned char) * len);
	std::string priority_str = getMsg(priority_s, priority_data);
	NBYTES[num_buckets - 1] -= padding;

	std::string hdr_payload_only = ""; 		// hdr and payload together in 1 bucket
	std::string hdr_only = "";
	std::string hdr_fec_only = "";
	std::string payload_only = "";
	std::string payload_fec_only = "";
	std::string payload_fec = "";
	std::string others = ""; 			// payload + hdr_fec + payload_fec
	std::string all_fec = "";
	int pad_hdr = 0;
	int size_payload_fec = 0;
	std::string hdr_fec_str = "";
	std::string deinterleaved_str = "";
	switch(scheme)
	{
		case 0:	deinterleaved_str = priority_str;
		break;

		case 1:	hdr_only = priority_str.substr(0, NBYTES[0]);
		payload_only = priority_str.substr(NBYTES[0], NBYTES[1]);

		// first few bytes of payload should go with the hdr (same fec group)
		pad_hdr = ((fec_K - hdr_data_size) < payload_only.size()) ? fec_K - hdr_data_size: payload_only.size();
		hdr_only.append(payload_only.substr(0, pad_hdr));	

		// printf("pad_hdr: %d, payload_only: %d\n", pad_hdr, payload_only.size());
		payload_only = payload_only.substr(pad_hdr, payload_only.size() - pad_hdr);

		if(fec_K == 0)
			return hdr_only + payload_only;
		// printf("size: %d %d %d %d\n", priority_str.size(), NBYTES[0], NBYTES[1], NBYTES[2]);
		hdr_fec_only = priority_str.substr(NBYTES[0]+ NBYTES[1], NBYTES[2]);

		payload_fec_only = priority_str.substr(NBYTES[0] + NBYTES[1] + NBYTES[2], NBYTES[3]);
		// printf("payload_fec_only: %d\n", payload_fec_only.size());

		payload_fec = reconstruct_payload_fec(payload_only, payload_fec_only, fec_N, fec_K);
		deinterleaved_str = hdr_only + hdr_fec_only + payload_fec;
		break;

		case 2: payload_only = priority_str.substr(0, NBYTES[0]);
		hdr_only = priority_str.substr(NBYTES[0], NBYTES[1]);

				// first few bytes of payload should go with the hdr (same fec group)
		pad_hdr = ((fec_K - hdr_data_size) < payload_only.size()) ? fec_K - hdr_data_size: payload_only.size();
		hdr_only.append(payload_only.substr(0, pad_hdr));
		payload_only = payload_only.substr(pad_hdr, payload_only.size() - pad_hdr);

		if(fec_K == 0)
			return hdr_only + payload_only;

		hdr_fec_only = priority_str.substr(NBYTES[0]+ NBYTES[1], NBYTES[2]);
		payload_fec_only = priority_str.substr(NBYTES[0] + NBYTES[1] + NBYTES[2], NBYTES[3]);
		payload_fec = reconstruct_payload_fec(payload_only, payload_fec_only, fec_N, fec_K);
		deinterleaved_str = hdr_only + hdr_fec_only + payload_fec;
		break;

		case 3: hdr_payload_only = priority_str.substr(0, NBYTES[0]);
		hdr_only = hdr_payload_only.substr(0, hdr_data_size);
		payload_only = hdr_payload_only.substr(hdr_data_size, NBYTES[0] - hdr_data_size);


				// first few bytes of payload should go with the hdr (same fec group)
		pad_hdr = ((fec_K - hdr_data_size) < payload_only.size()) ? fec_K - hdr_data_size: payload_only.size();
		hdr_only.append(payload_only.substr(0, pad_hdr));
		payload_only = payload_only.substr(pad_hdr, payload_only.size() - pad_hdr);

		if(fec_K == 0)
			return hdr_only + payload_only;

		hdr_fec_only = priority_str.substr(NBYTES[0], NBYTES[1]);
		payload_fec_only = priority_str.substr(NBYTES[0] + NBYTES[1], NBYTES[2]);
		payload_fec = reconstruct_payload_fec(payload_only, payload_fec_only, fec_N, fec_K);
		deinterleaved_str = hdr_only + hdr_fec_only + payload_fec;
		break;
		case 100: hdr_only = priority_str.substr(0, NBYTES[0]);					// !old!!!!!
		others = priority_str.substr(NBYTES[0], NBYTES[1]);

		payload_only = others.substr(0, expectedPayloadWithHdrLen - hdr_only.size());
		hdr_fec_only = others.substr(payload_only.size(), (fec_N - fec_K));
		payload_fec_only = others.substr(payload_only.size() + hdr_fec_only.size(), others.size() - (payload_only.size() + hdr_fec_only.size()));

		// first few bytes of payload should go with the hdr (same fec group)
		pad_hdr = ((fec_K - hdr_data_size) < payload_only.size()) ? fec_K - hdr_data_size: payload_only.size();
		hdr_only.append(payload_only.substr(0, pad_hdr));
		payload_only = payload_only.substr(pad_hdr, payload_only.size() - pad_hdr);

		if(fec_K == 0)
			return hdr_only + payload_only;

		payload_fec = reconstruct_payload_fec(payload_only, payload_fec_only, fec_N, fec_K);
		deinterleaved_str = hdr_only + hdr_fec_only + payload_fec; 		 
		break;

		case 4: hdr_payload_only = priority_str.substr(0, NBYTES[0]);
		hdr_only = hdr_payload_only.substr(0, hdr_data_size);
		payload_only = hdr_payload_only.substr(hdr_data_size, NBYTES[0] - hdr_data_size);

		// first few bytes of payload should go with the hdr (same fec group)
		pad_hdr = ((fec_K - hdr_data_size) < payload_only.size()) ? fec_K - hdr_data_size: payload_only.size();
		hdr_only.append(payload_only.substr(0, pad_hdr));
		payload_only = payload_only.substr(pad_hdr, payload_only.size() - pad_hdr);

		if(fec_K == 0)
			return hdr_only + payload_only;

		all_fec = priority_str.substr(NBYTES[0], NBYTES[1]);
		hdr_fec_only = all_fec.substr(0, (fec_N - fec_K));		
		size_payload_fec = all_fec.size() - hdr_fec_only.size();	

		payload_fec_only = all_fec.substr((fec_N - fec_K), size_payload_fec);

		payload_fec = reconstruct_payload_fec(payload_only, payload_fec_only, fec_N, fec_K);
		deinterleaved_str = hdr_only + hdr_fec_only + payload_fec;
		break;

		case 5: hdr_fec_str = priority_str.substr(0, NBYTES[0]);
		payload_fec = priority_str.substr(NBYTES[0], NBYTES[1]);

		hdr_only = hdr_fec_str.substr(0, hdr_data_size);
		hdr_fec_only = hdr_fec_str.substr(hdr_data_size, fec_N - fec_K);

		payload_only = payload_fec.substr(0, expectedPayloadWithHdrLen - hdr_only.size());
		payload_fec_only = payload_fec.substr(payload_only.size(), payload_fec.size() - payload_only.size());	

		if(fec_K == 0)
			return hdr_only + payload_only;		

		// first few bytes of payload should go with the hdr (same fec group)
		pad_hdr = ((fec_K - hdr_data_size) < payload_only.size()) ? fec_K - hdr_data_size: payload_only.size();
		hdr_only.append(payload_only.substr(0, pad_hdr));
		payload_only = payload_only.substr(pad_hdr, payload_only.size() - pad_hdr);

		payload_fec = reconstruct_payload_fec(payload_only, payload_fec_only, fec_N, fec_K);
		deinterleaved_str = hdr_only + hdr_fec_only + payload_fec;
		break;

		case 6: hdr_only = priority_str.substr(0, NBYTES[0]);
		hdr_fec_only = priority_str.substr(NBYTES[0], NBYTES[1]);

		payload_only = priority_str.substr(NBYTES[0] + NBYTES[1], NBYTES[2]);
		payload_fec_only = priority_str.substr(NBYTES[0] + NBYTES[1] + NBYTES[2], NBYTES[3]);	

		// first few bytes of payload should go with the hdr (same fec group)
		pad_hdr = ((fec_K - hdr_data_size) < payload_only.size()) ? fec_K - hdr_data_size: payload_only.size();
		hdr_only.append(payload_only.substr(0, pad_hdr));
		payload_only = payload_only.substr(pad_hdr, payload_only.size() - pad_hdr);

		payload_fec = reconstruct_payload_fec(payload_only, payload_fec_only, fec_N, fec_K);
		deinterleaved_str = hdr_only + hdr_fec_only + payload_fec;
		break;
	}


	if(mac_fec_flag)
	{
#ifdef DEBUG
		printf("rcvr: before deinterleaving.. \n");
		for(int i = 0; i < parity_bits_str.size(); i++)
			printf("%c", parity_bits_str[i]);
		printf("\n");
#endif

		std::string deinterleaved_bits = getBits((const unsigned char*) deinterleaved_str.data(), deinterleaved_str.size(), 8);
		int num_bits_before_deinterleaving = buf.size() * 8;
		int num_parity = parity_bits_str.size();
		num_parity = num_parity - (num_parity % 8);

		parity_bits_str = parity_bits_str.substr(0, num_parity);
	//test_parity_bits_str = test_parity_bits_str.substr(0, num_parity);	

	// deinterleave the parity and separate the padding from parity (look at interleave_mapping_best) //
		int parity_one_round = subcarriers * bps;

		unsigned char *final_parity_bits = (unsigned char*) malloc(sizeof(unsigned char) * mac_parity_added + 1);
		memset(final_parity_bits, 0, sizeof(unsigned char) * mac_parity_added + 1);

	// debug only //
		//unsigned char *test_final_parity_bits = (unsigned char*) malloc(sizeof(unsigned char) * mac_parity_added + 1);
		//memset(test_final_parity_bits, 0, sizeof(unsigned char) * mac_parity_added + 1);
	// debug ends //

		int interleaved_parity_pos[subcarriers * bps];
		initInterleavedParityPos(interleaved_parity_pos, bps, subcarriers);

		// printf("mac_parity_added: %d\n", mac_parity_added); 
		for(int i = 0; i < mac_parity_added; i++)
		{
			int index = i % parity_one_round;
			int blocks_mapped = i / parity_one_round;
			int new_pos = interleaved_parity_pos[index] + blocks_mapped * parity_one_round;
			assert(new_pos < ((blocks_mapped + 1) * parity_one_round));
			final_parity_bits[i] = parity_bits_str[new_pos];

		//test_final_parity_bits[i] = test_parity_bits_str[new_pos];
		}

		parity_bits_str.assign((const char*) final_parity_bits);	
	//test_parity_bits_str.assign((const char*) test_final_parity_bits);

#ifdef DEBUG
		printf("rcvr: after deinterleaving.. \n");
		for(int i = 0; i < parity_bits_str.size(); i++)
			printf("%c", parity_bits_str[i]);
		printf("\n");
#endif

		int symSize = 1;      // for LDPC, its immaterial, just keep it at 1
#ifdef RS_MAC_PARITY
		symSize = getSymSizeFromN_Syms(fec_N);
		fec_N = fec_N * symSize;
		fec_K = fec_K * symSize;

#endif
		decodeMACLayer_FEC_OPT(deinterleaved_bits,
			parity_bits_str,
			final_pos_map,
			orig_pos_map,
			subcarriers, fec_N,
			fec_K, bps, 
			num_bits_before_deinterleaving,
			fec_split_map, sorted_orig_pos,
			0, 0, 0, hint_flag);
	}

	free(orig_pos_map);
	free(final_pos_map);

	return deinterleaved_str;	
}

// alternate fec group delivery rate computation - //
/* use normal distribution to estimate the FEC group delivery rate * * X(i) = probability that codeword 'i' is lost * Y = sum X(i) */
double normal_cdf(double z)
{
	if (z > 6.0) { return 1.0; };
	if (z < -6.0) { return 0.0; };
	double b1 = 0.31938153;
	double b2 = -0.356563782;
	double b3 = 1.781477937;
	double b4 = -1.821255978;
	double b5 = 1.330274429;
	double p = 0.2316419;
	double c2 = 0.3989423;

	double a=fabs(z);
	double t = 1.0/(1.0+a*p);
	double b = c2*exp((-z)*(z/2.0));
	double n = (((((b5*t+b4)*t)+b3)*t+b2)*t+b1)*t;
	n = 1.0-b*n;
	if (z < 0.0) n = 1.0-n;
	return n;
}

float getFECGroupDeliveryRate(float *t_eff_delivery_rate, int nsymbols, int fec_N, int fec_K)
{

	float mean_loss_rate = 0.0;
	float var_loss_rate = 0.0;

	for(int i = 0; i < nsymbols; i++)
	{
		float Pi = 1 - t_eff_delivery_rate[i];
		mean_loss_rate += Pi;
		var_loss_rate += (Pi * (1 - Pi));
	}

	float sigma_loss_rate = sqrt(var_loss_rate);

	float threshold = (float) (fec_N - fec_K)/2;

	float normalized_threshold = (threshold-mean_loss_rate)/sigma_loss_rate;

	return(normal_cdf(normalized_threshold));
}

/* use Chebyshev's method to estimate the FEC group delivery rate
*  *
*   * X(i) = probability that codeword 'i' is lost
*    * Y = sum X(i)
*     * alpha = (n - k - mean(Y))/sigma(Y)
*      */
float getFECGroupDeliveryRate1(float *t_eff_delivery_rate, int nsymbols, int fec_N, int fec_K)
{
	float ub_loss = 0.0;

	float mean_loss_rate = 0.0;
	float var_loss_rate = 0.0;

	for(int i = 0; i < nsymbols; i++)
	{
		float Pi = 1 - t_eff_delivery_rate[i];
		mean_loss_rate += Pi;
		var_loss_rate += (Pi * (1 - Pi));
	}

	float sigma_loss_rate = sqrt(var_loss_rate);

	float threshold = (float) (fec_N - fec_K)/2;
	if(threshold - mean_loss_rate > 0)
	{
		float k = (threshold - mean_loss_rate) / sigma_loss_rate;
		ub_loss = 1 / (1 + pow(k, 2));
	}
	else
		ub_loss = 1;
	// printf("getFECGroupDelivery, nsymbols: %d, N: %d, K: %d\n", nsymbols, fec_N, fec_K);
	// printf("mean_loss: %f, var_loss: %f, threshold: %f, sigma_loss: %f\n", mean_loss_rate, var_loss_rate, threshold, sigma_loss_rate);
	return 1 - ub_loss;
}

/* to read the bits from the file */
std::string readBytes(int pkt_len, const char *filename1)
{
//const char *filename1 = "raw_bits.dat";

	int fd;
	FILE *d_fp;
	if ((fd = open (filename1, O_RDONLY|OUR_O_LARGEFILE, 0664)) < 0){
		perror(filename1);
	}else {
		if((d_fp = fdopen (fd, true ? "r" : "rb")) == NULL) {
			fprintf(stderr, "raw-bytes file cannot be opened\n");
			close(fd);
		}
		fprintf(stderr, "raw-bytes file opened!\n");
	}
	assert(d_fp);
	fseek(d_fp, 0, SEEK_END);
	int endPos = ftell(d_fp);
//printf("endPos: %d\n", endPos);

	int len = pkt_len * 8;
	fseek(d_fp, 0, SEEK_SET);
	unsigned char* d_data = (unsigned char*) malloc(sizeof(unsigned char) * endPos + 1);
	memset(d_data, 0, sizeof(unsigned char) * endPos + 1);
	int count = fread_unlocked(d_data, len, endPos/(sizeof(unsigned char) * len), d_fp);

// convert into bytes //
	std::string bits_str((const char*) d_data);
	unsigned char* bytes = (unsigned char*) malloc(sizeof(unsigned char) * pkt_len * count);
	memset(bytes, 0, sizeof(unsigned char) * pkt_len * count);

	std::string bytes_str = getMsg(bits_str, bytes);   
	return bytes_str;
}

std::string readBits(int pkt_len)
{
	const char *filename1 = "rx_bits.dat";

	int fd;
	FILE *d_fp;
	if ((fd = open (filename1, O_RDONLY|OUR_O_LARGEFILE, 0664)) < 0){
		perror(filename1);
	}else {
		if((d_fp = fdopen (fd, true ? "r" : "rb")) == NULL) {
			fprintf(stderr, "raw-bytes file cannot be opened\n");
			close(fd);
		}
		fprintf(stderr, "raw-bytes file opened!\n");
	}
	assert(d_fp);
	fseek(d_fp, 0, SEEK_END);
	int endPos = ftell(d_fp);

	int len = pkt_len * 8;
	fseek(d_fp, 0, SEEK_SET);
	unsigned char* d_data = (unsigned char*) malloc(sizeof(unsigned char) * endPos + 1);
	memset(d_data, 0, sizeof(unsigned char) * endPos + 1);
	int count = fread_unlocked(d_data, len, endPos/(sizeof(unsigned char) * len), d_fp);

	std::string bits_str((const char*) d_data);
	return bits_str;
}


void run_deinterleaver()
{
// 	int pkt_len = 1440;
// 	int mac_parity_bits = 0;
// 
// 	std::string bytes_str = readBytes(pkt_len, "raw_bits.dat"); 
// 	int num_subcarriers = 192;
// 	int fec_N = 255;
// 	int fec_K = 192;
// 	int bits_per_symbol = 4;
// 	std::string parity_bits_str;
// 	int payload_size = 1008;
// 
// 	mac_fec_flag = 0;
// 	int scheme = 0;
// 	int hint_flag = 1;
// 
// #ifdef RS_MAC_OPT_VERSION
// 	static std::map<int, split> fec_split_map;
// 	static std::vector<int> sorted_orig_pos;
// #endif
// 	static std::vector<int> orig_pos_indices;
// 
// 	int num_pkts = bytes_str.size()/pkt_len;
// 
// 	int n_total_fec_groups = 0;
// 	int n_correct_fec_groups = 0;
// 	int total = 0; 
// 
// 	std::string tmp_1, tmp_2;
// 
// 	// read the correct data bits for one packet: used for bit counting later //
// 	std::string correct_bytes = readBytes(pkt_len, "rx_bits.dat.1008.correct"); 
// 	int n_correct_bits = 0;
// 	float *csi = (float*) malloc(sizeof(float) * 192);
// 	memset(csi, 0, sizeof(float) * 192);
// 	readCSISubcarriers(csi, 192);
// 	int *final_pos_map = (int*) malloc(sizeof(int) * pkt_len * 8);
// 
// 	for(int i = 0; i < num_pkts; i++)
// 	{
// 		int total_fec_groups = 0;
// 		int correct_fec_groups = 0;
// 		std::string pkt_str; 
// 		if(mac_fec_flag == 1)
// 			pkt_str = bytes_str.substr(i * pkt_len, pkt_len);
// 		else
// 		{
// 			pkt_str = bytes_str.substr(i * pkt_len, pkt_len - mac_parity_bits/8); 
// 			mac_parity_bits = 0;
// 		}
// 
// 		std::string deinterleaved_string_std = deinterleave_mapping_best(pkt_str, num_subcarriers, bits_per_symbol, fec_N, fec_K,
// 			scheme, payload_size, parity_bits_str, mac_parity_bits, 
// 			fec_split_map, sorted_orig_pos, tmp_1, tmp_2, hint_flag
// #ifdef PARTIAL_TPUT
// 			, final_pos_map
// #endif
// 			);
// 
// #ifdef DEBUG_MAC_FEC
// 		printf(" <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< OUTPUT from deInterleaver , lt: %d >>>>>>>>>>>>>>>>>>>>>>>>>>>>>> \n", deinterleaved_string_std.size());
// 		for(int i = 0; i < deinterleaved_string_std.size(); i++)
// 			printf("%d ", deinterleaved_string_std.data()[i]);
// 		printf("\n");
// #endif
// 		int t_num_corr_bits = 0;
// 		std::string decoded_str = gr_FEC(deinterleaved_string_std, fec_N, fec_K, bits_per_symbol, false, parity_bits_str,
// #ifdef RS_MAC_OPT_VERSION
// 			fec_split_map,
// 			sorted_orig_pos,
// #endif
// 			orig_pos_indices,
// 			0, 0, false, payload_size,
// 			total_fec_groups, correct_fec_groups, correct_bytes, tmp_2
// #ifdef PARTIAL_TPUT
// 			, t_num_corr_bits, final_pos_map, 192, csi, hint_flag
// #endif
// 			);
// 
// #ifdef DEBUG_MAC_FEC
// 		printf(" <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< FINAL OUTPUT  , lt: %d >>>>>>>>>>>>>>>>>>>>>>>>>>>>>> \n", decoded_str.size());
// 		for(int i = 0; i < decoded_str.size(); i++)
// 			printf("%d ", ((unsigned char*) decoded_str.data())[i]);
// 		printf("\n");
// #endif
// 
// 	// do the crc check on the header //
// 		int hdr_data_size = 24;
// 		int hdr_size_wo_crc =  hdr_data_size - 4;
// 		std::string hdr_w_crc_str = decoded_str.substr(0, hdr_data_size);
// 		std::string exp_crc = hdr_w_crc_str.substr(hdr_size_wo_crc, 4);	
// 
// 		std::string hdr_wo_crc_str = hdr_w_crc_str.substr(0, 20);
// 		unsigned int rx_crc = gr_crc32(hdr_wo_crc_str);
// 		char tmp[5];
// 		snprintf(tmp, sizeof(tmp), "%ld", rx_crc);
// 		std::string actual_crc(tmp);
// 		std::string payload_only = decoded_str.substr(hdr_data_size, decoded_str.size() - hdr_data_size);
// 		if(actual_crc.compare(exp_crc) != 0)
// 		{
// 			printf("RX_HDR INCORRECT\n"); 
// 			correct_fec_groups = 0;
// 			t_num_corr_bits = 0;
// 		}
// 
// 		decoded_str = hdr_wo_crc_str + payload_only;
// 
// 	// do the crc check on the packet //
// 		std::string decoded_str_wo_crc = decoded_str.substr(0, decoded_str.size() - 4);
// 
// #ifdef DEBUG_MAC_FEC
// 		printf("decoded_str size: %d\n", decoded_str.size()); 
// 		printf(" <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< FINAL PAYLOAD ONLY  , lt: %d >>>>>>>>>>>>>>>>>>>>>>>>>>>>>> \n", decoded_str_wo_crc.size());
// 		for(int i = 0; i < decoded_str_wo_crc.size(); i++)
// 			printf("%d ", ((unsigned char*) decoded_str_wo_crc.data())[i]);
// 		printf("\n");
// #endif
// 
// 
// 		exp_crc = decoded_str.substr(decoded_str.size() - 4, 4);
// 		rx_crc = gr_crc32(decoded_str_wo_crc);
// 		char tmp1[5];
// 		snprintf(tmp1, sizeof(tmp1), "%ld", rx_crc);
// 		std::string _actual_crc(tmp1);
// 		if(_actual_crc.compare(exp_crc) == 0)
// 			total++;  
// 
// 		printf("actual_crc: \n");
// 		for(int i = 0; i < 4; i++)
// 			printf("%d ", ((unsigned char*) _actual_crc.data())[i]);
// 
// 		printf("\nexp_crc: \n");
// 		for(int i = 0; i < 4; i++)
// 			printf("%d ", ((unsigned char*) exp_crc.data())[i]);
// 		printf("\n");
// 		n_total_fec_groups += total_fec_groups;
// 		n_correct_fec_groups += correct_fec_groups;
// 		n_correct_bits += t_num_corr_bits;
// 	}
// 	free(csi);
// 	printf("rcvd: %d, correct: %d, total_fec_groups: %d, correct_fec_groups: %d correct_bits: %d\n", num_pkts, total, n_total_fec_groups, n_correct_fec_groups, n_correct_bits);
}

// both the strings input contain only the 'data' bits, no fec-parity bits //
int dumpActualFECGroupDeliveryRates(std::string encoded_str, std::string decoded_str, int fec_K, int symSize)
{
	// printf("dumpActualFECGroupDeliveryRates:: , fec_k: %d\n", fec_K); fflush(stdout);
	assert(encoded_str.size() == decoded_str.size());

	int size = decoded_str.size();
	std::string decoded_bits = getBitString((const unsigned char*) decoded_str.data(), decoded_str.size());
	std::string encoded_bits = getBitString((const unsigned char*) encoded_str.data(), encoded_str.size());

#ifdef DEBUG1
	printf("encoded bits: \n");
	for(int i = 0; i < encoded_bits.size(); i++)
		printf("%c", encoded_bits[i]); 
	printf("\n decoded bits: \n");
	for(int i = 0; i < decoded_bits.size(); i++)
		printf("%c", decoded_bits[i]);
	printf("\n");
#endif
	int n_bits = decoded_bits.size();

	int num_fec_groups = ceil((float)decoded_bits.size()/(fec_K * symSize));	

	int total_bits_delivered = 0;
	for(int i = 0; i < num_fec_groups; i++)
	{
		int bits_delivered = 0;				// tracks the bits delivered in an fec group //

		int bits_this_group = min(fec_K * symSize, n_bits - i * fec_K * symSize);
		for(int j = 0; j < bits_this_group; j++)		// all the bits in the current fec group //
		{
			if(decoded_bits[i*fec_K*symSize+j] == encoded_bits[i*fec_K*symSize+j])
				bits_delivered++;		
		}
		total_bits_delivered += bits_delivered;
		// printf("num_fec_groups: %d, fec_K: %d, fec_group: %d, actual_delivery_rate: %d, size: %d\n", num_fec_groups, fec_K, i, bits_delivered, decoded_bits.size());

		std::string decoded_this_group = decoded_bits.substr(i*fec_K*symSize, bits_this_group);
		std::string encoded_this_group = encoded_bits.substr(i*fec_K*symSize, bits_this_group);

	}
	return total_bits_delivered;
}

/* does not take the mac-fec into account, so should be called from appropriate locations */
float getHeaderDeliveryRate(float *eff_delivery_rate, float fec_group_delivery_rate, int *interleaved_pos, int hdr_size)
{
	// printf("fec_group_delivery_rate: %f\n", fec_group_delivery_rate); fflush(stdout);
	float delivery_rate = 1.0;
	for(int i = 0; i < hdr_size * 8; i++)
	{
		int tx_pos = interleaved_pos[i];
		delivery_rate *= eff_delivery_rate[tx_pos];		 
	}

	delivery_rate = 1 - (1 - delivery_rate) * (1 - fec_group_delivery_rate);

	return delivery_rate;
}

float getUtility_not_used(float total_fec_delivery_rate, float hdr_fec_group_delivery_rate, float *eff_delivery_rate, int *interleaved_pos, int num_bits)
{
	int hdr_size = 24; 				// 24 bytes

	float hdr_delivery_rate = getHeaderDeliveryRate(eff_delivery_rate, hdr_fec_group_delivery_rate, interleaved_pos, hdr_size);
	float utility = (total_fec_delivery_rate * hdr_delivery_rate)/num_bits;

	return utility;
}


float getUtility(float* fec_delivery_rates, float *eff_delivery_rate, int *interleaved_pos, int n_data_bits, int num_bits_out, int fec_K, int fec_N, float hdr_delivery_rate, int hint_flag)
{
	int num_fec_groups = ceil((float)n_data_bits/fec_K);

	float numerator = 0.0;				// estimated number of symbols delivered
	for(int i = 0; i < num_fec_groups; i++)
		numerator += fec_delivery_rates[i];

	//printf("numerator: %f, hdr_delivery_rate: %f, num_bits_out: %d\n", numerator, hdr_delivery_rate, num_bits_out);
	float utility = (numerator * hdr_delivery_rate)/num_bits_out;
	return utility;
}


int calcBERValuesOnSubcarriers(std::string tx_str, std::string rx_str, int modulation, int subcarriers, int *bit_errors, int *total_bits)
{
	assert(tx_str.size() == rx_str.size());

	int size = rx_str.size();
	std::string rx_bits = getBitString((const unsigned char*) rx_str.data(), rx_str.size());
	std::string tx_bits = getBitString((const unsigned char*) tx_str.data(), tx_str.size());

	for(int i = 0; i < size; i++)
	{
		int sym_index = i/modulation;
		int sc_index = sym_index%subcarriers;

		if(tx_bits[i] != rx_bits[i])
			bit_errors[sc_index]++;		

		total_bits[sc_index]++;
	}
}

float getFECSymbolDeliveryRateBySlope(int fec_N, float avg_ber)
{ 
	//printf("getFECSymbolDeliveryRateBySlope, fec_N: %d, avg_ber: %f\n", fec_N, avg_ber);
	int symSize = getSymSizeFromN_Bits(fec_N);
	float avg_slope = 0.5;                    // across symbol sizes

	// find the slope for this symbol size //
	float slope = symSize * avg_slope;

	// get the symbol-error-rate //
	float dr = 1 - (avg_ber * slope);
	//printf("dr: %f\n", dr);
	return dr; 
}

float getFECSymbolDeliveryRateByRoot(float del_rate_bits)
{
	float dr = sqrt(del_rate_bits);    
	return dr;
}

// used for the case where the bits of the FEC group can be delivered by themselves //
float getAvgDelRateOfFECGroup(int fec_group_index, float *eff_delivery_rate, int *interleaved_bit_pos, int fec_N, int num_bits)
{
	float avg_dr = 0.0;

	// calculate the avg ber of the 'data' bits in the fec-group //
	for(int j = 0; j < num_bits; j++)
	{
		int _orig_bit_pos = (fec_group_index * fec_N) + j;
		int _interleaved_bit_pos = interleaved_bit_pos[_orig_bit_pos];
		avg_dr += eff_delivery_rate[_interleaved_bit_pos];
	}
	return avg_dr/num_bits;
} 





void simulate_a_pkt(int fec_N, int fec_K, std::string combined_payload, int hdr_size, int num_subcarriers, int bits_per_symbol, 
	std::string parity_bits_str, 
#ifdef RS_MAC_OPT_VERSION
	std::map<int, split> &fec_split_map_sender,
	std::vector<int> &sorted_orig_pos_sender,
	std::map<int, split> &fec_split_map_rcvr,
	std::vector<int> &sorted_orig_pos_rcvr,
#endif
	std::vector<int> &orig_pos_indices,
	int mac_fec_N, int mac_fec_K, int scheme, bool useLDPC, int& total_bits_delivered, double& used_time, bool sampling, int hint_flag,
	int &mac_parity_bits)
{
		
		
		
	// encode the fec together for the 'combined_payload' //
	int t_fec_groups = 0;
	int c_fec_groups = 0;
	
	
	std::string fec_combined_str = gr_FEC(combined_payload, fec_N, fec_K, bits_per_symbol, true, parity_bits_str, 
#ifdef RS_MAC_OPT_VERSION
		fec_split_map_sender,
		sorted_orig_pos_sender,
#endif
		orig_pos_indices, 
		mac_fec_N, mac_fec_K, useLDPC, combined_payload.size(), 
		t_fec_groups, c_fec_groups, "", "");



	// interleave the combination //
	std::string interleaved_string = interleave_mapping_best(fec_combined_str, num_subcarriers, bits_per_symbol, fec_N, fec_K, scheme, mac_parity_bits, hint_flag);
	//std::string interleaved_string = fec_combined_str;			// bypasses the interleaver

	


	// debug only //
	std::string test_interleaved_string = interleaved_string;
	std::string test_parity_bits_str;
	// debug end //

	std::string err_string = injectErrorsProbabilistically(interleaved_string, g_csi, num_subcarriers, bits_per_symbol);
	interleaved_string = err_string;					// bypass the error injection
	
	
	

	// deinterleave the combination //
	std::string deinterleaved_string = deinterleave_mapping_best(interleaved_string, num_subcarriers, bits_per_symbol, fec_N, fec_K, 
		scheme, combined_payload.size(), parity_bits_str, 
		mac_parity_bits, fec_split_map_rcvr, sorted_orig_pos_rcvr, 
		test_interleaved_string, test_parity_bits_str, hint_flag);
	//std::string deinterleaved_string = interleaved_string;			// bypasses the de-interleaver


	
	

	// de-FEC on 'combined_payload' //
	std::string fec_decoded_combined = gr_FEC(deinterleaved_string, fec_N, fec_K, bits_per_symbol, false, parity_bits_str,
#ifdef RS_MAC_OPT_VERSION
		fec_split_map_rcvr,
		sorted_orig_pos_rcvr,
#endif
		orig_pos_indices,
		mac_fec_N, mac_fec_K, useLDPC, combined_payload.size(), 
		t_fec_groups, c_fec_groups, fec_combined_str, test_parity_bits_str);	

	// extract the hdr //
	std::string decoded_hdr = fec_decoded_combined.substr(0, hdr_size + 4);		// crc: '4' bytes

	// check crc on the header //
	std::string rx_crc_str = decoded_hdr.substr(hdr_size, 4);
	unsigned int hdr_crc = gr_crc32(decoded_hdr.substr(0, hdr_size));
	char hdr_crc_str1[5];
	snprintf(hdr_crc_str1, sizeof(hdr_crc_str1), "%ld", hdr_crc);
	std::string calc_hdr_crc_str(hdr_crc_str1);
	// crc complete on the header //



	std::string tx_bits = getBitString((const unsigned char*) combined_payload.data(), combined_payload.size());
	std::string rx_bits = getBitString((const unsigned char*) fec_decoded_combined.data(), fec_decoded_combined.size());
	// statistics: no hint
	// printf("before, %d\n", throughput);
	// with CRC
	if(MODE_GEN_HINT == NO_GEN_HINT) {
		// test header
		int header_fail = 0;
		if(calc_hdr_crc_str.compare(rx_crc_str) != 0) {
			header_fail ++;
		}

		if(header_fail == 0) {
			int n_bits = tx_bits.size();
			int num_fec_groups = ceil((float)n_bits/(fec_K * 8));	

			for(int i = 0; i < num_fec_groups; i++)
			{
				int bits_delivered = 0;				// tracks the bits delivered in an fec group //
				int bits_this_group = min(fec_K * 8, n_bits - i * fec_K * 8);

				for(int j = 0; j < bits_this_group; j++)		// all the bits in the current fec group //
				{
					if(tx_bits[i*fec_K*8+j] == rx_bits[i*fec_K*8+j])
						bits_delivered ++;		
				}
				if(bits_delivered == bits_this_group) {
					total_bits_delivered += bits_delivered;
				}
			}
		}
	}
	// statistics: hint
	else if(MODE_GEN_HINT == GEN_HINT) {
		int header_fail = 0;
		if(calc_hdr_crc_str.compare(rx_crc_str) != 0) {
			header_fail ++;
		}
		if(header_fail == 0) {
			int n_bits = tx_bits.size();
			int num_fec_groups = ceil((float)n_bits/(fec_K * 8));	

			// int total_bits_delivered = 0;
			for(int i = 0; i < num_fec_groups; i++)
			{
				int bits_this_group = min(fec_K * 8, n_bits - i * fec_K * 8);

				for(int j = 0; j < bits_this_group; j++)		// all the bits in the current fec group //
				{
					if(tx_bits[i*fec_K*8+j] == rx_bits[i*fec_K*8+j])
						total_bits_delivered ++;		
				}
			}
		}
	}	// end if with hint


	// int send_bits = (combined_payload.size() ) * 8 * fec_N / fec_K + mac_parity_bits;
	int send_bits = (combined_payload.size() ) * 8 * fec_N / fec_K;
	used_time += (send_bits / RATE + 28 + 2 * 9 + 16);
	
	
}



double auto_rate2(int fec_N, int fec_K, std::string combined_payload, int hdr_size, int num_subcarriers, int bits_per_symbol, 
	std::string parity_bits_str, 
#ifdef RS_MAC_OPT_VERSION
	std::map<int, split> &fec_split_map_sender,
	std::vector<int> &sorted_orig_pos_sender,
	std::map<int, split> &fec_split_map_rcvr,
	std::vector<int> &sorted_orig_pos_rcvr,
#endif
	std::vector<int> &orig_pos_indices,
	int mac_fec_N, int mac_fec_K, int &scheme, bool useLDPC, /*int& total_bits_delivered, double& used_time, bool sampling,*/ int hint_flag,
	int &mac_parity_bits)
{	// other output: FEC_RATE, RATE, MODE_MODULATION

	
	short best_scheme = 0; // 0: rand, 1: header, 2: data, 3: both
	bool *best_sw_map = (bool *)malloc(sizeof(bool) * MAX_ITR);
	
	double best_throughput = 0;
	
		
	// 4 mapping schemes
	memcpy(best_sw_map, SW_MAP, sizeof(bool) * MAX_ITR);
	for(scheme = 1; scheme <= 1; scheme ++) {
		for(int i = 0; i < MAX_ITR; i ++) SW_MAP[i] = false;
		// for(int i = 0; i < MAX_ITR; i ++) SW_MAP[i] = true;

		for(int itr = -1; itr < MAX_ITR; itr ++) {
		// for(int itr = -1; itr < 0; itr ++) {
			if(itr >= 0) SW_MAP[itr] = true;
			
			int tmp_throughput = 0;
			double tmp_used_time = 0;
			
			
			simulate_a_pkt(fec_N, fec_K, combined_payload, hdr_size, num_subcarriers, bits_per_symbol, parity_bits_str, 
			#ifdef RS_MAC_OPT_VERSION
				fec_split_map_sender,
				sorted_orig_pos_sender,
				fec_split_map_rcvr,
				sorted_orig_pos_rcvr,
			#endif
				orig_pos_indices,
				mac_fec_N, mac_fec_K, scheme, useLDPC, tmp_throughput, tmp_used_time, true, hint_flag, mac_parity_bits);

			// printf("%f\n", (tmp_throughput * 1.0 / tmp_used_time));

			if((tmp_throughput * 1.0 / tmp_used_time) > best_throughput) {
				// printf("better: %f > %f\n", (tmp_throughput * 1.0 / tmp_used_time), best_throughput);
				best_scheme = scheme;
				best_throughput = tmp_throughput * 1.0 / tmp_used_time;
				memcpy(best_sw_map, SW_MAP, sizeof(bool) * MAX_ITR);
			}
			else {
				if(itr >= 0) SW_MAP[itr] = false;
				break;
			}
			
		}
		
	}

	
	scheme = best_scheme;
	memcpy(SW_MAP, best_sw_map, sizeof(bool) * MAX_ITR);
	free(best_sw_map);
	
	return best_throughput;
}

