#include "rs.h"
#include <stdlib.h>     // char.h combines with gr_crc32.h
#include <string.h>
#include <strings.h>
#include <bitset>
#include <vector>
#include <map>
#include <math.h>
#include <stdio.h>      //for debugging
#include <assert.h>

#define min(a,b)        ((a) < (b) ? (a) : (b))

void gr_free_rs(void *p){
  struct rs *rs = (struct rs *)p;

  free(rs->alpha_to);
  free(rs->index_of);
  free(rs->genpoly);
  free(rs);
}

static struct rs 
*gr_init_rs(unsigned int symsize,unsigned int gfpoly,unsigned fcr,unsigned prim, unsigned int nroots)
{
  struct rs *rs;
  int i, j, sr,root,iprim;

  //printf("coming in......\n");

  if(symsize > 8*sizeof(DTYPE)){
    printf("Need version with ints rather than chars\n");
    exit(0);
    //return NULL; /* Need version with ints rather than chars */
  }
  if(fcr >= (1<<symsize)){
    printf("The fcr value is abnormal.\n");
    exit(0);
    //return NULL;
  }
  if(prim == 0 || prim >= (1<<symsize)){
    printf("The prim value is abnormal...\n");
    exit(0);
    //return NULL;
  }
  if(nroots >= (1<<symsize)){
    printf("Can't have more roots than symbol values!\n");
    exit(0);
    //return NULL; /* Can't have more roots than symbol values! */
  }

  rs = (struct rs *)calloc(1,sizeof(struct rs));
  rs->mm = symsize;
  rs->nn = (1<<symsize)-1;

  rs->alpha_to = (DTYPE *)malloc(sizeof(DTYPE)*(rs->nn+1));
  if(rs->alpha_to == NULL){
    free(rs);
    printf("Cannot set the values of 'alpha_to[]' of reedsolomon...\n");
    exit(0);
    //return NULL;
  }

  rs->index_of = (DTYPE *)malloc(sizeof(DTYPE)*(rs->nn+1));
  if(rs->index_of == NULL){
    free(rs->alpha_to);
    free(rs);
    printf("Cannot set the value of index of reedsolomon...\n");
    exit(0);
    //return NULL;
  }

  /* Generate Galois field lookup tables */
  rs->index_of[0] = A0; /* log(zero) = -inf */
  rs->alpha_to[A0] = 0; /* alpha**-inf = 0 */
  sr = 1;
  for(i=0;i<rs->nn;i++){
    rs->index_of[sr] = i;
    rs->alpha_to[i] = sr;
    sr <<= 1;
    if(sr & (1<<symsize))
      sr ^= gfpoly;
    sr &= rs->nn;
  }
  if(sr != 1){
    /* field generator polynomial is not primitive! */
    free(rs->alpha_to);
    free(rs->index_of);
    free(rs);
    printf("field generator polynomial is not primitive!\n");
    exit(0);
    //return NULL;
  }

  /* Form RS code generator polynomial from its roots */
  rs->genpoly = (DTYPE *)malloc(sizeof(DTYPE)*(nroots+1));
  if(rs->genpoly == NULL){
    free(rs->alpha_to);
    free(rs->index_of);
    free(rs);
    printf("Cannot set the values of generation polynomial of reedsolomon...\n");
    exit(0);
    //return NULL;
  }
  rs->fcr = fcr;
  rs->prim = prim;
  rs->nroots = nroots;

  /* Find prim-th root of 1, used in decoding */
  for(iprim=1;(iprim % prim) != 0;iprim += rs->nn)
    ;
  rs->iprim = iprim / prim;

  rs->genpoly[0] = 1;
  for (i = 0,root=fcr*prim; i < nroots; i++,root += prim) {
    rs->genpoly[i+1] = 1;

    /* Multiply rs->genpoly[] by  @**(root + x) */
    for (j = i; j > 0; j--){
      if (rs->genpoly[j] != 0)
        rs->genpoly[j] = rs->genpoly[j-1] ^ rs->alpha_to[modnn(rs,rs->index_of[rs->genpoly[j]] + root)];
      else
        rs->genpoly[j] = rs->genpoly[j-1];
    }
    /* rs->genpoly[0] can never be zero */
    rs->genpoly[0] = rs->alpha_to[modnn(rs,rs->index_of[rs->genpoly[0]] + root)];
  }
  /* convert rs->genpoly[] to index form for quicker encoding */
  for (i = 0; i <= nroots; i++)
    rs->genpoly[i] = rs->index_of[rs->genpoly[i]];

  return rs;
}

// apurv++ begins //
void gr_encode_rs(void *p, DTYPE *data, DTYPE *bb){

  struct rs *rs = (struct rs *)p;
  int i, j;
  DTYPE feedback;

  memset(bb,0,NROOTS*sizeof(DTYPE));

  for(i=0;i<NN-NROOTS;i++){
    feedback = INDEX_OF[data[i] ^ bb[0]];
    if(feedback != A0){      /* feedback term is non-zero */
#ifdef UNNORMALIZED
      /* This line is unnecessary when GENPOLY[NROOTS] is unity, as it must
       * always be for the polynomials constructed by init_rs()
       */
      feedback = MODNN(NN - GENPOLY[NROOTS] + feedback);
#endif
      for(j=1;j<NROOTS;j++)
        bb[j] ^= ALPHA_TO[MODNN(feedback + GENPOLY[NROOTS-j])];
    }
    /* Shift */
    memmove(&bb[0],&bb[1],sizeof(DTYPE)*(NROOTS-1));
    if(feedback != A0)
      bb[NROOTS-1] = ALPHA_TO[MODNN(feedback + GENPOLY[0])];
    else
      bb[NROOTS-1] = 0;
  }
}


std::string gr_encode_rs_fec(std::string s_bits, int n, int k)
{
   /*
    int syms;
    int gfpoly;
    int nn, nroots;

    switch(n){
	case 3:
		syms = 2; gfpoly = 0x7; break;
        case 15:
                syms = 4; gfpoly = 0x13; break;
        case 255:
                syms = 8; gfpoly = 0x11d; break;
        default:
                printf("the length of the block is out of range, FEC cannot be initiated.");
                break;
    }
    
    nn = (1<<syms)-1;
    nroots = ceil(nn*(1-(float)k/n));
    if((nroots%2)!=0)
    {
        nroots++;
    }
    
  
    nn = n;
    nroots = n - k;

    printf("nn = %d; nroots=%d, and the percentage is %f\n", nn, nroots, (float)k/n );

    // see if the current rs configuration reqd exists in the holder vector //
    static std::vector<struct rs*> rs_vec;
    
    std::vector<struct rs*>::iterator it = rs_vec.begin();
    static struct rs *rs; 
    bool found = false;
    while(it != rs_vec.end())
    {
	printf("searching!\n"); fflush(stdout);
	rs = *it;
	if(rs->nn == n && (rs->nroots - rs->nn) == k)
	{
		found = true;
		break;
	}
	it++;
    }
 
    if(!found)
    {
	rs = gr_init_rs(syms, gfpoly, 1, 1, nroots);
	rs_vec.push_back(rs);
    }
    */

    static std::vector<struct rs*> rs_vec;
    struct rs *rs = find_rs_handle(rs_vec, n, k);
    assert(rs);
    int syms = rs->mm;

    // printf("syms: %d, nn: %d, nroots: %d\n", syms, NN, NROOTS); 

    /*
    static struct rs *rs;
    static unsigned int handle_init = 0;
    	    

    // initializes the handle only once //
    if(handle_init == 0)
    {
       printf("entering initial stage......\n");
       rs = gr_init_rs(syms, gfpoly, 1, 1, nroots);
       printf("complete initial stage, entering gr_update_crc32 #1, handle%p\n", rs);
       handle_init++;
    }
    else if(handle_init == 1 && n != NN)
    {
	printf("initializing mac rs handle\n");
	rs = gr_init_rs(syms, gfpoly, 1, 1, nroots);
	handle_init++;
    }
    */
   
    /*
    const unsigned char *bytes = (const unsigned char*) buf.data();
    size_t nbytes = buf.size();
  
    // convert the data (8-bit) symbols into appropriate symSize symbols //
    std::string s_bits = getBits(bytes, nbytes, 8);
    */
    int num_bits = s_bits.size();

    unsigned char *data = (unsigned char*) malloc(sizeof(unsigned char) * ceil((float)num_bits/syms));
    memset(data, 0, sizeof(unsigned char) * ceil((float)num_bits/syms));

    // printf("s_bits size: %d\n", num_bits); fflush(stdout);
    int len = convertToSymbols(data, syms, s_bits); 

#ifdef DEBUG_FEC
    printf("raw buf %d symbols (symSize: %d)-\n", len, syms);
    for(int k = 0; k < len; k++)
    {
	printf("%d ", data[k]);
    }
    printf("\n");
#endif

    int num_blocks = ceil(((float)len/(rs->nn-rs->nroots)));
    int len_w_fec = (num_blocks) * NN;
    int rem = len % (NN-NROOTS);
    if(rem > 0)
    {
	len_w_fec = (num_blocks-1) * NN + rem + NROOTS;
    }

    /* encoded_data: the data that now contains the FEC as well */
    unsigned char *encoded_data = (unsigned char *) malloc(sizeof(unsigned char) * len_w_fec);
    memset(encoded_data, 0, sizeof(unsigned char) * len_w_fec);
    
#ifdef DEBUG_FEC
    printf("len: %d len_w_fec%d, num_blocks%d NN%d NROOTS%d\n", len, len_w_fec, num_blocks, NN, NROOTS); fflush(stdout);
#endif

    int dataLen = NN-NROOTS;
    int processedDataLen = 0;
    int pad_zeros = 0;
    unsigned char data_block[NN];	//temporary block of data//
    //memset(data_block, 0, NN);
    for (int i=0; i<num_blocks; i++)
    {
	memset(data_block, 0, NN);
        int num_bytes_to_encode = 0;

        if(processedDataLen + dataLen > len)
        {
             num_bytes_to_encode = len - processedDataLen;
             pad_zeros = dataLen - num_bytes_to_encode;
        }
        else
        {
             num_bytes_to_encode = dataLen;
        }
        processedDataLen += num_bytes_to_encode;
	
        // copy the data portion of this block into the encoded_data //
	memcpy(data_block, data, num_bytes_to_encode);

        data = data + num_bytes_to_encode;

        // use the shortened RS code if necessary, ie pad with 0s //
	memset(data_block + num_bytes_to_encode, 0, pad_zeros);

#ifdef DEBUG_FEC
       printf("\nraw data_block %d num_bytes_to_encode: %d pad_zeroes: %d \n", i, num_bytes_to_encode, pad_zeros);
       for(int k = 0; k < NN; k++)
	   printf("%d ", data_block[k]);
       printf("\n\n"); fflush(stdout);
#endif

        // encode this block using FEC //
	gr_encode_rs(rs, &data_block[0], &data_block[NN-NROOTS]);
 		
#ifdef DEBUG_FEC
       printf("\nencoded data_block %d - \n", i);
       for(int k = 0; k < NN; k++)
           printf("%d ", data_block[k]);
       printf("\n\n"); fflush(stdout);
#endif

       // copy the data part //
       memcpy(encoded_data + i * NN, data_block, num_bytes_to_encode);
	
       // copy the parity part //
       memcpy(encoded_data + i * NN + num_bytes_to_encode, data_block + (NN-NROOTS), NROOTS);
    }

#ifdef DEBUG_FEC
    // debug apurv start //
    printf("encoded data %d symbols (symSize: %d)- \n", len_w_fec, syms);
    for(int i = 0; i < len_w_fec; i++)
    {
        printf("%d ", encoded_data[i]);
    }
    printf("\n\n"); fflush(stdout);
    // debug end //
#endif

    /*
    // convert the symSize symbols back into bytes //
    unsigned char *encoded_bytes = (unsigned char *) malloc(sizeof(unsigned char) * len_w_fec);
    memset(encoded_bytes, 0, sizeof(unsigned char) * len_w_fec);

    std::string bits = getBits(encoded_data, len_w_fec, syms);
    int nbytes = convertToSymbols(encoded_bytes, 8, bits);

    free(encoded_data);
 
    std::string s1((const char*) encoded_bytes, nbytes);

    free(encoded_bytes);
    */

    std::string s1((const char*) encoded_data, len_w_fec);

    //printf("making packets complete...... size%d\n", s1.size());
    return s1;
}


int
gr_decode_rs(struct rs *rs, unsigned char *data)
{
  int deg_lambda, el, deg_omega;
  int i, j, r,k;
  DTYPE u,q,tmp,num1,num2,den,discr_r;
  DTYPE lambda[NROOTS+1], s[NROOTS];    /* Err+Eras Locator poly
                                         * and syndrome poly */
  DTYPE b[NROOTS+1], t[NROOTS+1], omega[NROOTS+1];
  DTYPE root[NROOTS], reg[NROOTS+1], loc[NROOTS];
  int syn_error, count;

  /* form the syndromes; i.e., evaluate data(x) at roots of g(x) */
  for(i=0;i<NROOTS;i++)
    s[i] = data[0];

  for(j=1;j<NN;j++){
    for(i=0;i<NROOTS;i++){
      if(s[i] == 0){
        s[i] = data[j];
      } else {
        s[i] = data[j] ^ ALPHA_TO[MODNN(INDEX_OF[s[i]] + (FCR+i)*PRIM)];
      }
    }
  }
  /* Convert syndromes to index form, checking for nonzero condition */
  syn_error = 0;
  for(i=0;i<NROOTS;i++){
    syn_error |= s[i];
    s[i] = INDEX_OF[s[i]];
  }

  if (!syn_error) {
    /* if syndrome is zero, data[] is a codeword and there are no
     * errors to correct. So return data[] unmodified
     */
    count = 0;
    goto finish;
  }
  memset(&lambda[1],0,NROOTS*sizeof(lambda[0]));
  lambda[0] = 1;

  for(i=0;i<NROOTS+1;i++)
    b[i] = INDEX_OF[lambda[i]];

  /*
   * Begin Berlekamp-Massey algorithm to determine error+erasure
   * locator polynomial
   */
  r = 0;
  el = 0;
  while (++r <= NROOTS) {       /* r is the step number */
    /* Compute discrepancy at the r-th step in poly-form */
    discr_r = 0;
    for (i = 0; i < r; i++){
      if ((lambda[i] != 0) && (s[r-i-1] != A0)) {
        discr_r ^= ALPHA_TO[MODNN(INDEX_OF[lambda[i]] + s[r-i-1])];
      }
    }
    discr_r = INDEX_OF[discr_r];        /* Index form */
    if (discr_r == A0) {
      /* 2 lines below: B(x) <-- x*B(x) */
      memmove(&b[1],b,NROOTS*sizeof(b[0]));
      b[0] = A0;
    } else {
      /* 7 lines below: T(x) <-- lambda(x) - discr_r*x*b(x) */
      t[0] = lambda[0];
      for (i = 0 ; i < NROOTS; i++) {
        if(b[i] != A0)
          t[i+1] = lambda[i+1] ^ ALPHA_TO[MODNN(discr_r + b[i])];
        else
          t[i+1] = lambda[i+1];
      }
      if (2 * el <= r - 1) {
        el = r - el;
        /*
         * 2 lines below: B(x) <-- inv(discr_r) *
         * lambda(x)
         */
        for (i = 0; i <= NROOTS; i++)
          b[i] = (lambda[i] == 0) ? A0 : MODNN(INDEX_OF[lambda[i]] - discr_r + NN);
      } else {
        /* 2 lines below: B(x) <-- x*B(x) */
        memmove(&b[1],b,NROOTS*sizeof(b[0]));
        b[0] = A0;
      }
      memcpy(lambda,t,(NROOTS+1)*sizeof(t[0]));
    }
  }


  /* Convert lambda to index form and compute deg(lambda(x)) */
  deg_lambda = 0;
  for(i=0;i<NROOTS+1;i++){
    lambda[i] = INDEX_OF[lambda[i]];
    if(lambda[i] != A0)
      deg_lambda = i;
  }
  /* Find roots of the error+erasure locator polynomial by Chien search */
  memcpy(&reg[1],&lambda[1],NROOTS*sizeof(reg[0]));
  count = 0;            /* Number of roots of lambda(x) */
  for (i = 1,k=IPRIM-1; i <= NN; i++,k = MODNN(k+IPRIM)) {
    q = 1; /* lambda[0] is always 0 */
    for (j = deg_lambda; j > 0; j--){
      if (reg[j] != A0) {
        reg[j] = MODNN(reg[j] + j);
        q ^= ALPHA_TO[reg[j]];
      }
    }
    if (q != 0)
      continue; /* Not a root */
    /* store root (index-form) and error location number */
    root[count] = i;
    loc[count] = k;
    /* If we've already found max possible roots,
     * abort the search to save time
     */
    if(++count == deg_lambda)
      break;
  }
  if (deg_lambda != count) {
    /*
     * deg(lambda) unequal to number of roots => uncorrectable
     * error detected
     */
    count = -1;
    goto finish;
  }
  /*
   * Compute err+eras evaluator poly omega(x) = s(x)*lambda(x) (modulo
   * x**NROOTS). in index form. Also find deg(omega).
   */
  deg_omega = 0;
  for (i = 0; i < NROOTS;i++){
    tmp = 0;
    j = (deg_lambda < i) ? deg_lambda : i;
    for(;j >= 0; j--){
      if ((s[i - j] != A0) && (lambda[j] != A0))
        tmp ^= ALPHA_TO[MODNN(s[i - j] + lambda[j])];
    }
    if(tmp != 0)
      deg_omega = i;
    omega[i] = INDEX_OF[tmp];
  }
  omega[NROOTS] = A0;

  /*
   * Compute error values in poly-form. num1 = omega(inv(X(l))), num2 =
   * inv(X(l))**(FCR-1) and den = lambda_pr(inv(X(l))) all in poly-form
   */
  for (j = count-1; j >=0; j--) {
    num1 = 0;
    for (i = deg_omega; i >= 0; i--) {
      if (omega[i] != A0)
        num1  ^= ALPHA_TO[MODNN(omega[i] + i * root[j])];
    }
    num2 = ALPHA_TO[MODNN(root[j] * (FCR - 1) + NN)];
    den = 0;

    /* lambda[i+1] for i even is the formal derivative lambda_pr of lambda[i] */
    for (i = min(deg_lambda,NROOTS-1) & ~1; i >= 0; i -=2) {
      if(lambda[i+1] != A0)
        den ^= ALPHA_TO[MODNN(lambda[i+1] + i * root[j])];
    }
    if (den == 0) {
      count = -1;
      goto finish;
    }
    /* Apply error to data */
    if (num1 != 0) {
      data[loc[j]] ^= ALPHA_TO[MODNN(INDEX_OF[num1] + INDEX_OF[num2] + NN - INDEX_OF[den])];
    }
  }

  finish:
     if(count > 0)
     {
	//printf("FEC tries to fix!\n"); fflush(stdout);
     }
#ifdef DEBUG_FEC
     printf(" --------------------------------------------\n");
     printf("\nin decode_rs: count:%d\n\n", count); fflush(stdout);
     printf(" --------------------------------------------\n");
#endif
     return count;
}

std::string
gr_decode_rs_fec(std::string buf, int n, int k, int dataLen, 
		 std::string parity_bits_str, 
#ifdef RS_MAC_OPT_VERSION
		 std::map<int, split> fec_split_map,
		 std::vector<int> sorted_orig_pos,
#else
		 std::vector<int> orig_pos_indices, 
#endif
		 int mac_fec_N, int mac_fec_K, 
		 int &total_fec_groups, int &correct_fec_groups, bool mac_fec_flag, 
		 std::string test_buf, std::string test_parity_bits_str
#ifdef PARTIAL_TPUT
		, int &corr_bits, int *final_pos_map, int bps, int subcarriers, float *csi, int hint_flag
#endif
		 )

{
  static std::vector<struct rs*> rs_vec;
  static struct rs *rs = find_rs_handle(rs_vec, n, k);
  assert(rs);
  int syms = rs->mm;

  // printf("nn = %d; nroots=%d, \n", rs->nn, rs->nroots);
  // get bits from the byte string //
  std::string s_bits = getBits((const unsigned char*) buf.data(), buf.size(), 8);

  int num_bits = s_bits.size();
  unsigned char *data = (unsigned char*) malloc(sizeof(unsigned char) * ceil(num_bits/syms));
  memset(data, 0, sizeof(unsigned char) * ceil(num_bits/syms));

  int len = convertToSymbols(data, syms, s_bits);

#ifdef PARTIAL_TPUT
  std::string test_bits = getBits((const unsigned char*) test_buf.data(), test_buf.size(), 8);
#endif

  if(0)
  {
     // debug only //
     std::string test_bits = getBits((const unsigned char*) test_buf.data(), test_buf.size(), 8);
     unsigned char *test_data = (unsigned char*) malloc(sizeof(unsigned char) * ceil(num_bits/syms));
     memset(test_data, 0, sizeof(unsigned char) * ceil(num_bits/syms));
     int len1 = convertToSymbols(test_data, syms, test_bits);
     // debug only ends //
  }

#ifdef DEBUG_FEC
  printf("before ERRORS nsymbols: %d---- bits: %d\n", len, num_bits);
  for(int i = 0; i < len; i++)
     printf("%d ", data[i]);
  printf("\n");
#endif
  //injectErrors((unsigned char*) s_bits.data(), num_bits, sorted_orig_pos); convertToSymbols(data, syms, s_bits);

#ifdef DEBUG_FEC
  printf("after ERRORS nsymbols: %d------ \n", len);
  for(int i = 0; i < len; i++)
     printf("%d ", data[i]);
  printf("\n");

  printf("received data %d symbols (symSize: %d)\n", len, syms);
  for(int i = 0; i < len; i++)
      printf("%d ", data[i]);
  printf("\n\n"); fflush(stdout);
#endif
 
  // printf("dataLen before: %d, syms: %d\n", dataLen, syms); 
  dataLen = dataLen * 8/syms; 
  assert(len > dataLen);
  
  int num_blocks = ceil((float)len/NN);
  // printf("len%d NN%d num_blocks%d dataLen%d\n", len, NN, num_blocks, dataLen);
  fflush(stdout);

  int processedDataLen = 0;
  unsigned char data_block[NN];
  //memset(data_block, 0, NN);
 
  unsigned char *decoded_data = (unsigned char*) malloc(dataLen);
  memset(decoded_data, 0, dataLen);
  int pad_zeros = 0;

  bool mac_parity_done = false;
  total_fec_groups += num_blocks;

  for(int i=0; i<num_blocks; i++)
  {
#ifdef DEBUG
     printf("Decoding block: %d\n", i); fflush(stdout);
#endif
     memset(data_block, 0, NN);
     int num_decoded = 0;
     if((processedDataLen + (NN - NROOTS))< dataLen)
     {
	num_decoded = NN-NROOTS;
     }
     else
     {
	if(i != num_blocks - 1)
	{
	   // flag as incorrect reception //
	   return buf;
	}
	assert(i == num_blocks - 1);
	num_decoded = dataLen - processedDataLen;
     }
    
     // perform the FEC decoding on the current <data_block> now //
     memcpy(data_block, data + i*NN, num_decoded);

     // copy the parity (NROOTS) to the data_block (leave space for padding if reqd)//
     memcpy(data_block + (NN-NROOTS), data + i*NN + num_decoded, NROOTS);
     memcpy(decoded_data + processedDataLen, data_block, num_decoded);		// added now

#ifdef DEBUG_FEC
     printf("undecoded block[%d]: \n", i);
     for(int k = 0; k < NN; k++)
        printf("%d ", data_block[k]);
     printf("\n");
#endif

     int retVal = gr_decode_rs(rs, data_block);

#ifdef PERFORM_MAC_FEC
     if(retVal < 0 && !mac_parity_done && mac_fec_flag)
     {
	bool ret = false;
#ifdef RS_MAC_OPT_VERSION
	//printf("NN: %d, NROOTS: %d, NN-NROOTS: %d\n", NN, NROOTS, NN-NROOTS); fflush(stdout);
	ret = repairUsingMacParity_split_OPT(data + i*NN, NN, (NN-NROOTS), i, num_decoded,
					     parity_bits_str, fec_split_map,
					     sorted_orig_pos,	
			      		     rs->mm, rs_vec); 
					     
					     //test_data + i*NN, test_parity_bits_str);	
#else
	ret = repairUsingMacParity(data + i*NN, NN, (NN-NROOTS), i, num_decoded,
         	 	              parity_bits_str, orig_pos_indices,
                		      mac_fec_N, mac_fec_K, rs->mm, rs_vec);
#endif
	mac_parity_done = true;
	if(ret)
	{
	    // printf("re-decode block: %d\n", i);
	    i--;
            continue;	
        }
     }
     mac_parity_done = false;
#endif	// PERFORM_MAC_FEC



#ifdef PARTIAL_TPUT
     if(retVal < 0 && hint_flag == 1)
     {
	// if the phy-fec group hasn't been repaired even after (possible) mac-fec, count partial throughput //
        int t_bytes = num_decoded;
        if(i == num_blocks - 1)
                t_bytes -= 4;
        std::string bits_block = getBits((const unsigned char*) data_block, t_bytes, 8);


        int start = 0;
        if(i == 0)
                start = 192;

        // printf("i: %d, bits_block.size: %d, NN: %d, NROOTS: %d\n", i, bits_block.size(), NN, NROOTS);
        fflush(stdout);
        int incorr_bits = 0;
        for(int b_index = start; b_index < bits_block.size(); b_index++)
        {
            int orig_pos = i*NN*8 + b_index;
            int t_index =  i*(NN-NROOTS)*8 + b_index + i*(NN-NROOTS);

            if(i > 0)
            {
                 orig_pos += 192;
                 t_index += 192;
            }

            int final_pos = final_pos_map[orig_pos];

            int sym_index = final_pos/bps;
            int sc_index = sym_index%subcarriers;
            float ber  = csi[sc_index];
            if(test_bits.data()[t_index] == bits_block.data()[b_index] && ber < 0.005)
                corr_bits++;
        }
     }
#endif

     if(retVal >= 0)
     {
        correct_fec_groups++;
        memcpy(decoded_data + processedDataLen, data_block, num_decoded);
     }


#ifdef DEBUG_FEC
     // debug start //
     printf("decoded block[%d]: \n", i); 
     for(int k = 0; k < NN; k++) {
	printf("%d ", data_block[k]);
     }
     printf("\n\n"); 
     // debug end //
#endif
     
     // copy the 'decoded' data_block contents into <decoded_data> // 
     //memcpy(decoded_data + processedDataLen, data_block, num_decoded); 
     processedDataLen += num_decoded;
  }
 
  if(processedDataLen != dataLen)
  {
     printf("nn%d nroots%d len%d num_blocks%d processedDataLen%d dataLen%d\n", NN, NROOTS,  
						len, num_blocks, processedDataLen, dataLen);
     assert(false);
  }

#ifdef DEBUG_FEC
  // debug apurv start //
  printf("decoded data - \n");
  for(int i = 0; i < dataLen; i++)
  {
      printf("%d ", decoded_data[i]);
  }
  printf("\n\n"); fflush(stdout);
  // debug end //
#endif

  // convert the symSize symbols back into bytes //
  unsigned char *decoded_bytes = (unsigned char *) malloc(sizeof(unsigned char) * dataLen);
  memset(decoded_bytes, 0, sizeof(unsigned char) * dataLen);

  std::string bits = getBits(decoded_data, dataLen, syms);
  int nbytes = convertToSymbols(decoded_bytes, 8, bits);

  free(decoded_data);

  std::string s1 = std::string((const char*) decoded_bytes, nbytes);
  //printf("processedDataLen%d decodedSize%d\n", processedDataLen, s1.size());
  return s1;
}

std::string
getBits(const unsigned char *data, int len, const int symSize)
{
  std::string orig_s;
  for(int i = 0; i < len; i++)
  {
      std::bitset<sizeof(unsigned char) * 8> s_bits(data[i]);
      std::string tmp_string = s_bits.to_string<char,std::char_traits<char>,std::allocator<char> >();
      orig_s.append(tmp_string.substr(8 - symSize, symSize));
  }
  return orig_s;
}

/* for reed solomon, converts the s_bits into symbols of 'symSize' bits each. Returns
 * the num symbols created in 'symbols'
 */
int 
convertToSymbols(unsigned char *symbols, int symSize, std::string s_bits)
{
   int k = 0;
   for(int i = 0; i < s_bits.size(); i+=symSize)
   {
        std::bitset<sizeof(unsigned char) * 8> bits(s_bits.substr(i, symSize));
        symbols[k] = (unsigned char) bits.to_ulong();
        k++;
   }
   return k;
}

void 
injectErrors(unsigned char *msg, int msg_length, std::vector<int> orig_pos)
{
   // printf("injecting errors at .. \n");
   int num_errors = 20;
   int k = 0;

   for(int i = 0; i < num_errors; i++)
   {
      int pos = orig_pos[i];
      msg[pos] = msg[pos] ^ 1;
      // printf("%d ", pos);
   }
  // printf("\n");
   /*
   for(int i = 0; i < msg_length;)
   {
	for(int j = 0; j < 4; j++)
	{
	    int pos = i + j;
	    msg[pos] = msg[pos] ^ 1;
	    printf("%d ", pos);
	    if(++k == num_errors)
	    {
		i = msg_length;
		break;
	    }
	}
	i+=32;
   }*/
}

/* tries to repair the current phy-FEC block, using mac-parity if it exists */
bool
repairUsingMacParity(unsigned char *data, int n, int k, int fec_group_index, int num_decoded,
		     std::string parity_bits_str, std::vector<int> orig_pos_indices, 
		     int mac_fec_N, int mac_fec_K, int symSize, std::vector<struct rs*> rsVector)
{
   // printf("Trying to repair using MAC parity!!! \n\n"); fflush(stdout);
   int start_pos_fec_group = fec_group_index * n * symSize;
   int end_pos_fec_group = (fec_group_index + 1) * num_decoded * symSize;

   std::string mac_fec_parity_str;
   std::string mac_fec_data_str; 
   std::vector<int> fec_parity_pos;

   int mac_symSize = 0;
   switch(mac_fec_N) {
        case 3:  mac_symSize = 2; break;
        case 15: mac_symSize = 4; break;
        case 255: mac_symSize = 8; break;
        default:
                assert(false);
   }

   // data bits of the phy-layer fec group //
   std::string bits = getBits(data, num_decoded, symSize);

   bool found = false;
   for(int i = 0; i < orig_pos_indices.size(); i++)
   {
	if(orig_pos_indices[i] >= start_pos_fec_group && orig_pos_indices[i] <= end_pos_fec_group)
	{
	    // starting point found for parity, mac_fec_K symbols protected  //
	    for(int j = 0; j < mac_fec_K * mac_symSize; j++)
	    {
		int pos = orig_pos_indices[i+j] - fec_group_index * n;
		mac_fec_data_str.append(1, bits[pos]);
		fec_parity_pos.push_back(pos);
	    }
		
	    // now capture the parity provided //
	    for(int j = 0; j < (mac_fec_N - mac_fec_K) * mac_symSize; j++)
	    {
		mac_fec_parity_str.append(1, parity_bits_str[i+j]);
            } 

	    found = true;
	    break;	
	}
   }

   if(!found)
	return false;

   // printf("MAC parity provided --- \n");
   // printf("indices protected .. \n");
   // for(int i = 0; i < orig_pos_indices.size(); i++)
   //      printf("%d ", orig_pos_indices[i]);
   // printf("\n");
   // 
   // printf("MAC data bits protected -- \n");
   // for(int i = 0; i < mac_fec_data_str.size(); i++)
   //      printf("%c ", mac_fec_data_str[i]);
   // printf("\n");
   // 
   // printf("parity_bits --- \n");
   // for(int i = 0; i < parity_bits_str.size(); i++)
   //      printf("%c ", parity_bits_str[i]);



   // get all the bits of the mac layer fec group //
   mac_fec_data_str.append(mac_fec_parity_str);
  
   assert(mac_fec_data_str.size() == (mac_fec_N) * mac_symSize);
 
   // convert into the symbols //  
   unsigned char mac_fec_symbols[mac_fec_N]; 
   convertToSymbols(mac_fec_symbols, mac_symSize, mac_fec_data_str); 

   struct rs *rs = find_rs_handle(rsVector, mac_fec_N, mac_fec_K);
   assert(rs);

   int retVal = gr_decode_rs(rs, mac_fec_symbols);

	// whoa.. mac FEC passed //
   if(retVal != -1)
   {
	// printf("mac fec passed!!\n"); fflush(stdout);
	// convert the data symbols of the mac-fec-block into bits //
	std::string mac_data_bits = getBits(mac_fec_symbols, mac_fec_K, mac_symSize);
	
	assert(mac_data_bits.size() == fec_parity_pos.size());
	assert(mac_data_bits.size() == mac_fec_K * mac_symSize);

	// printf("data bits of the MAC fec group!.. \n");
	// for(int i = 0; i < mac_data_bits.size(); i++)
	// 	printf("%c ", mac_data_bits[i]);
	// printf("\n");

	// replace the 'bits' in the phy fec group now //
	for(int i = 0; i < fec_parity_pos.size(); i++)
	{
	    bits.replace(fec_parity_pos[i], 1, 1, mac_data_bits[i]); 
	}
   }
   else
   {
	// printf("MAC parity fails.. \n"); 
	return false;
   }
   convertToSymbols(data, symSize, bits);
   return true;
}

/* split version */
bool
repairUsingMacParity_split_OPT(unsigned char *data, int n, int k, int fec_group_index, int num_decoded, 
                               std::string parity_bits_str, std::map<int, split> fec_split_map,
			       std::vector<int> sorted_orig_pos,
                               int symSize,
			       std::vector<struct rs*> rsVector)
			       //unsigned char *test_data, std::string test_parity_bits_str)
{
   // data bits in the phy-layer FEC group //
   int n_data_bits = num_decoded * symSize;
   // printf("@@@@@@@@@@@@@ Trying to repair FEC-group: %d using MAC parity.. n: %d, k: %d n_data_bits: %d!!! \n\n", fec_group_index, n, k, n_data_bits); 
   fflush(stdout);

   // search in the fec_split_map, if this fec group has an entry //
   std::map<int, split>::iterator it = fec_split_map.find(fec_group_index); 
   if(it == fec_split_map.end())
   {
	// printf("fec split map does not have an entry for fec_group: %d\n", fec_group_index);
	return false;
   }

   split split_pair = (*it).second;   
   int r_bad = split_pair.r_bad;
   int d_bad = split_pair.d_bad;
   int r_good = split_pair.r_good;
   int d_good = split_pair.d_good;		// it is possibly padded!
   int d_good_pad = d_good - (n_data_bits - d_bad);
   int d_good_ideal = d_good - d_good_pad;	// actual number of bits!

   int offset = split_pair.indices_offset; 

#ifdef DEBUG
    printf("repairUsingMacParity_split_OPT, parity : \n");
    for(int i = 0; i < parity_bits_str.size(); i++)
         printf("%c", parity_bits_str[i]);
    printf("\n");

   printf("Rfec_group: %d bad: %d, r_bad: %d, good: %d, r_good: %d good_ideal: %d parity_offset: %d\n", fec_group_index, d_bad, r_bad, d_good, r_good, d_good_ideal, split_pair.parity_offset);
   fflush(stdout);
#endif

   // extract the total (bad+good) mac parity bits for this fec_group //
   std::string mac_parity_bits = parity_bits_str.substr(split_pair.parity_offset, (r_bad + r_good));
   //printf("mac_parity_bits size: %d\n", r_bad + r_good);

   // offset found, now extract the actual bad and good data bits //
   std::string bits = getBits(data, num_decoded, symSize);
   bool mac_correction = false;
   std::string bad_bits;
   std::string good_bits;

   /*
   // debug only //
   std::string test_bits = getBits(test_data, num_decoded, symSize);
   std::string test_mac_parity_bits = test_parity_bits_str.substr(split_pair.parity_offset, (r_bad + r_good));
   std::string test_bad_bits;
   std::string test_good_bits;
   // debug end // */
   
   if(r_bad > 0)
   {
	// printf("mac-fec: bad part, r_bad: %d offset: %d n: %d\n", r_bad, offset, n); fflush(stdout);
	//printf("bad data bits --- \n");
	for(int i = 0; i < d_bad; i++)
	{
	    int pos = sorted_orig_pos[offset + i] - fec_group_index * n * symSize;
	    //printf("%d ", pos); //sorted_orig_pos[offset + i]);
	    bad_bits.append(1, bits[pos]);
	    //test_bad_bits.append(1, test_bits[pos]);
	    //printf("%c", bad_bits[i]);
	}
	std::string bad_parity = mac_parity_bits.substr(0, r_bad);

#ifdef DEBUG_MAC_FEC
        printf("bad parity part : \n");
        for(int i = 0; i < bad_parity.size(); i++)
                printf("%c", bad_parity[i]);
        printf("\n");
#endif

	bad_bits.append(bad_parity);
	mac_correction = do_MAC_FEC_decoding(bits, bad_bits, fec_group_index, rsVector,
	    		 		    d_bad + r_bad, d_bad, offset, sorted_orig_pos, 
		 			    0, n * symSize);

	//std::string test_bad_parity = test_mac_parity_bits.substr(0, r_bad);
	//test_bad_bits.append(test_bad_parity);
   }

   if(r_good > 0)
   {
	// printf("mac-fec: good part, r_good: %d \n", r_good); fflush(stdout);
	for(int i = 0; i < d_good_ideal; i++)
	{		
	    int pos = sorted_orig_pos[offset + d_bad + i] - fec_group_index * n * symSize;
	    good_bits.append(1, bits[pos]);
	    //test_good_bits.append(1, test_bits[pos]);
	}
	
	// if d_good_pad > 0, then append '0' to the good_bits
	if(d_good_pad > 0)
	{
		good_bits.append(d_good_pad, '0');	
		//test_good_bits.append(d_good_pad, '0');
	}
   
#ifdef DEBUG_MAC_FEC
	printf("good data part : \n");
	for(int i = 0; i < good_bits.size(); i++)
		printf("%c", good_bits[i]);
	printf("\n");
#endif
	   
        assert(good_bits.size() == d_good);
        // printf(" d_good: %d\n", d_good); fflush(stdout);
 
	std::string good_parity = mac_parity_bits.substr(r_bad, r_good);
	good_bits.append(good_parity);

#ifdef DEBUG_MAC_FEC
        printf("good parity part : \n");
        for(int i = 0; i < good_parity.size(); i++)
                printf("%c", good_parity[i]);
        printf("\n");
#endif

	int _symSize = 2;
	switch(d_good + r_good)
        {
             case 6: _symSize = 2; break;
	     case 21:_symSize = 3; break;
	     case 60: _symSize = 4; break;
	     case 155: _symSize = 5; break;
	     case 378: _symSize = 6; break;
	     case 889: _symSize = 7; break;
	     case 2040: _symSize = 8; break;
	     default: assert(false);

	}

        unsigned char symbols[n];
        // printf("good_bits.size: %d, symSize: %d\n", good_bits.size(), _symSize);
	fflush(stdout);

	mac_correction = do_MAC_FEC_decoding(bits, good_bits, fec_group_index, rsVector,
					     d_good + r_good, d_good, 
		       			     offset + d_bad, sorted_orig_pos, d_good_pad, n * symSize);

        //std::string test_good_parity = test_mac_parity_bits.substr(r_bad, r_good);
        //test_good_bits.append(test_good_parity);
	
   }

   // printf("r_good: %d, d_good: %d, r_bad: %d, d_bad: %d, d_good_pad: %d\n", r_good, d_good, r_bad, d_bad, d_good_pad);
   fflush(stdout);

   // convert back the (corrected) phy-fec bits into the FEC-group symbols //
   if(mac_correction)
   {
#ifdef DEBUG_MAC_FEC
       printf("before correction data block .. \n");
       for(int i = 0; i < num_decoded; i++)
         printf("%d ", data[i]);
       printf("\n");

#endif
       int num = convertToSymbols(data, symSize, bits);
       assert(num == num_decoded);
#ifdef DEBUG_MAC_FEC
       printf("corrected data block .. \n");
       for(int i = 0; i < num; i++)
	 printf("%d ", data[i]);
       printf("\n");
#endif
   }
   /*
   printf("Result fec_group: %d\n", fec_group_index);
   // debug -  print good, bad mac-fec groups and the symbols in error for both //
   printf("bad group: ");
   dumpSymbolsInError(test_bad_bits, bad_bits, d_bad + r_bad, d_bad);
   printf("good group: ");
   dumpSymbolsInError(test_good_bits, good_bits, d_good + r_good, d_good);	
   */

   return true; 
}

bool
do_MAC_FEC_decoding(std::string &phy_fec_bits, std::string mac_fec_bits, int fec_group_index,
		    std::vector<struct rs*> rsVector, int n, int k, int offset, 
		    std::vector<int> sorted_orig_pos, int pad_remove, int phy_fec_N)
{
   // printf("do_MAC_FEC_decoding, n: %d, k: %d, pad_remove: %d, phy_fec_N: %d\n", n, k, pad_remove, phy_fec_N); fflush(stdout);
   int symSize = 2;
   switch(n)
   {
        case 6: symSize = 2; break;
        case 21: symSize = 3; break;
        case 60: symSize = 4; break;
        case 155: symSize = 5; break;
        case 378: symSize = 6; break;
        case 889: symSize = 7; break;
        case 2040: symSize = 8; break;
        default: assert(false);
   }

   int N = n/symSize;
   int K = k/symSize;
   assert(n % symSize == 0);
   assert(k % symSize == 0);
 
   struct rs *rs = find_rs_handle(rsVector, N, K);     
   assert(rs);

   unsigned char symbols[N];
   convertToSymbols(symbols, symSize, mac_fec_bits);

#ifdef DEBUG
   printf("mac-fec block before possible correction .., symSize: %d\n", symSize);
   for(int i = 0; i < mac_fec_bits.size(); i++)
	printf("%c", mac_fec_bits[i]);
   printf("\n");
#endif

#ifdef DEBUG
   printf("mac-fec block before correction :: %d symbols \n", N);
   for(int i = 0; i < N; i++)
     printf("%d ", symbols[i]);
   printf("\n");
#endif

   int retVal = gr_decode_rs(rs, symbols);

   if(retVal >= 0)
   {
#ifdef DEBUG
   printf("mac-fec block after correction :: %d symbols, %d bits\n", N, mac_fec_bits.size());

   for(int i = 0; i < N; i++)
     printf("%d ", symbols[i]);
   printf("\n");
#endif

      std::string corrected_data_bits = getBits(symbols, K, symSize);

      // printf(" ********************* %d MAC FEC BLOCK IS FIXED, size: %d*********************************** \n", retVal, corrected_data_bits.size());
#ifdef DEBUG
      printf("\nmac-fec block after possible correction .. \n");
      for(int i = 0; i < corrected_data_bits.size(); i++)
         printf("%c", corrected_data_bits[i]);
      printf("\n");
      printf("correcting :::: \n");
#endif
      for(int j = 0; j < corrected_data_bits.size() - pad_remove; j++)
      {
          int pos = sorted_orig_pos[offset + j] - fec_group_index * phy_fec_N;
          phy_fec_bits.replace( pos, 1, 1, corrected_data_bits[j]);
	  //printf("%d ", pos);
      }
      return true;
   }
   else
   {
      // printf(" ----------------------- %d MAC FEC BLOCK IS NOT FIXED, --------------------------------------\n", retVal); fflush(stdout);
   }
   return false;
}


struct rs*
find_rs_handle(std::vector<struct rs*> rs_vec, int n, int k)
{
    //printf("finding n: %d, k: %d\n", n, k); fflush(stdout);
    int syms;
    int gfpoly;
    int nn, nroots;

    switch(n){
        case 3:
                syms = 2; gfpoly = 0x7; break;
	case 7:
                syms = 3; gfpoly = 0xb; break;
        case 15:
                syms = 4; gfpoly = 0x13; break;
	case 31:
                syms = 5; gfpoly = 0x25; break;
	case 63:
                syms = 6; gfpoly = 0x43; break;
	case 127:
                syms = 7; gfpoly = 0x89; break;
        case 255:
                syms = 8; gfpoly = 0x11d; break;
        default:
                printf("the length of the block is out of range, FEC cannot be initiated.");
                break;
    }

    std::vector<struct rs*>::iterator it = rs_vec.begin();
    struct rs *rs;

    while(it != rs_vec.end())
    {
        //printf("searching!\n"); fflush(stdout);
        rs = *it;
        if(rs->nn == n && (rs->nroots - rs->nn) == k)
        {
		return rs;
        }
        it++;
    }

    nroots = n - k;
    rs = gr_init_rs(syms, gfpoly, 1, 1, nroots);
    rs_vec.push_back(rs);
    return rs;
}

split
make_split(int r_bad, int d_bad, int r_good, int d_good, int d_indices_offset, int d_parity_offset)
{
    struct split s;
    s.r_bad = r_bad;
    s.d_bad = d_bad;
    s.r_good = r_good;
    s.d_good = d_good;
    s.indices_offset = d_indices_offset;
    s.parity_offset = d_parity_offset;
    return s;
}

void
dumpSymbolsInError(std::string tx_bits, std::string rx_bits, int n, int k)
{
   if(n - k == 0)
	return; 

   int symSize = 2;
   switch(n)
   {
        case 6: symSize = 2; break;
        case 21: symSize = 3; break;
        case 60: symSize = 4; break;
        case 155: symSize = 5; break;
        case 378: symSize = 6; break;
        case 889: symSize = 7; break;
        case 2040: symSize = 8; break;
        default: assert(false);
   }

   int N = n/symSize;
   int K = k/symSize;
   assert(n % symSize == 0);
   assert(k % symSize == 0);

   unsigned char tx_symbols[N];
   convertToSymbols(tx_symbols, symSize, tx_bits);

   unsigned char rx_symbols[N];
   convertToSymbols(rx_symbols, symSize, rx_bits);

   int err = 0;
   for(int i = 0; i < N; i++)
   {
	if(tx_symbols[i] != rx_symbols[i])
	{
		err++;	
	}
   }

#ifdef DEBUG   
   printf("tx_symbols: \n"); 
   for(int i = 0; i < N; i++)
	printf("%d ", tx_symbols[i]);
   
   printf("\n rx_symbols: \n");
   for(int i = 0; i < N; i++)
	printf("%d ", rx_symbols[i]);
   printf("\n");
#endif

   // printf("N: %d, K: %d, R: %d, err: %d\n", N, K, N-K, err); 
   return; 
}
