#include <string>
#include <map>
#include <vector>

// apurv++ starts //
#define DTYPE unsigned char

#define PERFORM_MAC_FEC 1
#define RS_MAC_OPT_VERSION 1
//#define PARTIAL_TPUT 1

/* Reed-Solomon codec control block */
struct rs {
  unsigned int mm;              /* Bits per symbol */
  unsigned int nn;              /* Symbols per block (= (1<<mm)-1) */
  unsigned char *alpha_to;      /* log lookup table */
  unsigned char *index_of;      /* Antilog lookup table */
  unsigned char *genpoly;       /* Generator polynomial */
  unsigned int nroots;     /* Number of generator roots = number of parity symbols */
  unsigned char fcr;        /* First consecutive root, index form */
  unsigned char prim;       /* Primitive element, index form */
  unsigned char iprim;      /* prim-th root of 1, index form */
};

struct split {
  unsigned int r_bad;
  unsigned int d_bad;
  unsigned int r_good;
  unsigned int d_good;
  unsigned int indices_offset;
  unsigned int parity_offset;
};

typedef struct split split;

static inline int modnn(struct rs *rs,int x){
  while (x >= rs->nn) {
    x -= rs->nn;
    x = (x >> rs->mm) + (x & rs->nn);
  }
  return x;
}

#define MODNN(x) modnn(rs,x)

#define MM (rs->mm)
#define NN (rs->nn)
#define ALPHA_TO (rs->alpha_to)
#define INDEX_OF (rs->index_of)
#define GENPOLY (rs->genpoly)
#define NROOTS (rs->nroots)
#define FCR (rs->fcr)
#define PRIM (rs->prim)
#define IPRIM (rs->iprim)
#define A0 (rs->nn)

static struct rs 
*gr_init_rs(unsigned int symsize,unsigned int gfpoly,unsigned int fcr,
            unsigned int prim,unsigned int nroots);
void 
gr_free_rs(void *p);

std::string
gr_decode_rs_fec(std::string buf, int n, int k, int dataLen,
		 std::string parity_bits_str, 
#ifdef RS_MAC_OPT_VERSION
		 std::map<int, split> fec_split_map,
		 std::vector<int> sorted_orig_indices,
#else
		 std::vector<int> orig_pos_indices, 
#endif
                 int mac_fec_N, int mac_fec_K, 
		 int &total_fec_groups, int &correct_fec_groups, bool mac_fec_flag, 
		 std::string test_buf, std::string test_parity_bits_str
#ifdef PARTIAL_TPUT
		, int &corr_bits, int *final_pos_map, int bps, int subarriers, float *csi, int hint_flag
#endif
		 );

std::string
gr_encode_rs_fec(std::string buf, int n, int k);

std::string
getBits(const unsigned char *data, int len, const int symSize);

int
convertToSymbols(unsigned char *symbols, int symSize, std::string bits);

void convertToBytes(unsigned char *symbols, int symSize, std::string s_bits);

void
injectErrors(unsigned char *msg, int msg_length, std::vector<int> orig_pos);

struct rs*
find_rs_handle(std::vector<struct rs*> rs_vec, int n, int k);

bool
repairUsingMacParity(unsigned char *data, int n, int k, int fec_group_index, int num_decoded,
                     std::string parity_bits_str, std::vector<int> orig_pos_indices,
                     int mac_fec_N, int mac_fec_K, int symSize, std::vector<struct rs*> rsVector);

bool
repairUsingMacParity_split_OPT(unsigned char *data, int n, int k, int fec_group_index, int num_decoded, 
                               std::string parity_bits_str, std::map<int, split> fec_split_map,
			       std::vector<int> sorted_orig_pos,
                               int symSize,
			       std::vector<struct rs*> rsVector); 
			       //unsigned char *test_data, std::string test_parity_bits_str);

bool
do_MAC_FEC_decoding(std::string &phy_fec_bits, std::string mac_fec_bits, int fec_group_index,
                    std::vector<struct rs*> rsVector, int n, int k, int offset,
                    std::vector<int> sorted_orig_pos, int pad_remove, int phy_fec_N);

split
make_split(int r_bad, int d_bad, int r_good, int d_good, int d_indices_offset, int parity_offset);

void
dumpSymbolsInError(std::string tx_bits, std::string rx_bits, int n, int k);
